/***************************************************************************/
// Copyright (c) 1996-97 by Matthias Schonlau
/***************************************************************************/

/***************************************************************************/
// GENERAL NOTES FOR THIS FILE
//
// MATRIX ALLOCATION FUNCTIONS:
//      Each of these produces an m by n matrix of the requested
//      type, returning a pointer to the matrix.
//      The matrix is referenced with indexing: [0..m-1][0..n-1]
//
// size_t VERSUS int :
//  the size of the matrix should be size_t rather than int . This
//  may effect the performance when we move to another operating system,
//  epecially when we move to a PC
//
// AUXILIARY ALLOCATION:
//      All of these function have their codes simply copied for different
//      types (*int, *real, **int , **real).  For each type also a list of 
//      addresses (or pointers), a list of contents of the pointers, and 
//	the size of the list need to be established. 
// free : need to return NULL pointers , otherwise an old pointer might 
//        accidentally point to a spot which was allocated for another 
//        pointer subsequently					
// check_alloc : changes consist only of real change to int .
/***************************************************************************/

#include <stdio.h>
#include <math.h>
#include <malloc.h>

#include "general.h"
#include "matrix.h"
#include "alloc.h"

//#define SPEED 0   /* 0 if auxiliary allocation used, 1 otherwise */
int SPEED=1;  // global variable
/* initialization *real does not return a NULL pointer */

/* counters for space allocated / freed */
static  int  rnow=0, inow=0; /* real's allocated right now */  

/* list of pointers */
static void **ptr_list=NULL;/* list keeping track of allocated pointers*/
static int *ptr_size=NULL;  /* list keeping track of length of allocated pointers*/
static int size_list=0; /* size of both lists */

static void ***dptr_list=NULL; /* double pointer list */
static int *dptr_size=NULL;   /* double pointer size */
static int size_dlist=0; /* size of double lists */

static void **ptr_int_list=NULL; /* list allocated pointers for int */
static int *ptr_int_size=NULL; //list with length of allocated pointers for int 
static int size_int_list=0; /* size of both lists for int */

static void ***dptr_int_list=NULL; /* double pointer list for int  */
static int *dptr_int_size=NULL;   /* double pointer size for int */
static int size_int_dlist=0; /* size of double lists for int */


/***************************************************************************/
void print_rnow()
{
  fprintf(stderr,"Allocated right now : %d reals, %d ints \n",rnow,inow);
}
/***************************************************************************/
/*Version:  July 06, 1996 						   */
/*Purpose: allocate int matrix 					  	   */
void alloc(int m, int n, int ***mat)
{
    int	i, **M=NULL;

    M = *mat;
    if (SPEED==0 && seek_list(m,M)!=0)
       nrerror("alloc_int_matrix:Trying to allocate existing matrix pointer.");

    if (SPEED==1)
    {
	/* Produce array of m pointers to the rows */
	M = (int **)malloc(m*sizeof(int*));
	if(M==NULL)
		nrerror("alloc Int Matrix Failure (1)");

        /* Allocate m rows of n reals */
        for (i=0; i<m; i++)
        {
          M[i] = (int *)malloc(n*sizeof(int));
          if(NULL == M[i])
                nrerror("alloc int Matrix Failure");
        }
    }
    else
    {
        M = generic_alloc(0,M,m);
        for (i=0; i<m; i++)
        { 
           /*M[i] might contain some leftover junk that
             happens to be a valid address */
           M[i] = NULL ;
           M[i] = generic_alloc(0,M[i],n);
        }
    }
    *mat= M;
}
/***************************************************************************/
/*Version :  July 06, 1996 						   */
/*Purpose : allocate real matrix					   */
void alloc(int m, int n, real ***mat)
{
    int   i;
    real **M=NULL;

    M= *mat;
    if (SPEED==0) 
      if (seek_list(m,M)!=0)
       nrerror("alloc_real_matrix:Trying to allocate existing matrix pointer.");

    if (SPEED==1)  
    {
        /* Allocate an array of m pointers to the rows */
        M = (real **)malloc(m*sizeof(real*));
        if(NULL == M)
            nrerror("alloc real Matrix Failure (1)");

	/* Allocate m rows of n reals */
        for (i=0; i<m; i++)
        {
	  M[i] = (real *)malloc(n*sizeof(real));
	  if(NULL == M[i])
          {
                //M[20000*m][-2000]=.2;
		nrerror("alloc real Matrix Failure (2)");
          }
        }
    }
    else
    {
        M = generic_alloc(0,M,m);
        for (i=0; i<m; i++)
        {
           /*M[i] might contain some leftover junk that 
             happens to be a valid address */ 
           M[i] = NULL;
           M[i] = generic_alloc(0,M[i],n);
        }
       
    }

    *mat= M;
}
/***************************************************************************/
/*Version:  Jan 29, 1997 						   */
/*Purpose : allocate char matrix 					   */
void alloc(int m,int n,char ***mat)
{
    int   i;
    char **M=NULL;

    M= *mat;

        /* Allocate an array of m pointers to the rows */
        M = (char **)malloc(m*sizeof(char*));
        if(NULL == M)
            nrerror("alloc char Matrix Failure (1)");

        /* Allocate m rows of n chars*/
        for (i=0; i<m; i++)
        {
          M[i] = (char *)malloc(n*sizeof(char));
          if(NULL == M[i])
          {
                nrerror("alloc char Matrix Failure (2)");
          }
        }
    *mat= M;
}
/***************************************************************************/
/*Version:  Feb 10, 1997 						   */
/*Purpose : allocate string matrix 					   */
void alloc(int m,int n,string ***mat)
{
    int   i;
    string **M=NULL;

    M= *mat;

        /* Allocate an array of m pointers to the rows */
        M = (string **)malloc(m*sizeof(string*));
        if(NULL == M)
            nrerror("alloc string Matrix Failure (1)");

        /* Allocate m rows of n strings*/
        for (i=0; i<m; i++)
        {
          M[i] = (string *)malloc(n*sizeof(string));
          if(NULL == M[i])
          {
                nrerror("alloc string Matrix Failure (2)");
          }
        }
    *mat= M;
}
/***************************************************************************/
/*Version:  July 04, 1996 						   */
/*Purpose : allocate int vector 					   */
void alloc(int n, int **v)
{
   int *vec=NULL;
  
   vec = *v;
   if (n==0) nrerror("alloc_int_vector: dimensions == 0");

   if (SPEED == 1)
   {
	vec = (int *)malloc(n*sizeof(int));
	if(vec == NULL)
		nrerror("alloc_int_vector: Failure.");
   }
   else
   {
        vec = generic_alloc(0,vec,n);
   }
   *v = vec;
}

/***************************************************************************/
/*Version:  July 03, 1996 						   */
/*Purpose : allocate real vector                                           */
void alloc(int n, real **v)
{
   real *vec=NULL;

   vec = *v;
   if (n==0) nrerror("alloc_real_vector: dimensions == 0");
   
   if (SPEED == 1)
   {
	vec = (real *)malloc(n*sizeof(real));
	if(vec == NULL)
		nrerror("alloc_real_vector: Failure.");
   }
   else
   {
      vec = generic_alloc(0,vec,n);
   }
   *v = vec;
}
/***************************************************************************/
/*Version:  January 30,1997						   */
/*Purpose : allocate char vector                                           */
void alloc(int n, char **v)
{
   char *vec=NULL;

   vec = *v;
   if (n==0) nrerror("alloc_char_vector: dimensions == 0");

        vec = (char *)malloc(n*sizeof(char));
        if(vec == NULL)
                nrerror("alloc_char_vector: Failure.");
   *v = vec;
}
/***************************************************************************/
/*Version:  Feb 10,1997    	                                            */
/*Purpose : allocate string vector                                           */
void alloc(int n, string **v)
{
   string *vec=NULL;

   vec = *v;
   if (n==0) nrerror("alloc string vector: dimensions == 0");

        vec = (string *)malloc(n*sizeof(string));
        if(vec == NULL)
                nrerror("alloc string vector: Failure.");
   *v = vec;
}
/***************************************************************************/
/*Version:  July 05, 1996 						   */
/*Purpose : free int vector 						   */
void frei(int n, int **v)
{
   int *vec=NULL;

   vec = *v;
   if (SPEED == 1)
      free(vec);
   else
      vec = generic_alloc(n,vec,0);
   *v = vec;
}
/***************************************************************************/
/*Version:  July 05, 1996 						   */
/*Purpose : free real vector 						   */
void frei(int n, real **v)
{
   real *vec=NULL;
  
    vec = *v;
    if (SPEED==1) 
       free(vec);
    else
       vec = generic_alloc(n,vec,0);
    *v = vec;
}
/***************************************************************************/
/*Version:  January 30, 1997                                                  */
/*Purpose : free char vector                                               */
void frei(int n, char **v)
{
   char *vec=NULL;

    vec = *v;
       free(vec);
    *v = vec;
}
/***************************************************************************/
/*Version:  January 30, 1997                                                  */
/*Purpose : free string vector                                               */
void frei(int n, string **v)
{
   string *vec=NULL;

    vec = *v;
       free(vec);
    *v = vec;
}
/***************************************************************************/
/*Version:  July 05, 1996 						   */
/*Purpose : free int Matrix 						   */
void frei(int m, int n, int ***mat)
{
   int i, **M=NULL;

   M = *mat;
   if (SPEED ==1)
   {
        /* Free the m rows of n reals */
        for (i=0; i<m; i++)
      
          free(M[i]);

        /* Free the vector of m pointers to the rows */
	free(M);
   }
   else
   {
        for (i=0; i<m; i++)
           M[i] = generic_alloc(n,M[i],0);
        M = generic_alloc(m,M,0);
   }
   *mat = M;
}
/***************************************************************************/
/*Version:  July 05, 1996 						   */
/*Purpose : free real Matrix 						   */
void frei(int m, int n, real ***mat)
{
    int i;
    real **M=NULL;

    M = *mat;
    if (SPEED ==1) 
    {
        /* Free the m rows of n reals */
        for (i=0; i<m; i++)
	  free(M[i]);
        /* Free the vector of m pointers to the rows */
        free(M);
    }
    else
    {
        for (i=0; i<m; i++)
          M[i] = generic_alloc(n,M[i],0);
        M = generic_alloc(m,M,0);   
    }
    *mat = M;
}
/***************************************************************************/
/*Version:  Jan 29, 1997                                                  */
/*Purpose : free char Matrix                                               */
void frei(int m, int n, char ***mat)
{
    int i;
    char **M=NULL;

    M = *mat;
        /* Free the m rows of n chars*/
        for (i=0; i<m; i++)
          free(M[i]);
        /* Free the vector of m pointers to the rows */
        free(M);
    *mat = M;
}
/***************************************************************************/
/*Version:  Feb 10, 1997                                                  */
/*Purpose : free string Matrix                                               */
void frei(int m, int n, string ***mat)
{
    int i;
    string **M=NULL;

    M = *mat;
        /* Free the m rows of n strings*/
        for (i=0; i<m; i++)
          free(M[i]);
        /* Free the vector of m pointers to the rows */
        free(M);
    *mat = M;
}

/***************************************************************************/
/*Version:  June 28, 1996 						   */
/*Purpose : reallocate real vector					   */
void morealloc(int oldn, real **v, int newn)
{
   real *vec=NULL ;

   vec = *v;
   if (newn==0) nrerror("morealloc: dimensions == 0");

   if (SPEED ==1)
   {
     vec = (real *)realloc(vec,newn*sizeof(real));
     if (vec == NULL)
       nrerror("realloc real Vector Failure.");
   }
   else
     vec = generic_alloc(oldn,vec,newn);
 
   *v = vec;
}
/***************************************************************************/
/*Version:  July 04, 1996 						   */
/*Purpose : reallocate int vector					   */
void morealloc(int oldn, int **v, int newn)
{
   int *vec=NULL ;

   vec = *v;
   if (newn==0) nrerror("realloc_int_vector: dimensions == 0");
   
   if (SPEED ==1)
   {
        vec = (int *)realloc(vec,newn*sizeof(int));
        if (vec == NULL)
                nrerror("realloc int Vector Failure.");
   }
   else 
     vec = generic_alloc(oldn,vec,newn);

   *v = vec;
}
/***************************************************************************/
/*Version:  March 05, 1997                                                  */
/*Purpose : reallocate string vector                                          */
void morealloc(int oldn, string **v, int newn)
{
   string *vec=NULL ;

   vec = *v;
   if (newn==0) nrerror("realloc_string_vector: dimensions == 0");

        vec = (string *)realloc(vec,newn*sizeof(string));
        if (vec == NULL)
                nrerror("realloc string Vector Failure.");

   *v = vec;
}
/************************************************************************/
/*Version: July 06, 1996 						*/
/*Purpose: Reallocation of REAL Matrix while preserving old data        */
void morealloc(int oldm, int oldn, real ***mat, int newm, int newn)
{
  int i;
  real **M=NULL;

  M = *mat;
  if (newm*newn*oldn*oldm==0) 
     nrerror("realloc real matrix: 0 columns and / or 0 rows");

  if (SPEED==1)
  {
      if (newm<oldm) 
      { /* free the superfluous rows, realloc the rest, realloc M */ 
        for (i=0; i<newm; i++)
           if ((M[i]= (real *)realloc(M[i],newn*sizeof(real)))==NULL)
              nrerror("realloc real mat: Allocation failure (1)");
        for (i=newm; i<oldm; i++)
           free(M[i]);
        M = (real **)realloc(M,newm*sizeof(real*));
        if (M==NULL) 
           nrerror("realloc real mat: Allocation failure (2)");
      }
      else
      { /* realloc M, realloc existing rows , alloc the new ones */
        M = (real **)realloc(M,newm*sizeof(real*));
        if (M==NULL)
           nrerror("realloc real mat: Allocation failure (3)");
        for (i=0; i<oldm; i++)
           if ((M[i]= (real *)realloc(M[i],newn*sizeof(real)))==NULL)
              nrerror("realloc real mat: Allocation failure (4)");
        for (i=oldm; i<newm; i++)
           if ((M[i]= (real *)malloc(newn*sizeof(real)))==NULL)
              nrerror("realloc real mat: Allocation failure (5)");;
      }
  }
  else
  {
      if (newm<oldm)
      { /* free the superfluous rows, realloc the rest, realloc M */ 
        for (i=0; i<newm; i++)
          M[i] = generic_alloc(oldn,M[i],newn); 
        for (i=newm; i<oldm; i++)
          M[i] = generic_alloc(oldn,M[i],0);
        M = generic_alloc(oldm,M,newm); 
      } 
      else
      { /* realloc M, realloc existing rows , alloc the new ones */
        M = generic_alloc(oldm,M,newm); 
        for (i=0; i<oldm; i++)
          M[i] = generic_alloc(oldn,M[i],newn);
        for (i=oldm; i<newm; i++)
        {
          /* prevent some junk in M[i] to be on the list accidentally */
          M[i] = NULL;
          M[i] = generic_alloc(0,M[i],newn);
        }
      }
  }
  *mat = M;
}
/************************************************************************/
/*Version: July 06, 1996                                                */
/*Purpose: Reallocation of INT Matrix while preserving old data		*/
void morealloc(int oldm, int oldn, int ***mat, int newm, int newn)
{
  int i, **M=NULL;

  M = *mat;
  if (newm*newn*oldn*oldm==0) 
     nrerror("realloc  int matrix: 0 columns and / or 0 rows");

  if (SPEED==1)
  {
      if (newm<oldm)
      { /* free the superfluous rows, realloc the rest, realloc M */
        for (i=0; i<newm; i++)
           if ((M[i]= (int *)realloc(M[i],newn*sizeof(int)))==NULL)
              nrerror("realloc int mat: Allocation failure (1)");
        for (i=newm; i<oldm; i++)
           free(M[i]);
        M = (int **)realloc(M,newm*sizeof(int*));
        if (M==NULL)
           nrerror("realloc int mat: Allocation failure (2)");
      }
      else
      { /* realloc M, realloc existing rows , alloc the new ones */
        M = (int **)realloc(M,newm*sizeof(int*));
        if (M==NULL)
           nrerror("realloc int mat: Allocation failure (3)");
        for (i=0; i<oldm; i++)
           if ((M[i]= (int *)realloc(M[i],newn*sizeof(int )))==NULL)
              nrerror("realloc int mat: Allocation failure (4)");
        for (i=oldm; i<newm; i++)
           if ((M[i]= (int *)malloc(newn*sizeof(int)))==NULL)
              nrerror("realloc int mat: Allocation failure (5)");;
      }
  }
  else
  {
      if (newm<oldm)
      { /* free the superfluous rows, realloc the rest, realloc M */
        for (i=0; i<newm; i++)
          M[i] = generic_alloc(oldn,M[i],newn);
        for (i=newm; i<oldm; i++)
          M[i] = generic_alloc(oldn,M[i],0);
        M = generic_alloc(oldm,M,newm);
      }
      else
      { /* realloc M, realloc existing rows , alloc the new ones */
        M = generic_alloc(oldm,M,newm);
        for (i=0; i<oldm; i++)
          M[i] = generic_alloc(oldn,M[i],newn);
        for (i=oldm; i<newm; i++)
        {
          /* prevent some junk in M[i] to be on the list accidentally */
          M[i] = NULL;
          M[i] = generic_alloc(0,M[i],newn);
        }
      }
  }
  *mat = M;
}
/***************************************************************************/
/*Version: March 05, 1997                                                  */
/*Purpose: Reallocation of STRING Matrix while preserving old data            */
void morealloc(int oldm, int oldn, string ***mat, int newm, int newn)
{
  int i; 
  string **M=NULL;

  M = *mat;
  if (newm*newn*oldn*oldm==0)
     nrerror("realloc  string matrix: 0 columns and / or 0 rows");

      if (newm<oldm)
      { /* free the superfluous rows, realloc the rest, realloc M */
        for (i=0; i<newm; i++)
           if ((M[i]= (string *)realloc(M[i],newn*sizeof(string )))==NULL)
              nrerror("realloc string mat: Allocation failure (1)");
        for (i=newm; i<oldm; i++)
           free(M[i]);
        M = (string **)realloc(M,newm*sizeof(string*));
        if (M==NULL)
           nrerror("realloc string mat: Allocation failure (2)");
      }
      else
      { /* realloc M, realloc existing rows , alloc the new ones */
        M = (string **)realloc(M,newm*sizeof(string*));
        if (M==NULL)
           nrerror("realloc string mat: Allocation failure (3)");
        for (i=0; i<oldm; i++)
           if ((M[i]= (string *)realloc(M[i],newn*sizeof(string )))==NULL)
              nrerror("realloc string  mat: Allocation failure (4)");
        for (i=oldm; i<newm; i++)
           if ((M[i]= (string *)malloc(newn*sizeof(string)))==NULL)
              nrerror("realloc string mat: Allocation failure (5)");;
      }
  *mat = M;
}
/***************************************************************************/
/*Version: July 03, 1996 -Feb 13, 97                                   	   */
/*Purpose : print list of pointers to real			           */
void print_list(char *comment=" ")
{
   int i;
   
   printf("Pointer List:%s \n",comment);
   for (i=0;i<size_list;i++) {
     // cout << ptr_list[i] << " " << ptr_size[i] << space ;
     printf("%p %d ",ptr_list[i],ptr_size[i]);
   }
   printf("\n");
} 
/***************************************************************************/
/*Version: July 03, 1996 -Feb 13, 97                                       */
/*Purpose : print list of double pointers to real                          */
void print_dlist(char *comment=" ")
{
   int i;

   printf("Double Pointer List: %s \n",comment); 
   for (i=0;i<size_dlist;i++) {
     // cout << dptr_list[i] << " " << dptr_size[i] << space ;
     printf("%p %d ",dptr_list[i],dptr_size[i]);
   }
   printf("\n");
}
/***************************************************************************/
/*Version: July 03, 1996                                                   */
/*Purpose : print list of pointers to int                                  */
void print_int_list(char *comment=" ")
{
   int i;

   printf("Pointer List to int:%s ",comment);
   for (i=0;i<size_int_list;i++) {
     // cout << ptr_int_list[i] << space ;
     printf("%p ",ptr_int_list[i]);
   }
   printf("\n");
   printveci(size_int_list,ptr_int_size,"lengths:");
}
/***************************************************************************/
/*Version: July 03, 1996                                                   */
/*Purpose : print list of double  pointers to int                          */
void print_int_dlist(char *comment=" ")
{
   int i;

   printf("Double Pointer List to int: %s ",comment);
   for (i=0;i<size_int_dlist;i++) {
     // cout << dptr_int_list[i] << space ;
     printf("%p ",dptr_int_list[i]);
   }
   printf("\n");
   printveci(size_int_dlist,dptr_int_size,"lengths:");
}
/***************************************************************************/
/*Version: June 28, 1996                                              	   */
/*Purpose : establish whether POINTER-to-REAL is already allocated, and,   */
/*          if so , whether the size of the pointer is correct 		   */ 
/*input   : n : length of pointer (if exists in list)			   */
/*          ptr : pointer looking for 					   */ 
/*return :  0  pointer not found 					   */
/*          1  pointer found and size is correct			   */
/*          2  pointer found and size is incorrect			   */
/***************************************************************************/
int seek_list(int n,real *vec)
{
   int i;
   
   for (i=0;i< size_list; i++)
      if (ptr_list[i] == (void *) vec)
      {
        if (ptr_size[i] == n ) 
          return 1;
        else 
          return 2;
      }
   return 0;
}
/***************************************************************************/
/*Version: March 03, 1997                                                  */
/*Purpose : establish whether POINTER-to-REAL is already allocated, and,   */
/*          if so , whether the size of the pointer is correct             */
/*input:    ptr : pointer looking for                                      */
/*return :  -1  pointer not found                                           */
/*          length of pointer otherwise 				   */
/***************************************************************************/
int find_list(real *vec)
{
   int i;

   for (i=0;i< size_list; i++)
      if (ptr_list[i] == (void *) vec)
      {
         return ptr_size[i];
      }
   return -1;
}
/***************************************************************************/
/*Version: July 03, 1996                                                   */
/*Purpose : establish whether POINTER-to-int is already allocated, and,   */
/*          if so , whether the size of the pointer is correct             */
/*input   : n : length of pointer (if exists in list)                      */
/*          ptr : pointer looking for                                      */
/*return :  0  pointer not found                                           */
/*          1  pointer found and size is correct                           */
/*          2  pointer found and size is incorrect                         */
/***************************************************************************/
int seek_list(int n,int *vec)
{
   int i;

   for (i=0;i< size_int_list; i++)
      if (ptr_int_list[i] == (void *) vec)
      {
        if (ptr_int_size[i] == n )
          return 1;
        else
          return 2;
      }
   return 0;
}
/***************************************************************************/
/*Version: March 03, 1997                                                  */
/*Purpose : establish whether POINTER-to-int is already allocated, and,   */
/*          if so , whether the size of the pointer is correct             */
/*input:    ptr : pointer looking for                                      */
/*return :  -1  pointer not found                                           */
/*          length of pointer otherwise 				   */
/***************************************************************************/
int find_list(int *vec)
{
   int i;

   for (i=0;i< size_int_list; i++)
      if (ptr_int_list[i] == (void *) vec)
      {
          return ptr_int_size[i];
      }
   return -1;
}
/***************************************************************************/
/*Version: July 03, 1996                                                   */
/*Purpose : establish whether DOUBLE POINTER-to-int is already allocated,  */
/*	    and, if so , whether the size of the pointer is correct        */
/*input   : n : length of pointer (if exists in list)                      */
/*          ptr : pointer looking for                                      */
/*return :  0  pointer not found                                           */
/*          1  pointer found and size is correct                           */
/*          2  pointer found and size is incorrect                         */
/***************************************************************************/
int seek_list(int n,int **vec)
{
   int i;

   for (i=0;i< size_int_dlist; i++)
      if (dptr_int_list[i] == (void **) vec)
      {
        if (dptr_int_size[i] == n )
          return 1;
        else
          return 2;
      }
   return 0;
}
/***************************************************************************/
/*Version: March 03, 1997                                                  */
/*Purpose : establish whether DOUBLE POINTER-to-int is already allocated,  */
/*          and, if so , whether the size of the pointer is correct        */
/*input:    ptr : pointer looking for                                      */
/*return :  -1  pointer not found                                           */
/*          length of pointer otherwise                                    */
int find_list(int **vec)
{
   int i;

   for (i=0;i< size_int_dlist; i++)
      if (dptr_int_list[i] == (void **) vec)
      {
          return dptr_int_size[i];
      }
   return -1;
}
/***************************************************************************/
/*Version: July 03, 1996                                                   */
/*Purpose : establish whether DOUBLE-POINTER-to-real is already allocated, */ 
/*          and, if so , whether the size of the pointer is correct        */
/*input   : n : length of pointer (if exists in list)                      */
/*          ptr : pointer looking for                                      */
/*return :  0  pointer not found                                           */
/*          1  pointer found and size is correct                           */
/*          2  pointer found and size is incorrect                         */
/***************************************************************************/
int seek_list(int n,real **vec)
{
   int i;

   for (i=0;i< size_dlist; i++)
      if (dptr_list[i] == (void **)vec)
      {
        if (dptr_size[i] == n )
          return 1;
        else
        {
          /* cout<<"dpointer: "<<dptr_list[i]<<"old n "<<dptr_size[i]
              <<"#elem "<<i+1<<endl; */
          printf("dpointer: %p old n %d #elem %d\n",dptr_list[i],dptr_size[i],
            i+1);
          return 2;
        }
      }
   return 0;
}
/***************************************************************************/
/*Version: March 03, 1997                                                   */
/*Purpose : establish whether DOUBLE-POINTER-to-real is already allocated, */
/*          and, if so , whether the size of the pointer is correct        */
/*input:    ptr : pointer looking for                                      */
/*return :  -1  pointer not found                                           */
/*          length of pointer otherwise                                    */
int find_list(real **vec)
{
   int i;

   for (i=0;i< size_dlist; i++)
      if (dptr_list[i] == (void **)vec)
      {
        return dptr_size[i];
      }
   return -1;
}
/***************************************************************************/
/* Version: June 28, 1996                                              	   */
/* Purpose : remove / add  the pointer given by vec			   */
/*           and take care of allocation/ reallocation of two global arrays*/
/* Input   :  add  : TRUE add elem				           */
/*                   FALSE remove elem					   */
/*            n    : if TRUE , size of vector to be added  		   */
/*            vec  : pointer used for identification in list               */
void add_list(boolean add,int n,real *vec) 
{
  int i;
  boolean copy;

  if (size_list ==0 && add)
  { /* allocate the global lists with 1 elem */
    size_list++; 
    ptr_list = (void **)malloc(size_list*sizeof(void*));
    if(ptr_list == NULL) nrerror("add_list: Failure (1)");
    ptr_list[0] = (void *) vec;

    ptr_size = (int *)malloc(size_list*sizeof(int));
    if(ptr_size == NULL) nrerror("add_list: Failure (3)");
    ptr_size[0] = n;
  }
  else if (size_list ==1 && !add) 
  {  /* free the global lists*/ 
    free(ptr_size); free(ptr_list); 
    size_list=0; 
  }
  else if (size_list !=0) 
  { /* reallocate the global lists*/
    if (add) size_list++; 
    else     size_list--;
    if (!add)
    { /* copy all elems after 'vec' once spot earlier, overwriting 'vec' */ 
      for (copy=FALSE,i=0; i< size_list+1; i++)
         if (ptr_list[i]== (void *) vec) 
            copy=TRUE;
         else if (copy)
         {
            ptr_list[i-1] = ptr_list[i]; 
            ptr_size[i-1] = ptr_size[i]; 
         }
    } 
    ptr_size = (int *)realloc(ptr_size,size_list*sizeof(int));
    if (ptr_size == NULL) nrerror("add_list: Failure (4)");
    ptr_list = (void **)realloc(ptr_list,size_list*sizeof(void*));
    if (ptr_size == NULL) nrerror("add_list: Failure (5)");
    if (add) 
    { /* add last elem*/
      ptr_list[size_list-1] = (void *) vec;
      ptr_size[size_list-1] = n; 
    }
  }
  else  nrerror("add_list: Code Bug (size_list ==0, add=FALSE ");
}
/***************************************************************************/
/* Version: July 04, 1996                                                  */
/* Purpose : remove / add  the pointer to INT given by vec                 */
/*           and take care of allocation/ reallocation of two global arrays*/
/* Input   :  add  : TRUE add elem                                         */
/*                   FALSE remove elem                                     */
/*            n    : if TRUE , size of vector to be added                  */
/*            vec  : pointer used for identification in list               */
void add_list(boolean add,int n,int *vec)
{
  int i;
  boolean copy;

  if (size_int_list ==0 && add)
  { /* allocate the global lists with 1 elem */
    size_int_list++;
    ptr_int_list = (void **)malloc(size_int_list*sizeof(void*));
    if(ptr_int_list == NULL) nrerror("add_list: Failure (1)");
    ptr_int_list[0] = (void *) vec;

    ptr_int_size = (int *)malloc(size_int_list*sizeof(int));
    if(ptr_int_size == NULL) nrerror("add_list: Failure (3)");
    ptr_int_size[0] = n;
  }
  else if (size_int_list ==1 && !add)
  {  /* free the global lists*/
    free(ptr_int_size); free(ptr_int_list);
    size_int_list=0;
  }
  else if (size_int_list !=0)
  { /* reallocate the global lists*/
    if (add) size_int_list++;
    else     size_int_list--;
    if (!add)
    { /* copy all elems after 'vec' once spot earlier, overwriting 'vec' */
      for (copy=FALSE,i=0; i< size_int_list+1; i++)
         if (ptr_int_list[i]== (void *) vec)
            copy=TRUE;
         else if (copy)
         {
            ptr_int_list[i-1] = ptr_int_list[i];
            ptr_int_size[i-1] = ptr_int_size[i];
         }
    }
    ptr_int_size = (int *)realloc(ptr_int_size,size_int_list*sizeof(int));
    if (ptr_int_size == NULL) nrerror("add_list: Failure (4)");
    ptr_int_list = (void **)realloc(ptr_int_list,size_int_list*sizeof(void*));
    if (ptr_int_size == NULL) nrerror("add_list: Failure (5)");
    if (add)
    { /* add last elem*/
      ptr_int_list[size_int_list-1] = (void *) vec;
      ptr_int_size[size_int_list-1] = n;
    }
  }
  else  nrerror("add_list: Code Bug (size_int_list ==0, add=FALSE ");
}
/***************************************************************************/
/* Version: July 03, 1996                                                  */
/* Purpose : remove / add  the DOUBLE POINTER to REAL given by vec         */
/*           and take care of allocation/ reallocation of two global arrays*/
/* Input   :  add  : TRUE add elem                                         */
/*                   FALSE remove elem                                     */
/*            n    : if TRUE , size of vector to be added                  */
/*            vec  : pointer used for identification in list               */
void add_list(boolean add,int n,real **vec)
{
  int i;
  boolean copy;

  if (size_dlist ==0 && add)
  { /* allocate the global lists with 1 elem */
    size_dlist++;
    dptr_list = (void ***)malloc(size_dlist*sizeof(void**));
    if(dptr_list == NULL) nrerror("add_list dr: Failure (1)");
    dptr_list[0] = (void **) vec; 

    dptr_size = (int *)malloc(size_dlist*sizeof(int)); 
    if(dptr_size == NULL) nrerror("add_list dr: Failure (3)");
    dptr_size[0] = n; 
  }
  else if (size_dlist ==1 && !add) 
  {  /* free the global lists*/
    free(dptr_size); free(dptr_list); 
    size_dlist=0; 
  }
  else if (size_dlist !=0) 
  { /* reallocate the global lists*/
    if (add) size_dlist++; 
    else     size_dlist--; 
    if (!add)
    { /* copy all elems after 'vec' once spot earlier, overwriting 'vec' */
      for (copy=FALSE,i=0; i< size_dlist+1; i++)
         if (dptr_list[i]== (void **) vec) 
            copy=TRUE;
         else if (copy)
         {
            dptr_list[i-1] = dptr_list[i];
            dptr_size[i-1] = dptr_size[i];
         }
    }
    dptr_size = (int *)realloc(dptr_size,size_dlist*sizeof(int));
    if (dptr_size == NULL) nrerror("add_list dr: Failure (4)");
    dptr_list = (void ***)realloc(dptr_list,size_dlist*sizeof(void**));
    if (dptr_size == NULL) nrerror("add_list dr: Failure (5)");
    if (add)
    { /* add last elem*/
      dptr_list[size_dlist-1] = (void **) vec;
      dptr_size[size_dlist-1] = n;
    }
  }
  else  nrerror("add_list dr: Code Bug (size_dlist ==0, add=FALSE ");
}
/***************************************************************************/
/* Version: July 04, 1996                                                  */
/* Purpose : remove / add  the DOUBLE POINTER to INT given by vec          */
/*           and take care of allocation/ reallocation of two global arrays*/
/* Input   :  add  : TRUE add elem                                         */
/*                   FALSE remove elem                                     */
/*            n    : if TRUE , size of vector to be added                  */
/*            vec  : pointer used for identification in list               */
void add_list(boolean add,int n,int **vec)
{
  int i;
  boolean copy;

  if (size_int_list ==0 && add)
  { /* allocate the global lists with 1 elem */
    size_int_dlist++;
    dptr_int_list = (void ***)malloc(size_int_dlist*sizeof(void**));
    if(dptr_int_list == NULL) nrerror("add_list dr: Failure (1)");
    dptr_int_list[0] = (void **) vec;

    dptr_int_size = (int *)malloc(size_int_dlist*sizeof(int));
    if(dptr_int_size == NULL) nrerror("add_list dr: Failure (3)");
    dptr_int_size[0] = n;
  }
  else if (size_int_dlist ==1 && !add)
  {  /* free the global lists*/
    free(dptr_int_size); free(dptr_int_list);
    size_int_dlist=0;
  }
  else if (size_int_dlist !=0)
  { /* reallocate the global lists*/
    if (add) size_int_dlist++;
    else     size_int_dlist--;
    if (!add)
    { /* copy all elems after 'vec' once spot earlier, overwriting 'vec' */
      for (copy=FALSE,i=0; i< size_int_dlist+1; i++)
         if (dptr_int_list[i]== (void **) vec)
            copy=TRUE;
         else if (copy)
         {
            dptr_int_list[i-1] = dptr_int_list[i];
            dptr_int_size[i-1] = dptr_int_size[i];
         }
    }
    dptr_int_size = (int *)realloc(dptr_int_size,size_int_dlist*sizeof(int));
    if (dptr_int_size == NULL) nrerror("add_list dr: Failure (4)");
    dptr_int_list = (void ***)realloc(dptr_int_list,size_int_dlist*sizeof(void**));
    if (dptr_int_size == NULL) nrerror("add_list dr: Failure (5)");
    if (add)
    { /* add last elem*/
      dptr_int_list[size_int_dlist-1] = (void **) vec;
      dptr_int_size[size_int_dlist-1] = n;
    }
  }
  else  nrerror("add_list dr: Code Bug (size_int_dlist ==0, add=FALSE ");
}
/***************************************************************************/
/* Version: July 23, 1996                                                  */
/* Purpose: (1) allocate/reallocate pointer to REAL 			   */
/*          (2) maintain a list with all presently allocated pointers and  */
/*	    and their (vector) sizes,				           */
/*          (3) check list for inconsistencies				   */
/* Input:   oldn ==0   => allocate					   */
/*          newn ==0   => free 						   */
/*          oldn !=0, newn !=0   => reallocate			           */
/* Note :   We do not check for NULL pointers since free() does not produce*/
/*  	    a NULL pointer (and vec is not passed up again) 		   */
real *generic_alloc(int oldn,real *vec,int newn)
{
    int seek;

    if (vec != NULL) seek = seek_list(oldn,vec);
    else             seek = 0;

    if (oldn==0 && newn!=0) 
    { /* check list, ALLOCATE vec, add to list */
      if (seek!= 0)
         nrerror("generic_alloc: (1) Trying to allocate existing pointer. \n \t (2) Pointer not initialized to NULL");
     
      vec = (real *)malloc(newn*sizeof(real));
      if(vec == NULL)
         nrerror("generic_real_alloc: Failure.");
      add_list(TRUE,newn,vec);
      rnow += newn;
    }
    else if (vec!=NULL && oldn!=0 && newn!=0)
    {  /* check list , REALLOCATE , change list */
      if (seek==0)
        nrerror("generic_alloc: Trying to reallocate nonexisting pointer.");
      else if (seek==2)
        nrerror("generic_alloc: Trying to reallocate with incorrect old n");

      add_list(FALSE,oldn,vec); /* removing the old pointer */
      vec = (real *)realloc(vec,newn*sizeof(real));
      if (vec == NULL)
        nrerror("generic_alloc: realloc real Vector Failure.");
      add_list(TRUE,newn,vec);  /* adding the (different) new pointer */
      rnow += -oldn+newn;
    }
    else if (oldn!=0 && newn==0)   // don't add && vec!= NULL, leads to error
    { /*  check list, FREE vector, remove from list */
      if (seek==2) 
        nrerror("generic_alloc: Trying to free pointer with incorrect old n");
      else if (seek==0)
        nrerror("generic_alloc: (1) Trying to free nonexisting pointer. \n \t (2) Pointer reAssignment without change in list");
      add_list(FALSE,oldn,vec);  /* removing the pointer */
      free(vec); 
      vec = NULL;
      rnow -= oldn;
    }
    else  
    {
      /*fprintf(stderr,"oldn:" << oldn << " newn:" << newn << " vec address:" << vec << " seek: " << seek << endl; */  
      fprintf(stderr,"oldn: %d newn: %d vec address: %p seek: %d\n",
        oldn,newn,vec,seek);
      nrerror("generic_alloc: Code Bug");
    }
   return(vec);
}
/***************************************************************************/
/* Version: July 04, 1996                                                  */
/* Purpose: (1) allocate/reallocate  pointer to INT                        */
/*          (2) maintain a list with all presently allocated pointers and  */
/*          and their (vector) sizes,                                      */
/*          (3) check list for inconsistencies                             */
/* Input:   oldn ==0   => allocate                                         */
/*          newn ==0   => free                                             */
/*          oldn !=0, newn !=0   => reallocate                             */
/* Note :   We do not check for NULL pointers since free() does not produce*/
/*          a NULL pointer (and vec is not passed up again)                */
int *generic_alloc(int oldn,int *vec,int newn)
{
    int seek;

    if (vec != NULL) seek = seek_list(oldn,vec);
    else             seek=0;

    if (oldn==0 && newn!=0)
    { /* check list, ALLOCATE vec, add to list */
      if (seek!= 0)
         nrerror("generic_alloc i: (1) Trying to allocate existing pointer. \n (2) Pointer not initialized to NULL ");
      vec = (int *)malloc(newn*sizeof(int));
      if(vec == NULL)
         nrerror("generic_real_alloc i: Failure.");
      add_list(TRUE,newn,vec);
      inow += newn;
    }
    else if (vec!=NULL && oldn!=0 && newn!=0)
    {  /* check list , REALLOCATE , change list */
      if (seek==0)
        nrerror("generic_alloc i: Trying to reallocate nonexisting pointer.");
      else if (seek==2)
        nrerror("generic_alloc i: Trying to reallocate with incorrect old n");

      add_list(FALSE,oldn,vec); /* removing the old pointer */
      vec = (int *)realloc(vec,newn*sizeof(int));
      if (vec == NULL)
        nrerror("generic_alloc i: realloc real Vector Failure.");
      add_list(TRUE,newn,vec);  /* adding the (different) new pointer */
      inow += -oldn+newn;
    }
    else if (vec!=NULL && oldn!=0 && newn==0)
    { /*  check list, FREE vector, remove from list */
      if (seek==2)
        nrerror("generic_alloc i: Trying to free pointer with incorrect old n");
      else if (seek==0)
        nrerror("generic_alloc i: (1) Trying to free nonexisting pointer. \n (2) Pointer reassignment without change in list ");
      /*Note: nonexisting pointer could mean forgot to pass pointer back */
      add_list(FALSE,oldn,vec);  /* removing the pointer */
      free(vec);
      vec = NULL;
      inow -= oldn;
    }
    else  nrerror("generic_alloc i: Code Bug");
   return(vec);
}
/***************************************************************************/
/* Version: July 03, 1996                                                  */
/* Purpose: (1) allocate/reallocate DOUBLE ptr to REAL 			   */
/*          (2) maintain a list with all presently allocated pointers and  */
/*          and their (vector) sizes,                                      */
/*          (3) check list for inconsistencies                             */
/* Input:   oldn ==0   => allocate                                         */
/*          newn ==0   => free                                             */
/*          oldn !=0, newn !=0   => reallocate                             */
/* Note :   We do not check for NULL pointers since free() does not produce*/
/*          a NULL pointer (and vec is not passed up again)                */
real **generic_alloc(int oldn,real **mat,int newn)
{
    int seek;

    if (mat != NULL) seek = seek_list(oldn,mat);
    else             seek = 0;

    if (oldn==0 && newn!=0)
    { /* check list, ALLOCATE mat, add to list */
      if (seek!= 0)
         nrerror("generic_alloc dr: (1) Trying to allocate existing pointer. \n (2) Pointer not initialized to NULL");
      mat = (real **)malloc(newn*sizeof(real *));
      if(mat == NULL)
         nrerror("generic_real_alloc dr: Failure.");
      add_list(TRUE,newn,mat);
    }
    else if (mat!=NULL && oldn!=0 && newn!=0)
    {  /* check list , REALLOCATE , change list */
      if (seek==0)
        nrerror("generic_alloc dr: Trying to reallocate nonexisting pointer.");
      else if (seek==2)
      {
        print_dlist();
        nrerror("generic_alloc dr: Trying to reallocate with incorrect old n");
      }

      add_list(FALSE,oldn,mat); /* removing the old pointer */
      mat = (real **)realloc(mat,newn*sizeof(real *));
      if (mat == NULL)
        nrerror("generic_alloc dr: realloc real Vector Failure.");
      add_list(TRUE,newn,mat);  /* adding the (different) new pointer */
    }
    else if (mat!=NULL && oldn!=0 && newn==0)
    { /*  check list, FREE matrix, remove from list */
      if (seek==2)
        nrerror("generic_alloc dr:Trying to free pointer with incorrect old n");
      else if (seek==0)
        nrerror("generic_alloc dr: (1) Trying to free nonexisting pointer. \n (2) pointer reassignment without change in list");
      /*Note: nonexisting ptr could mean forgot to pass pointer back earlier*/
      add_list(FALSE,oldn,mat);  /* removing the pointer */
      free(mat);
      mat = NULL;
    }
    else  nrerror("generic_alloc dr: Code Bug.");
   return(mat);
}
/***************************************************************************/
/* Version: July 03, 1996                                                  */
/* Purpose: (1) allocate/reallocate DOUBLE ptr to INT 		   	   */
/*          (2) maintain a list with all presently allocated pointers and  */
/*          and their (vector) sizes,                                      */
/*          (3) check list for inconsistencies                             */
/* Input:   oldn ==0   => allocate                                         */
/*          newn ==0   => free                                             */
/*          oldn !=0, newn !=0   => reallocate                             */
/* Note :   We do not check for NULL pointers since free() does not produce*/
/*          a NULL pointer (and vec is not passed up again)                */
int **generic_alloc(int oldn,int **mat,int newn)
{
    int seek;

    if (mat != NULL) seek = seek_list(oldn,mat);
    else             seek = 0;

    if (oldn==0 && newn!=0)
    { /* check list, ALLOCATE mat, add to list */
      if (seek!= 0)
         nrerror("generic_alloc di: (1) Trying to allocate existing pointer. \n (2) Pointer not initialized to NULL");
      mat = (int **) malloc(newn*sizeof(int *));
      if(mat == NULL)
         nrerror("generic_int_alloc di: Failure.");
      add_list(TRUE,newn,mat);
    }
    else if (mat!=NULL && oldn!=0 && newn!=0)
    {  /* check list , REALLOCATE , change list */
      if (seek==0)
        nrerror("generic_alloc di: Trying to reallocate nonexisting pointer.");
      else if (seek==2)
        nrerror("generic_alloc di: Trying to reallocate with incorrect old n");

      add_list(FALSE,oldn,mat); /* removing the old pointer */
      mat = (int **)realloc(mat,newn*sizeof(int *));
      if (mat == NULL)
        nrerror("generic_alloc di: realloc int Matrix Failure.");
      add_list(TRUE,newn,mat);  /* adding the (different) new pointer */
    }
    else if (mat!=NULL && oldn!=0 && newn==0)
    { /*  check list, FREE matrix, remove from list */
      if (seek==2)
        nrerror("generic_alloc di:Trying to free pointer with incorrect old n");
      else if (seek==0)
        nrerror("generic_alloc di: (1) Trying to free nonexisting pointer. \n (2) POinter reassignment without change in list");
      /*Note: nonexisting ptr could mean forgot to pass pointer back earlier*/
      add_list(FALSE,oldn,mat);  /* removing the pointer */
      free(mat);
      mat = NULL;
    }
    else  nrerror("generic_alloc di: Code Bug.");
   return(mat);
}
/***************************************************************************/
/* Version: March 03, 1997                                                 */
/* Purpose: (1) check whether real Matrix is in allocated-list		   */
/*          (2) check whether indices are in allowed range for that matrix */
/* input  : i, j    indices for use in matrix			           */	
/*   	    M	    pointer of matrix					   */
/* Note : valid range for indices is presumed to be 0<=i<imax, 0<=j<jmax   */
void check_alloc(real **M, int i,int j, char label[])
{
   int imax, jmax;

   if (SPEED==1) return; // allocation lists don't exist 
   if (M==NULL)
   {
      printf("%s\n",label);
      nrerror("check_alloc real matrix: M == NULL");
   }
   imax=find_list(M);
   if (imax== -1 && i==0)
   { // needs to come before next if because changes imax 
       fprintf(stdout,"WARNING: %s:1st row not allocated, \n\tsuspected use of &vec as real matrix\n",label);  
       imax=1;
   }
   if (0<=i && i<imax) 
      jmax = find_list(M[i]);
   else jmax= -2; // not available because i out of range
   if (imax== -1 || jmax== -1 || j>=jmax || i>=imax || i<0 || j<0 || jmax== -2) 
   { 
     printf("%s\n",label);
     fprintf(stdout,"Indices Used (start at 0) : i     = %d j     = %d \n",i,j);
     if (jmax == -2)
       fprintf(stdout,"Allocated Dimensions      : n_rows= %d n_cols= NA\n",imax);
     else
       fprintf(stdout,"Allocated Dimensions      : n_rows= %d n_cols= %d \n",imax,jmax);
     nrerror("check_alloc real matrix:\n1 or more indeces out of range or matrix not allocated (-1).");
   }
}
/***************************************************************************/
/* Version: March 03, 1997                                                 */
/* Purpose: (1) check whether int Matrix is in allocated-list             */
/*          (2) check whether indices are in allowed range for that matrix */
/* input  : i, j    indices for use in matrix                              */
/*          M       pointer of matrix                                      */
/* Note : valid range for indices is presumed to be 0<=i<imax, 0<=j<jmax   */
void check_alloc(int **M, int i,int j, char label[])
{
   int imax, jmax;

   if (SPEED==1) return; // allocation lists don't exist
   if (M==NULL)
   {
      printf("%s\n",label);
      nrerror("check_alloc int matrix: M == NULL");
   }
   imax=find_list(M);
   if (imax== -1 && i==0)
   { // needs to come before next if because changes imax
       fprintf(stdout,"WARNING: %s:1st row not allocated, \n\tsuspected use of &vec as int matrix\n",label);
       imax=1;
   }
   if (0<=i && i<imax)
      jmax = find_list(M[i]);
   else jmax= -2; // not available because i out of range
   if (imax== -1 || jmax== -1 || j>=jmax || i>=imax || i<0 || j<0 || jmax== -2)
   {
     printf("%s\n",label);
     fprintf(stdout,"Indices Used (start at 0) : i     = %d j     = %d \n",i,j);
     if (jmax == -2)
       fprintf(stdout,"Allocated Dimensions      : n_rows= %d n_cols= NA\n",imax);
     else
       fprintf(stdout,"Allocated Dimensions      : n_rows= %d n_cols= %d \n",imax,jmax);
     nrerror("check_alloc int matrix:\n1 or more indeces out of range or matrix not allocated (-1).");
   }
}
/***************************************************************************/
/* Version: March 03, 1997                                                 */
/* Purpose: (1) check whether real Vector is in allocated-list             */
/*          (2) chack whether indices are in allowed range for that vector */
/* input  : i       indices for use in vector 				   */
/*          v       pointer of vector 					   */
/* Note : valid range for indices is presumed to be 0<=i<imax	 	   */
void check_alloc(real *v, int i,char label[])
{
   int imax;

   if (SPEED==1) return; // allocation lists don't exist 
   if (v==NULL) 
   {
      printf("%s\n",label);
      nrerror("check_alloc real vector: v == NULL");
   }
   if ((imax=find_list(v))== -1)
   {
         printf("%s\n",label);
         nrerror("check_alloc real vector: Vector is not allocated.");
   }
   if (i>=imax || i<0)
   {
      printf("%s\n",label);
      fprintf(stdout,"i= %d n_elems= %d (need 0<i<n_elems)\n",i,imax);
      nrerror("check_alloc real Vector: index out of range");
   }
}
/***************************************************************************/
/* Purpose: (1) check whether int Vector is in allocated-list              */
/*          (2) chack whether indices are in allowed range for that vector */
/* input  : i       indices for use in vector 				   */
/*          v       pointer of vector 				      	   */
/* Note : valid range for indices is presumed to be 0<=i<imax              */
void check_alloc(int *v, int i,char label[])
{
   int imax;

   if (SPEED==1) return; // allocation lists don't exist
   if (v==NULL)
   {
      printf("%s\n",label);
      nrerror("check_alloc int vector: v == NULL");
   }
   if ((imax=find_list(v))== -1)
   {
         printf("%s\n",label);
         nrerror("check_alloc int vector: Vector is not allocated.");
   }
   if (i>=imax || i<0)
   {
      printf("%s\n",label);
      fprintf(stdout,"i= %d n_elems= %d (need 0<i<n_elems)\n",i,imax);
      nrerror("check_alloc int Vector: index out of range");
   }
}
/***************************************************************************/
/*
void main()
{
 real *vec, *vec2;
 real **mat, **mat2;
 int n,m;

SPEED = 0;
n=2; m=3;
print_dlist("2d"); print_list("2");
 alloc(n,m,&mat); 
 alloc(n,m,&mat2); 
 mat[0][0] = 1.0; mat[0][1] =2.0;  mat[0][2] =3.0; 
 mat[1][0] = 4.0; mat[1][1] =5.0;  mat[1][2] =6.0;
printmatr(n,m,mat,"mat");
print_dlist("2d"); print_list("2");
check_alloc(mat,1,2,"mat1");
	alloc(n,&vec);
	vec[0]= 2; vec[1]=3;
	alloc(m,&vec2);
	vec2[0]= 21; vec2[1]=31; vec2[2]=40;
check_alloc(&vec,0,1,"vec");
}
*/
