#include <math.h>
#include <stdio.h> // for fprintf error

/* only needed to define real */
#include "general.h"
#include "alloc.h"
#include "matrix.h"

/************************************************************************/
/* Version : Nov 25, 1996 						*/
/* Purpose : choleski decomposition for a pos. definite symmetric matrix*/
/* input   : **mat  n,n sym. pos. definite matrix 			*/
/* output  : mat    choleski Decomposition (T'T=m) 			*/
/* WARNING : mat is overwritten   					*/
/* aux funcs: choldc							*/
void choleski(real **mat, int n)
{
  int i,j;
  real *diag=NULL; 

      alloc(n,&diag);

      /* get the choleski decomposition and put into mat*/
      choldc(mat,n,diag);

      /* put diag into mat and clean mat to make it lower triangular */
      for (i=0; i<n; i++)
      {
        mat[i][i]= diag[i];
        for (j=i+1; j<n; j++)
          mat[i][j]= 0.0;
      }
      frei(n, &diag);
}

/************************************************************************/
/* Version : Jan 23, 1996 						*/
/* c program that obtains a choleski decomposition for a pos. definite matrix */
/* NOTE : all [x] have been changed to [x-1] in all arrays 		*/
/*           floats changed to real					*/
/* WARNING: (sum <= 1^e-30) allows the root of slightly negative numbers*/
/*  	(the results will be 0). This had to be done becuase otherwise  */
/*	we were crashing several times					*/
void choldc(real **a, int n, real p[])
/* only the upper triangle of "a" need to be given. It is not modified. 
 The choleski decomposition is returned  in the lower triangle of a, with 
 the exception of the diagonal elements which are returned in p[0..(n-1)]*/ 
{
   void nrerror(char error_text[]);
   int i,j,k;
   real sum;

   for(i=1 ; i<=n; i++) 
   {
     for (j=i;j<=n; j++) 
     {
       for (sum=a[i-1][j-1], k=i-1; k>=1; k--) 
         sum -= a[i-1][k-1] * a[j-1][k-1];
       if (i==j) 
       {
         if (sum <= 1.0e-30) 
         {
           fprintf(stderr,"sum: %f ",sum);
           nrerror("choldc failed, numerically not positive definite");
         }
         p[i-1]=sqrt(sum);
       } 
       else a[j-1][i-1] = sum/ p[i-1];
     }
   }
}
  
