/* eigenvalues and eigenvectors */

#include <math.h>
#include <malloc.h>  // otherwise won't compile on setosa 
#include <stdlib.h>

#include "general.h"
#include "matrix.h"
#include "alloc.h"

/***********************************************************************/
/* computes (a**2 + b**2)^(.5) without destructive underflow or overflow */
/* SQR defined in general.h */
#define NRANSI
real pythag(real a, real b)
{
 	real absa,absb;
	absa=fabs(a);
	absb=fabs(b);
	if (absa > absb) return absa*sqrt(1.0+SQR(absb/absa));
	else return (absb == 0.0 ? 0.0 : absb*sqrt(1.0+SQR(absa/absb)));
}
#undef NRANSI
/***********************************************************************/
#define NRANSI
/* SIGN defined in general.h */

void tqli(real d[], real e[], int n, real **z)
{
	real pythag(real a, real b);
	int m,l,iter,i,k;
	real s,r,p,g,f,dd,c,b;

	for (i=1;i<=n-1;i++) e[i-1]=e[i];
	e[n-1]=0.0;
	for (l=0;l<=n-1;l++) {
		iter=0;
		do {
			for (m=l;m<=n-2;m++) {
				dd=fabs(d[m])+fabs(d[m+1]);
				if ((real)(fabs(e[m])+dd) == dd) break;
			}
			if (m != l) {
				if (iter++ == 30) nrerror("Too many iterations in tqli");
				g=(d[l+1]-d[l])/(2.0*e[l]);
				r=pythag(g,1.0);
				g=d[m]-d[l]+e[l]/(g+SIGN(r,g));
				s=c=1.0;
				p=0.0;
				for (i=m-1;i>=l;i--) {
					f=s*e[i];
					b=c*e[i];
					e[i+1]=(r=pythag(f,g));
					if (r == 0.0) {
						d[i+1] -= p;
						e[m]=0.0;
						break;
					}
					s=f/r;
					c=g/r;
					g=d[i+1]-p;
					r=(d[i]-g)*s+2.0*c*b;
					d[i+1]=g+(p=s*r);
					g=c*r-b;
					for (k=0;k<=n-1;k++) {
						f=z[k][i+1];
						z[k][i+1]=s*z[k][i]+c*f;
						z[k][i]=c*z[k][i]-s*f;
					}
				}
				if (r == 0.0 && i >= l) continue;
				d[l] -= p;
				e[l]=g;
				e[m]=0.0;
			}
		} while (m != l);
	}
}
#undef NRANSI
/***********************************************************************/
void tred2(real **a, int n, real d[], real e[])
{
	int l,k,j,i;
	real scale,hh,h,g,f;

	for (i=n-1;i>=1;i--) {
		l=i-1;
		h=scale=0.0;
		if (l > 0) {
			for (k=0;k<=l;k++)
				scale += fabs(a[i][k]);
			if (scale == 0.0)
				e[i]=a[i][l];
			else {
				for (k=0;k<=l;k++) {
					a[i][k] /= scale;
					h += a[i][k]*a[i][k];
				}
				f=a[i][l];
				g=(f >= 0.0 ? -sqrt(h) : sqrt(h));
				e[i]=scale*g;
				h -= f*g;
				a[i][l]=f-g;
				f=0.0;
				for (j=0;j<=l;j++) {
					a[j][i]=a[i][j]/h;
					g=0.0;
					for (k=0;k<=j;k++)
						g += a[j][k]*a[i][k];
					for (k=j+1;k<=l;k++)
						g += a[k][j]*a[i][k];
					e[j]=g/h;
					f += e[j]*a[i][j];
				}
				hh=f/(h+h);
				for (j=0;j<=l;j++) {
					f=a[i][j];
					e[j]=g=e[j]-hh*f;
					for (k=0;k<=j;k++)
						a[j][k] -= (f*e[k]+g*a[i][k]);
				}
			}
		} else
			e[i]=a[i][l];
		d[i]=h;
	}
	d[0]=0.0;
	e[0]=0.0;
	/* Contents of this loop can be omitted if eigenvectors not
			wanted except for statement d[i]=a[i][i]; */
	for (i=0;i<=n-1;i++) {
		l=i-1;
		if (d[i]) {
			for (j=0;j<=l;j++) {
				g=0.0;
				for (k=0;k<=l;k++)
					g += a[i][k]*a[k][j];
				for (k=0;k<=l;k++)
					a[k][j] -= g*a[k][i];
			}
		}
		d[i]=a[i][i];
		a[i][i]=1.0;
		for (j=0;j<=l;j++) a[j][i]=a[i][j]=0.0;
	}
}

/**********************************************************************/
/* Version : May 19, 1996 */
/* on input :   mat : real symmetric squared matrix 
 		n   : dimension of matrix
                d   : junk vector size n 
   on outut :   d :   eigenvalues (not sorted by size [eigsort would do that]) 
		mat : normalized eigenvectors (in columns) in same order as  
			eigenvalues
 		      input matrix is DESTROYED
		      
*/
void eigen(real **mat, int n, real *d)
{
  void tqli(real d[], real e[], int n, real **z);
  void tred2(real **a, int n, real d[], real e[]);
  real *e=NULL; /* intermediate junk */

  //if (!symmetry_realmat_check(n,mat))
  //  nrerror("eigen: Matrix is not symmetric.");
  alloc(n,&e);
  tred2(mat, n, d,e);
  tqli(d,e,n ,mat);
  frei(n,&e);
}
/**********************************************************************/
/*
void main()
{
  int size;
  real **data=NULL;
  real *mean=NULL;
  real *d=NULL;
 
  size=readmat(&data,&mean);
  alloc(size,&d);
  eigen(data,size,d);
  printvecr(size,d,"main : eigenvalues");
}
*/
