/* Version : April 03, 1995 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <unistd.h> /* needed e.g. for getpid*/
#include <sys/types.h>

#include "general.h"
#include "matrix.h"
#include "alloc.h"

#define RandomGenerator drand48() 
//#define RandomGenerator RandUnif() 

double drand48( void );
void srand48( long );

/***********************************************************************/
/* Version : Early 1996
/* returns a nomally distributed deviate with mean 0 and unit variance */
/* test normality with plot(qnorm(ppoints(out)),sort(out)) in Splus */
/* reference : Numerical Recipes in c , p 289-290  */
float gasdev()
{
static int iset=0;
static float gset;
float  fac,rsq,v1,v2;

 if (iset ==0)
 {
   /* Pick to uniform numbers in the square exptending from -1 to 1 in each
   direction, see if they are in the unit circle, and if they are not ,
   try again. */
   do {
     v1=2.0 * RandomGenerator -1.0;
     v2=2.0 * RandomGenerator -1.0;
     rsq = v1* v1 + v2*v2;
   } while (rsq >=1.0 || rsq ==0.0);
   fac = sqrt(-2.0 *log(rsq)/rsq);
   /* Now make the Box-Muller transformation to get two normal deviates. Return    one and save the other for next time. */
   gset=v1*fac;
   iset=1; /* set flag*/
   return v2*fac;
 }
 else
 { 
   iset=0; /* we have an extra deviate handy , so unset flag and return it */
   return gset;
 }
}
/***********************************************************************/
/* Purpose : generate random multivariate normal data with covariance R */
/*              and mean 0                                              */
/* input :   R          N,N covariance matrix                           */
/*           N          dimension of R                                  */
/*           number     scalar, number of multiv. obs to be generated   */
/* output :  data       number,N matrix of data                         */
/* Method : R= T*T',  q=N(0,1) indep.                                   */
/*      x= T*q . Hence x~ MN(0,R)                                       */
/* NOTE   : converges  slowly to R, need huge sample (say 5000 or more) */
void normal_data(int number, real **R, int N, real **data)
{
    int i,k;
    real *q=NULL;

    alloc(N,&q);
    choleski(R,N);
    //printmatr(N,N,R,"T");

    for (k=0; k<number; k++)
    {
      for (i=0; i<N;i++)
        q[i]= gasdev();
      //printvecr(N,q,"q");
      MatMultVecReal(N,R,N,q,data[k]); // R is really T here
    }

    //printmatr(number,N,data,"Data");
    frei(N,&q);
}
/***********************************************************************/
/*
void main()
{
 int i;
 float r;

 srand48( getpid() ); 
 cputime(1);
 for(i=0; i<10; i++) 
 {
   r= gasdev();
   printf(" %f \n ", r );  
 }
 cputime(2);
}
*/
