/***************************************************************************/
// Copyright (c) 1996 by Matthias Schonlau
/***************************************************************************/

#ifndef general 
#define general 

#define   TRUE 		1
#define   FALSE		0
#define   EPS		3.0e-7
#define   REAL_EPS	3.0e-14    /* set this according to double or float*/
#define   REAL_INFINITY 1.0e14
#define   INFINITY 	1.0e7
#define   PI		3.14159265359
#define   space 	"  "
#define   OK 		0   // for permlex
#define   All_Done 	1  // for permlex
#define MAXLINE 150   /* need for readmat.cc , and krigmin.cc */
#define MAXTERMS 30 /* in readmat.cc : maximal number of terms in header line */
#define STRING_LEN 30


typedef double real;
typedef int boolean;
typedef char string[STRING_LEN];
/* note : if I read something longer than [30], this will result into a 
          segmentation error */ 

struct Matrix
{
  int  rows; /* care: NumRow */
  int  cols; /* care:  NumCol */
  boolean init;     /* initialized : TRUE : elements were  assigned */  
  real **mat; /* elements */
  string  name;
  boolean labelled; /* TRUE : colname were assigned */
  string  *colname;
} ; 

struct Vector
{
  int  dim; /* care:  NumCol */
  boolean init;
  real *vec;
  boolean labelled;
  string  name;
  string  *dimname;
};

//#define MAX(a,b) ((b) > (a) ? (b) : (a))    	/* max(a,b)*/
//#define MIN(a,b) ((b) < (a) ? (b) : (a))    	/* min(a,b)*/
#define SIGN(a,b) ((b) > 0.0 ? fabs(a) : -fabs(a))    /* a* sign(b)*/
static real sqrarg;
#define SQR(a) ((sqrarg=(a)) == 0.0 ? 0.0 : sqrarg*sqrarg)   /*  a*a */

double drand48( void );
void srand48( long );

/*********************************/
/*linear algebra                 */
/*********************************/
void inverse(real **a, real **y, int N);
real deter(real **a, int N);
void choldc(real **a, int n, real p[]);
void choleski(real **mat, int n);
void eigen(real **mat, int n, real *d);
/********************************/
/*mathematical operations	*/
/********************************/
int choose(int n,int i);
int factorial(int n);
/*********************************/
/*optimization                   */
/*********************************/
int Minimize(real (*ObjFunc)(real *x, int nDims),
          real AbsTol, real RelTol, unsigned MaxFuncs,
          int nDims, real **startx, real *starty, string method);
unsigned Simplex(real (*ObjFunc)(real *x, int nDims),
          real AbsTol, real RelTol, unsigned MaxFuncs,
          int nDims, real **p, real *y);
int Powell(real (*func)(real [],int nDims), real ftol,
       int ITMAX, int n, real p[], real *fret, real **xi);
/*********************************/
/*statistical distributions      */
/*********************************/
float gammp(float a, float x);
real  pdfgauss(real z);
real cdfinvgauss(real p);
real cdfgauss(real z, boolean Lower);
float betai(float a, float b, float x);

/*********************************/
/* Random Generators             */
/*********************************/
int RandInit(int _xcomp, int _ycomp, int _zcomp);
real RandUnif(void);
float gasdev();
real betadev(int a, int b);
float gamdev(int ia);
void create_cor_mat(int p,int N, real **R);
void normal_data(int number, real **R, int N, real **data);

/********************************/
/* Objective Functions 	 	*/
/********************************/
double BraninFunc(double *x,int n);
double GPFunc(double *x,int n);
double Hartman6Func(double *x,int n);
double SixHumpFunc(double *x,int n);

/********************************/
/* Transformations 		*/
/********************************/
void trafo_ab_inf(int n,int d,real **x,real **range,real **trafox,string method);
void trafo_inf_ab(int n,int d,real **x,real **range,real **trafox, string method);

/********************************/
/* Error, Warning, Info cover functions*/
/********************************/
void nrerror(char error_text[]);
void nrerror(char *error_text,real value,char *value_text);
void nrinfo(char *s);
void nrwarning(char *s);
/********************************/
/* Miscellaneous		*/
/********************************/
void exec_system(char command[MAXLINE]);
void  cputime(int ruf, char test[]="");
real maximum_two(real a, real b);
void min_distance_each_dim(int n,int d,real *pt,real **all_pts,real **range,int *direction,real *sm_dist);
int str_to_real( char* str , real *val);
int na_str_to_real( char str[], real *val,real na_value= -REAL_INFINITY);
#endif
