#include <stdio.h>
#include <string.h>
#include <stdlib.h> /* also used  for atoi */
#include <math.h>

#include "general.h"  
#include "alloc.h"  
#include "matrix.h"  
#include "krig.h"

extern FILE *ofp;

/************************************************************************/
/* Version : Nov 17, 1997                                               */
/* Purpose : handle tokens for prediction                               */
int token_predict(char **token,int n_token, char infile[] , char outfile[])
{
  if (n_token==1)
  {
     strcpy(infile,"xpred.mat");
     strcpy(outfile,"ypred.mat");
     nrinfo("Default input filename is used: xpred.mat");
     nrinfo("Default output filename is used: ypred.mat");
  }
  else if (n_token==3)
  {
    if (strcmp(token[1],">")==0)
    {
      strcpy(infile,"xpred.mat");
      strcpy(outfile,token[2]);
      nrinfo("Default input filename is used: xpred.mat");
    }
    else if (strcmp(token[1],"<")==0)
    {
      strcpy(infile,token[2]);
      strcpy(outfile,"ypred.mat");
      nrinfo("Default output filename is used: ypred.mat");
    }
    else return 1;
  }
  else if (n_token==5)
  {
    if (strcmp(token[1],">")==0 && strcmp(token[3],"<")==0)
    {
      strcpy(infile,token[4]);
      strcpy(outfile,token[2]);
    }
    else if (strcmp(token[1],"<")==0 && strcmp(token[3],">")==0)
    {
      strcpy(infile,token[2]);
      strcpy(outfile,token[4]);
    }
    else return 1;
  }
  else  return 1;

  if (strcmp(infile,"screen")==0 || strcmp(infile,"s")==0)
  {
     fprintf(ofp,"Sorry, cannot input from the screen\n");
     return 1;
  }

  return 0;
}
/************************************************************************/
/* Version : January 30, 1997 - Nov 25,97				*/
/* Purpose : cover function for prediction 				*/
/* NOte    : don't worry about screen at this point			*/	
/* NOTE    : only doing this for very first KrigModel			*/
#define USAGE_PREDICT "Usage: Predict {<input_filename} {>output_filename}\n"
#define NO_RESPONSE "Sorry, no response variables.\n"
void init_predict(KrigModel *kr,KrigArray *ka,char **token, int n_token)
{
  char infile[MAXLINE], outfile[MAXLINE];
  
  if (token_predict(token,n_token, infile, outfile)!=0)
    {printf(USAGE_PREDICT); return;}
  if (ka->p+ka->c == 0) 
    { printf(NO_RESPONSE); return;}
  database(kr, ka);
  krigpredict(kr,ka,infile,outfile);
}
#undef USAGE_PREDICT
#undef NO_RESPONSE
/************************************************************************/
/* Version : January 30, 1997                                           */
/* Purpose : cover function for designtomin				*/
#define USAGE_DESMIN "Usage: MinimizeStage number_points  >output_filename\n"
void init_MinStage(KrigModel *kr,KrigArray *ka,char **token,int n_token)
{
  int number, error ;

  if (n_token==4 && strcmp(token[2],">")==0)
  {
     if ((number=atoi(token[1]))<=0) 
     {
      fprintf(ofp,"Sorry, command ignored. Integer value needed for MinimizeStage.\n"); return;
     }
     error = MinimizeStage(kr,ka,number,token[3]);
  } 
  else printf(USAGE_DESMIN);
}
#undef USAGE_DESMIN
/************************************************************************/
/* Version : Nov 30, 1997                                           	*/
/* Purpose : cover function for Minimze 				*/
void init_Minimize(KrigModel *kr,KrigArray *ka,char **token,int n_token,
   boolean RandomError)
{
    int error;

    if (n_token==1)
    {
      error = minimize(kr,ka);
      if (error==1)
        nrwarning("Couldn't find Expected Improvement > 0.0");
      else if (error==2)
        nrwarning("Local Optimizer failed to make 3 successful searches");          }
    else
      fprintf(ofp,"Unexpected Input. No tokens should follow 'Minimize'\n");

}
#undef INCOMPATIBLE_RANDOM
/************************************************************************/
/* Version : Feb 25, 1997        	                                */
/* Purpose : cover function for MainEffects				*/
#define USAGE_MAINEFFECTS "Usage: MainEffects > output_filename\n"
void init_MainEffects(KrigModel *kr,KrigArray *ka,char **token,int n_token)
{
  int i;

  if (n_token==3 && strcmp(token[1],">")==0)
  {
     if (ka==NULL)
       MainEffects(kr,token[2]); 
     else
     {
       MainEffects(&kr[0],token[2],'w');
       for (i=1;i< ka->p+ka->c; i++)
         MainEffects(&kr[i],token[2],'a');
     }
  }
  else printf(USAGE_MAINEFFECTS);
}
#undef USAGE_MAINEFFECTS
/************************************************************************/
/* Version : Feb 25, 1997	                                        */
/* Purpose : cover function for JointEffects				*/
#define USAGE_JOINTEFFECTS "Usage: JointEffects > output_filename\n"
void init_JointEffects(KrigModel *kr,KrigArray *ka,char **token,int n_token)
{
  int i;

  if (n_token==3 && strcmp(token[1],">")==0)
  {
     if (ka==NULL) 
       JointEffects(kr,token[2]);
     else 
     {
       JointEffects(&kr[0],token[2],'w');
       for (i=1;i< ka->p+ka->c; i++)
         JointEffects(&kr[i],token[2],'a');
     }
  }
  else printf(USAGE_JOINTEFFECTS);
}
#undef USAGE_JOINTEFFECTS
/************************************************************************/
/* Version : Feb 12, 1997                                           	*/
/* Purpose : cover function for fit 					*/
void init_fit(KrigModel *kr,KrigArray *ka,char **token,int n_token)
{
  int i;
   if (n_token==1)
   {
      if (initialized(kr->init,"X","y",NULL,NULL,NULL))
      {
         database(kr, ka);
         if (ka->p+ka->c == 0) fprintf(ofp,"Sorry, no response variables.\n"); 
         for (i=0;i< ka->p+ka->c;i++)
         {
            fprintf(ofp,"Fitting %s ... \n",kr[i].ylabel);
            krig_mle(&kr[i],TRUE);
         } 
      }
    }
    else 
      fprintf(ofp,"Unexpected Input. No tokens should follow 'Fit' (other than comments).\n");
}
/************************************************************************/
/* Version : Nov 5, 1997                                           	*/
/* Purpose : read xcv.mat 						*/
/* on input:  xfile   name of x's to be read  				*/
/*            yfile   name of y 's to be read				*/
/* on output: anyn  (1)vector, contains length of anyX			*/	
/*            anyX  (anyn,d) matrix read				*/
/*            anyy  (anyn,1) matrix read				*/
/* NOTE : cannot call more than once because read_matrix(0,0, 		*/
/*        would cause problems						*/
/* NOTE : if exit with error anyX and anyY are freed . Otherwise not    */
#define COL_MISMATCH "ERROR: Incorrect number of columns\n"
#define ROW_MISMATCH "ERROR: Incorrect number of rows\n"
int read_anyXY(char *xfile,char *yfile,int *ptr_anyn, int d, real ***ptr_anyX,
   real **ptr_anyyvec)
{
  int i, actual_n,actual_d;
  string *label=NULL;
  real **anyy=NULL;//anyy is a ptr to a vector, 
                   //which causes problems since we read matrices 

  //reading xfile
  if (read_matrix(0,0,ptr_anyX,&actual_n,&actual_d,xfile,&label)!=0) {  
     if (actual_n*actual_d!=0) frei(actual_n,actual_d,ptr_anyX);
     if (actual_d!=0)       frei(actual_d,&label);return 1;
  }
  if (actual_d != d) {
     fprintf(ofp,COL_MISMATCH); 
     frei(actual_n,actual_d,ptr_anyX); frei(actual_d,&label);return 1;
  }
  //not checking whether labels are identical 
  *ptr_anyn = actual_n;
  frei(d,&label);

  // reading yfile
  if (read_matrix(0,0,&anyy,&actual_n,&actual_d,yfile,&label)!=0) {  
     if (actual_n*actual_d!=0) frei(actual_n,actual_d,&anyy);
     if (actual_d!=0)       frei(actual_d,&label);
     frei(*ptr_anyn,d,ptr_anyX); return 1;
  }
  if (actual_n != *ptr_anyn) { 
     fprintf(ofp,ROW_MISMATCH);
     frei(*ptr_anyn,d,ptr_anyX); 
     frei(actual_n,actual_d,&anyy); frei(actual_d,&label);return 1;
  }
  if (actual_d != 1) {
     fprintf(ofp,"ERROR: More than one y is not yet programmed \n");
     frei(*ptr_anyn,d,ptr_anyX); 
     frei(actual_n,actual_d,&anyy); frei(actual_d,&label);return 1;
  }
  //not checking whether labels are identical   
  frei(actual_d,&label);

  //copy anyy to anyyvec
  alloc(*ptr_anyn,ptr_anyyvec);
  for (i=0;i<(*ptr_anyn);i++)
    ptr_anyyvec[0][i] = anyy[i][0];
  frei(*ptr_anyn,1,&anyy);

  // anyyvec and anyX need to be freed later
  return 0;
}
#undef COL_MISMATCH 
#undef ROW_MISMATCH 
/************************************************************************/
/* Version : Nov 5, 1997                                                */
/* Purpose : call read_anyXY with the right parameters from token       */
/* on input  : i    =0 if first time , so */ 
/* on return : outputfile set 						*/
#define TOO_MANY_ROWS "ERROR: Cross Validation files contain more rows than model.\n" 
#define ATTEMPT_OVERWRITE "ERROR: Attempting to overwrite input file\n" 
int get_anyXY(int n_token,char **token,char **outputfile,
   int *ptr_anyn,KrigModel *kr, real ***ptr_anyX,real **ptr_anyy)
{
     int n,d;

        n= *kr[0].n;
        d= *kr[0].d;

        if (n_token==3) 
        { // normal case; crossvalidation with X and y  
         alloc(n,d,ptr_anyX);
         alloc(n,ptr_anyy);
         copy_real_matrix(n,d,kr[0].X,*ptr_anyX);
         copy_real_vector(n,kr[0].y,*ptr_anyy);
         *ptr_anyn = n;
         strcpy(*outputfile,token[2]);
        }
        else if (n_token==6)
        { // read xcv.mat (first time)
          if (strcmp(token[1],"<")==0) { 
            if (read_anyXY(token[2],token[3],ptr_anyn,d,ptr_anyX,ptr_anyy)!=0) 
              return 1;
            strcpy(*outputfile,token[5]);
            if (strcmp(*outputfile,token[2])==0 || 
                strcmp(*outputfile,token[3])==0) {
                fprintf(ofp,ATTEMPT_OVERWRITE);  frei(*ptr_anyn,ptr_anyy);
                frei(*ptr_anyn,d,ptr_anyX);
                return 1;
            }
          }
          else if (strcmp(token[3],"<")==0) { 
            if (read_anyXY(token[4],token[5],ptr_anyn,d,ptr_anyX,ptr_anyy)!=0) 
              return 1; 
            strcpy(*outputfile,token[2]);
            if (strcmp(*outputfile,token[4])==0 || 
                strcmp(*outputfile,token[5])==0) {
                fprintf(ofp,ATTEMPT_OVERWRITE);  frei(*ptr_anyn,ptr_anyy);
                frei(*ptr_anyn,d,ptr_anyX);
                return 1;
            }
          }
          else nrerror("CodeBug : get_anyXY token problem inner loop");
          if (*ptr_anyn > *kr[0].n) {
            fprintf(ofp,TOO_MANY_ROWS);
            frei(*ptr_anyn,ptr_anyy);
            frei(*ptr_anyn,d,ptr_anyX);
            return 1;
          }
        }
        else nrerror("get_anyXY: unrecognized option for token");
     return 0;
}
#undef TOO_MANY_ROWS
#undef ATTEMPT_OVERWRITE
/************************************************************************/
/* Version : May 06, 1997                                           	*/
/* Purpose : run all krigmodels for Cross  Validation			*/
/* NOTE : same X/anyX matrix used for all Krigmodels !!!	        */
/* NOTE : assumption that d is the same for all p KrigModels   		*/  
#define USAGE_KRIGCV "Usage: CrossValidate {< x_filename y_filename} > filename \n"
void init_krigcv(KrigModel *kr,KrigArray *ka,char **token,int n_token)
{
  real **cvmat=NULL, **oneypred=NULL;
  string *label=NULL;
  char *outputfile=NULL;
  int col_cvmat=3; //number of columns for one cv (Pred,se,cvei)
                   // needs to be changed in unison with eicv in krigblup.cc
  int i, pc,n, anyn,d ;
  real **anyX=NULL, *anyy=NULL;

   if (
        ((n_token==3 && strcmp(token[1],">")==0)  
         || n_token==6 && strcmp(token[1],"<")==0 && strcmp(token[4],">")==0
         || n_token==6 && strcmp(token[1],">")==0 && strcmp(token[3],"<")==0)
        && strcmp(token[0],"CrossValidate")==0
      )
   {
    if (initialized(kr->init,"X","y","alpha","theta","beta")
    && initialized(kr->init,"var",NULL,NULL,NULL,NULL)  )
    {
     pc = ka->p+ka->c; n = *kr[0].n; d= *kr[0].d;
     alloc(MAXLINE,&outputfile);
     alloc(n,pc*col_cvmat,&cvmat); 
     alloc(n,col_cvmat,&oneypred); 
     alloc(pc*col_cvmat,&label);
     database(kr, ka);
     if (get_anyXY(n_token,token,&outputfile,&anyn,kr,&anyX,&anyy)!=0) 
     { 
         frei(n,pc*col_cvmat,&cvmat); frei(n,col_cvmat,&oneypred);
         frei(MAXLINE,&outputfile); frei(pc*col_cvmat,&label); return;
     }
     for (i=0;i< pc;i++)
     {
        fprintf(ofp,"CrossValidating %s ... \n",kr[i].ylabel);
        krigcv_anyXY(&kr[i],anyX,anyy,anyn,*kr[i].d,oneypred,col_cvmat);
        //copy oneypred into  appropriate cols of cvmat
        copy_matrix(n,col_cvmat,oneypred,0,cvmat,i*col_cvmat);
     } 

     strcpy(label[0],"CvPred");strcpy(label[1],"CvSe");strcpy(label[2],"CvEi");
     label_suffix(kr,pc,col_cvmat,label);
     outputmatr_cover(anyn,pc*col_cvmat,cvmat,outputfile,label, 
        "CrossValidationMatrix");

     frei(n,pc*col_cvmat,&cvmat); 
     frei(n,col_cvmat,&oneypred);
     frei(pc*col_cvmat,&label); 
     frei(MAXLINE,&outputfile);
     frei(anyn,&anyy);
     frei(anyn,d,&anyX);

    }
   }
   else fprintf(ofp,USAGE_KRIGCV); 
}
#undef USAGE_KRIGCV
