#include <stdio.h>
#include <math.h>
#include <string.h>

#include "general.h"
#include "alloc.h"
#include "matrix.h"
#include "krig.h"

#define DEBUG 0

extern FILE *ofp;
extern int GridSize;
extern boolean PtsOutOfRange;

#define OUT_OF_RANGE_WARN "WARNING: Effects are based on the range as given in Xdescription.\n    Not all X points are within that range.\n" 
/************************************************************************/
/* Version : Feb 25, 1997                                               */
/* Purpose : compute MainEffects and output into outfile                */
/* Note    : GridSize determines #oflevels for each main effect		*/
void MainEffects(KrigModel *kr,char *outfile, char mode)
{
  int i,j,d;
  real **allmat=NULL;  // all main effects
  real **gridmat=NULL; // (grid_n,1) matrix for grid only
  real **ymat=NULL;   // (grid_n,2) matrix for predictions 
  string **effectmat=NULL; //(grid_n*d,3) matrix for the effects
  int grid_n;
 
 if (initialized(kr->init,"X","y","alpha","theta","beta")
 && initialized(kr->init,"var","range",NULL,NULL,NULL)  )
 { 
  cputime(11,"Main Effects");
  if (PtsOutOfRange) fprintf(ofp,OUT_OF_RANGE_WARN);
  ensure_init(kr,"R");
  ensure_init(kr,"F");
  kr_cover_augmR(kr); // calc . U, eigval, eigvalinv
  d = *kr->d;
  alloc(d*GridSize,d+2,&allmat); //+2 for pred. and se
  alloc(d*GridSize,2,&ymat); //for pred. and se
  alloc(d*GridSize,4,&effectmat); // for (term,value,effect,se)
  for (i=0; i<d*GridSize; i++)
    for (j=0; j<d; j++)
      allmat[i][j] = (kr->range[j][0] + kr->range[j][1])/2.0;
  //printmatr(d*GridSize,d+2,allmat); // last two cols undefined
  for (i=0; i<d; i++) 
  { // main effects for variable i
    grid(GridSize,1,&(kr->range[i]),&gridmat,&grid_n);
    copy_matrix(GridSize,1,gridmat,0,&allmat[i*GridSize],i);
    // copy gridmat into effectmat starting from i*GridSize in column 2
    copy_matrix(GridSize,1,gridmat,0,&effectmat[i*GridSize],1); 
    // fill up effectmat with xlabel
    for (j=0;j<GridSize;j++)
       strcpy(effectmat[j+i*GridSize][0],kr->xlabel[i]);
    frei(grid_n,1,&gridmat);
  }
  krigblup(kr,d*GridSize,allmat,ymat);
  copy_matrix(d*GridSize,2,ymat,0,allmat,d); // copy predictions into allmat
  // copy ymat into effectmat starting from 0 into cols 3 and 4 
  copy_matrix(d*GridSize,2,ymat,0,effectmat,2);
  if (strcmp(outfile,"screen")==0 || strcmp(outfile,"s")==0)
  {
    printvecs(d,kr->xlabel,"Labels:");
    printmatr(d*GridSize,d+2,allmat); // this is different than from outputmat
    //printmats(d*GridSize,4,effectmat);
  }
  else
  {
    fprintf(ofp,"Writing Main Effects to '%s' file...\n",outfile);
    //outputmatr(d*GridSize,d+2,allmat,"test",kr->xlabel,NULL,"MainEffects");
    outputmats(d*GridSize,4,effectmat,outfile,NULL,
       "    Term     Value       Effect       StandardError", "MainEffects", 
     mode);
  }

  frei(d*GridSize,d+2,&allmat); 
  frei(d*GridSize,2,&ymat); 
  frei(d*GridSize,4,&effectmat); 
  cputime(12,"Main Effects");
 }
}
/************************************************************************/
/* Version : Feb 25, 1997                                               */
/* Purpose : compute JointEffects and output into outfile               */
/* Note    : GridSize determines #oflevels for each joint effect	*/
void JointEffects(KrigModel *kr,char *outfile,char mode)
{
  int i,j,k,d, counter; 
  int n_effectmat; // number of elements for all effects together
  int grid_n; // number of points in one grid
  real **allmat=NULL;  // all joint effects
  real **gridmat=NULL; // (grid_n,1) matrix for grid only
  real **ymat=NULL;   // (grid_n,2) matrix for predictions
  string **effectmat=NULL; //(grid_n*d,3) matrix for the effects
  real **je_range=NULL;// range for joint effect
  char line[MAXLINE];

 if (initialized(kr->init,"X","y","alpha","theta","beta")
 && initialized(kr->init,"var","range",NULL,NULL,NULL)  )
 {
  cputime(11,"Joint Effects");
  if (PtsOutOfRange) fprintf(ofp,OUT_OF_RANGE_WARN);
  ensure_init(kr,"R");
  ensure_init(kr,"F");
  kr_cover_augmR(kr); // calc . U, eigval, eigvalinv
  d = *kr->d; 
  n_effectmat = (d-1)*d/2 *GridSize*GridSize;
  alloc(n_effectmat,d+2,&allmat); //+2 for pred. and se
  alloc(n_effectmat,2,&ymat); //for pred. and se
  alloc(n_effectmat,5,&effectmat); //for (term,value1,value2,effect,se) 
  alloc(2,2,&je_range); // range for joint effect
  for (i=0; i<n_effectmat; i++)
    for (j=0; j<d; j++)
      allmat[i][j] = (kr->range[j][0] + kr->range[j][1])/2.0;
  for (i=0,counter=0; i<d; i++)
  {
    je_range[0][0]= kr->range[i][0]; je_range[0][1]= kr->range[i][1]; 
    for (k=i+1; k<d; k++)
    { // joint effects for variables i k
      je_range[1][0]= kr->range[k][0]; je_range[1][1]= kr->range[k][1]; 
      grid(GridSize,2,je_range,&gridmat,&grid_n); 
      //copy col 1,2 of gridmat into col i, k  of effectmat row(counter*grid_n)
      copy_matrix(grid_n,1,gridmat,0,&allmat[counter*grid_n],i); 
      copy_matrix(grid_n,1,gridmat,1,&allmat[counter*grid_n],k); 
      // copy col 1,2 of gridmat into col 2,3 of effectmat row (counter*grid_n) 
      copy_matrix(grid_n,1,gridmat,0,&effectmat[counter*grid_n],1);
      copy_matrix(grid_n,1,gridmat,1,&effectmat[counter*grid_n],2);
      // fill up effectmat with xlabel
      for (j=0;j<grid_n;j++)
      {
         sprintf(effectmat[j+counter*grid_n][0],"%s  %s",kr->xlabel[i],
           kr->xlabel[k]); // note we are putting to labels in one spot !!
      }
      frei(grid_n,2,&gridmat);
      counter++;
    }
  }
  krigblup(kr,n_effectmat,allmat,ymat);
  copy_matrix(n_effectmat,2,ymat,0,allmat,d); // copy predictions into allmat
  // copy ymat into effectmat starting from 0 into cols 4,5
  copy_matrix(n_effectmat,2,ymat,0,effectmat,3);
  if (strcmp(outfile,"screen")==0 || strcmp(outfile,"s")==0)
  {
    printvecs(d,kr->xlabel,"Labels:");
    printmatr(n_effectmat,d+2,allmat); // this is differnt than outputmatr
    //printmats(n_effectmat,d+2,allmat);
  }
  else
  {
    fprintf(ofp,"Writing Joint Effects to '%s' file...\n",outfile);
    outputmats(n_effectmat,5,effectmat,outfile,NULL,
       "    Term1 Term2  Value1       Value2     Effect  StandardError", 
       "JointEffects" , mode);
  }

  frei(n_effectmat,d+2,&allmat); //+2 for pred. and se
  frei(n_effectmat,2,&ymat); //for pred. and se
  frei(n_effectmat,5,&effectmat); //for (term,value1,value2,effect,se) 
  frei(2,2,&je_range); // range for joint effect
  cputime(12,"Joint Effects");
 }

}

/**************************************************************************/
#undef OUT_OF_RANGE_WARN
