#include <stdio.h>
#include <math.h>
#include "general.h"
#include "alloc.h"
#include "matrix.h"
#include "krig.h"

//#define RandomGenerator drand48()
#define RandomGenerator RandUnif()
	// DEBUG: get start values and loglikelihood at each function evaluation
	// only for the first two dimensions
#define DEBUG 0
extern int MleTries;
extern int RandomError;
extern FILE *ofp;

/* declare vecs and matrices file-global to avoid repeated allocation 
  for each function evaluation for the local optimizer*/ 
static KrigModel krfit;  

/* for the decomposition of R (not of augmR) */
static real 	**Ufit=NULL, *diaginvfit=NULL, *eigvalfit=NULL;  

/* for calc_var */
static real *yminFbeta=NULL, *UyminFbeta=NULL, *tempn=NULL;

/* for calc_beta */
static real *yU=NULL, *tempr=NULL;
static real **transFU=NULL, **FU=NULL, **transF=NULL, **temprr=NULL, 
            **temprrinv=NULL, **FUDinv=NULL;

static int info_infty;
static int info_infty_warn;
static boolean warnflag; //TRUE if warnflag printed in a try
static int MAX_ITER_FIT=2000;//max iter for correlation param. optimization
                // method for local optimization (simplex|powell)
static string optim_method = "simplex";


/****************************************************************/
/* Version : June 26, 1996 					*/
/* Purpose : allocate file-global vecs and matrices 		*/ 
/* Note    : this function must be called prior to calc_var*/
/*           and calc_beta 					*/
void fit_global_alloc()
{
   int n,r;

   n= *krfit.n;
   r = *krfit.r;

   /* for calc_var */
   alloc(n,&tempn); 
   alloc(n,&yminFbeta);
   alloc(n,&UyminFbeta); 

   /* for calc_beta */
   alloc(r,&tempr); 
   alloc(n,&yU);
   alloc(n,r,&transFU); 
   alloc(r,n,&transF);
   alloc(r,n,&FU);
   alloc(r,n,&FUDinv);
   alloc(r,r,&temprr); 
   alloc(r,r,&temprrinv);

   alloc(n,n,&Ufit);
   alloc(n,&eigvalfit); 
   alloc(n,&diaginvfit);
}
/****************************************************************/
/* Version : June 26, 1996                                      */
/* Purpose : free file-global vecs and matrices                 */
/****************************************************************/
void fit_global_free()
{
   int n,r;

   n= *krfit.n;
   r = *krfit.r;

  /* for calc_var */
  frei(n,&tempn);
  frei(n,&yminFbeta);
  frei(n,&UyminFbeta);

  /* for calc_beta */
  frei(r,&tempr);
  frei(n,&yU);

  /* for calc_beta */
  frei(n,r,&transFU);
  frei(r,n,&transF);
  frei(r,n,&FU);
  frei(r,n,&FUDinv);
  frei(r,r,&temprr);
  frei(r,r,&temprrinv);

  frei(n,n,&Ufit);
  frei(n,&diaginvfit); 
  frei(n,&eigvalfit);
}
/****************************************************************/
/* Version : June 26, 1996 					*/
/* Purpose : calc stochastic variance hat as given by the mle   */ 
/* var         = 1/n ( (y-mu)' invR (y-mu) )			*/
/*             = 1/n [U'(y-F beta)]' D^(-1) [U'(y-F beta)] 	*/
/*             = 1/n [(y-F beta)' U] D^(-1) [(y-F beta)' U]' 	*/
/* Note    : file_global_alloc(kr)  must be called prior to this*/ 
/* Note    : F, U must be calc. prior to this 		        */
/****************************************************************/
void  calc_var(real **U, real *diaginv)
{
   int n ,r;

   n = *krfit.n;  r = *krfit.r;
   calc_beta(U,diaginv); /* calculate beta */

   MatMultVecReal(n,krfit.F,r,krfit.beta,tempn); /* tempn is Fbeta here */
   VecAddVecNewReal(-1.0,krfit.y,n,tempn,yminFbeta);
   VecMultMatReal(yminFbeta,n,U,n,UyminFbeta);
   VecEwMultVecNewReal(UyminFbeta,n,diaginv,tempn); /* tempn overwritten */
   *krfit.var = VecMultVecReal(tempn,n,UyminFbeta)  / n; 
}
/****************************************************************/
/* Version : June 26, 1996 					*/
/* compute beta hat as given by the mle				*/
/*   beta  = (F' Rinv F)^(-1) (F' Rinv y)			*/	
/*         = (F' U Dinv U' F)^(-1) (F' U Dinv U' y)		*/
/* 	 = ( F' U Dinv (F' U)' )^(-1)  ( F' U Dinv (y' U)' )	*/
/* Note    : file_global_alloc(kr)  must be called prior to this*/
/* Note    : F,U need to calc. prior to this 		        */
/****************************************************************/
void calc_beta(real **U,real *diaginv)
{
  int n,r;

  n = *krfit.n; r = *krfit.r;

  TransRealMat(n,r,krfit.F,transF);
  TransRealMat(n,r,krfit.F,transF);
  MatMultMatReal(r,transF,n,U,n,FU);
  MatMultDiagMatReal(r,FU,n,diaginv,FUDinv); 
  TransRealMat(r,n,FU,transFU);
  MatMultMatReal(r,FUDinv,n,transFU,r,temprr);
  inverse(temprr,temprrinv,r); /* temprr is overwritten and useless*/ 
  VecMultMatReal(krfit.y,n,U,n,yU);
  MatMultVecReal(r,FUDinv,n,yU,tempr);
  MatMultVecReal(r,temprrinv,r,tempr,krfit.beta);
}
/************************************************************************/
/* Version : May 22, 1996  						*/
/* Purpose : Perform an eigenvalue decomposition of kr.R   		*/
/* on output : U,  diaginv  						*/
/* Note    : R need to be calc. prior to this				*/
/************************************************************************/
void decompose(real **U, real *eigval, real *eigvalinv)
{
  int i ,n ;

  n= *krfit.n; 
  copy_real_matrix(n,n,krfit.R,U);
  eigen(U,n,eigval);
  for (i=0; i<n; i++)
  { // construct eigvalinv (check ev==0)
    if (eigval[i]> REAL_EPS || eigval[i] < - REAL_EPS)
      eigvalinv[i] = 1.0/eigval[i];
    else
    { //Correlation Matrix is singular, set ev=0
      eigvalinv[i] = 0.0;
      if (!warnflag)
      {
        nrwarning("Correlation Matrix is singular. Proceed with fingers crossed. (T2)"); 
        warnflag = TRUE;
      }
    }
  }
}
/************************************************************************/
/* Version : June 14, 1996 - Jan 30, 1997 				*/
/* Purpose : computes log likelihood of the multivariate normal 	*/
/*           kriging model  up to a constant factor 			*/  
/* Side Effects :  also updates						*/
/*                 beta, var, R, F , U , diagvec , diagvecinv 		*/
/* Input : parvec: Vector of transformed (!) parameters alpha's		*/ 
/*                 and theta's						*/
/*         n_params : dimension of parvec				*/
/* Output: fn returns loglikelihood					*/ 
/* Note : stochratio has to be one when RandomError=FALSE		*/ 
real krigloglike(real *parvec, int n_params)
{
  void decompose(real **U, real *eigval, real *diaginv);
  void invtrafo_corpar(KrigModel *kr, real *vec, int oned);
  int n ;
  real detR, s;
  
  n = *krfit.n;

  /*retransform to alpha's and theta's*/
  invtrafo_corpar(&krfit, parvec, *krfit.d); 
  kr_calcR(&krfit);
  decompose(Ufit, eigvalfit, diaginvfit); /* Decompose kr.R */
  //printvecr( *krfit.d,krfit.alpha,"loglike alpha :");
  //printvecr( *krfit.d,krfit.theta,"loglike theta:");

  calc_var(Ufit,diaginvfit); /* calculate total variance  */
  
  //note: if eigval set to zero then detR will be 0 and then give an error
  detR = ProdVecReal(n, eigvalfit);
  if (detR > 0.0)
  {
     if (*krfit.var<=0.0) 
     {  
        nrwarning("Stochastic Variance estimated to be 0 or smaller");
        *krfit.var = 1.0e-10;
     } 
     s =  - .5 * ( n * log(*krfit.var) + log(detR) );
  }
  else if (detR == 0.0) 
  {
     s = - .5 * ( n * log(*krfit.var) +REAL_INFINITY);
     /*log(detR) would yield -INFINITY, multiplied by -.5 this would encourage
       rather than discourage (detR == 0.0) */ 
     info_infty++;
     //printvecr(n,eigvalfit,"Eigenvalues ");
  }
  else if (-1.0e-4<detR && detR<0.0) 
  {
    s = - .5 * ( n * log(*krfit.var) +REAL_INFINITY);
    info_infty_warn++;
  }
  else if (detR <  0.0) 
    nrerror("loglike: detR<0.0 , R is negative definite",detR,"Determinant R");
  else nrerror("Fitting: log likelihood is complex?");

  if (DEBUG)
  { 
    fprintf(ofp,"loglike:  %f \n", s);
    print_corpars(&krfit);
  }
  return -s ; /* negative log likelihood */
}
/************************************************************************/
/* Version : May 17, 1996 - Apr 16, 1997 				*/
/* Purpose : retransform vec to alpha and theta				*/
/*           trafo for alpha: sin(x) +1.0				*/
/*           trafo for theta: x*x /10					*/
/*           trafo for stochratio: sin(x)/2.0 +.5			*/
/* on output : alpha, theta, stochratio (if RandomError) calculated 	*/
/* NOTE : if this is changed , change corresp. function in trafo.c	*/
void invtrafo_corpar(KrigModel *kr, real *vec, int oned)
{
  int i;

  for (i=0; i<oned; i++)
  {
    (*kr).alpha[i] = sin(vec[i]) + 1.0;
    (*kr).theta[i] = vec[oned+i] * vec[oned+i] / 10.0;
    if (RandomError)
      *kr->stochratio = sin(vec[2*oned])/2.0 + .5;
  }
}
/************************************************************************/
/* Version : Jan 24 , 1997	 	                                */
/* Purpose : transform from alpha and theta to infinity (in vec)        */
/*           trafo for alpha: arcsin(alpha-1.0)                         */
/*           trafo for theta: sqrt(10*theta)                            */
/*           trafo for stochratio: arcsin((2.0*(ratio-.5))		*/
/* on input  : alpha vector, theta vector, size of the vector 		*/
/* on output : vec  calculated			                        */
/* NOTE : must match trafo from invtrafo_corpar				*/
void  trafo_corpar(real *alpha, real *theta, real var, real stochratio, 
  real *vec, int oned)
{
  int i;
  real temp;

  for (i=0; i<oned; i++)
  {
      temp = alpha[i] -1.0;
      if (temp>= -1.0 && temp<=1.0)
        vec[i] = asin(temp);
      else
        nrerror("krigfit, trafo_corpar: arcsin argument out of range",temp,
        "arcsin argument: ");
      temp= theta[i];
      if (temp>=0.0)
        vec[oned+i] = sqrt(10.0* theta[i]); 
      else nrerror("trafo_corpar: theta<0.0",temp,"theta[i]: ");
  }
  if (RandomError)
  {
    if (var>0.0)
      vec[2*oned] = asin(2.0*(stochratio-.5));
    else nrerror("trafo_corpar: var",var," ");
  } 
}
/************************************************************************/
/* Version : November 14, 1996                                       	*/
/* Purpose : update R,F,U, beta, sigma					*/ 
void update()
{
   kr_calcR(&krfit);
   kr_calcF(&krfit); // only depends on x and # of elems in beta
   kr_cover_augmR(&krfit);//kr_calc_augmR(&krfit,NoCV);after calcF and kr_calcR
   decompose(Ufit, eigvalfit, diaginvfit); // before calc_var
   calc_var(Ufit,diaginvfit); /* calculate stochastic variance  */

   krfit.init->beta = TRUE; 
   krfit.init->var= TRUE;
   krfit.init->U = TRUE;  krfit.init->F = TRUE;
   krfit.init->R = TRUE;
}
/************************************************************************/
/* Version : Jan 24, 1997                                       	*/
/* Purpose : Find the (unconstrained) start values for mle optimization	*/  
/* NOTE : if try==1 and old mle's exist, use them as start values	*/
/* on input :  	x  dim_y+1,dim_y matrix for the  starting values	*/
/*	 	y  dim_y+1 vector for function values thereof		*/ 
/*		k  MLE-Try numbers, starting with 1 			*/  
/* NOTE: can't use  trafo_ab_inf because vectors are transposed		*/
/* NOTE: for theta random start values  cover very small range		*/ 
/* 	because otherwise simplex fails to start. Will also restricts   */
/*	start thetas to be small.					*/ 
void start_values_krfit(int try_num, int dim_y,real **x,real *y)
{
   void trafo_corpar(real *alpha,real *theta,real var,real stochratio,
      real *vec,int oned);
   int i,j, d;

   if (try_num>1 || 
      !initialized(krfit.init,"theta","alpha",NULL,NULL,NULL,FALSE))
   { // random start values
     krfit.init->alpha =TRUE;//needed for first kr_calcR call in krigloglike 
     krfit.init->theta =TRUE;//start corpar values have been assigned 
     for (i=0 ; i<(dim_y+1); i++)
     {
       for (j=0; j< *krfit.d; j++)
          x[i][j] = PI * (RandomGenerator - .5 );
       // PI is chosen to cover the complete range for the sinus 
       for (j= *krfit.d; j<dim_y; j++)
          x[i][j] = PI * (RandomGenerator - .5);
       // when large theta values are chosen simplex often doesn't start up
       // form experience 'PI' works well, even though it should be INFINITY
       if (RandomError) 
           x[i][dim_y-1] = PI* RandomGenerator /2.0;  
           // .5 <= ratio <= 1.0  only cover half of range of sinus
       y[i] = krigloglike(x[i],dim_y); 
     }
   }
   else if (try_num==1)
   { //if old mle's exist, use them as starting values
     //this only works with try==1,later theta and alpha are always initialized
     if (krfit.init->theta && krfit.init->alpha)
     { // use old mle's
       fprintf(ofp,"Supplied Parameter values used as Starting values for mle.\n"); 
       d = *krfit.d;
       //transform first row of x to infinity
       trafo_corpar(krfit.alpha,krfit.theta,*krfit.var,*krfit.stochratio,
          x[0],d); 
       y[0] = krigloglike(x[0],dim_y);//get function value for old mle
       // do the rest of the rows random
       for (i=1 ; i<(dim_y+1); i++)
       {
         for (j=0; j< *krfit.d; j++)
            x[i][j] = PI * (RandomGenerator - .5 );
         // PI is chosen to cover the complete range for the sinus
         for (j= *krfit.d; j<dim_y; j++)
            x[i][j] = PI * (RandomGenerator - .5);
         if (RandomError)
           x[i][dim_y-1] = PI* RandomGenerator /2.0;  
           // .5 <= ratio <= 1.0  only cover half of range of sinus
         y[i] = krigloglike(x[i],dim_y);
       }
     }
   }
   else nrerror("start_values_krfit: Code Bug. try_num<1 or 'initalized' bug");
   if (DEBUG) printmatr(dim_y+1,dim_y,x,"xstart");
   if (DEBUG) printvecr(dim_y+1,y,"starty:"); 

}
/************************************************************************/
/* Purpose : calculate the total number of parameters to be  optimized 	*/
inline int calc_n_params(int d)
{
   int n_params ; 

   if (RandomError) 
   { 
     n_params = d * 2+1;
   }
   else 
   { // not RandomError
     n_params = d * 2;
     if (*krfit.stochratio != 1.0) 
     { nrwarning("Setting ErrorVariance to 0.0 for Fitting");} 
     *krfit.stochratio = 1.0; // in either case make sure its 0
   }
   return n_params;
}
/************************************************************************/
/* Purpose : print out details about mle fit 				*/
void mle_details_print(int k,real y0,int iter)
{
    fprintf(ofp,"MLE-Try %2d, MaxLogLike: %.2f, Evals:%4d",k+1,-y0,iter);
    if (RandomError)
      fprintf(ofp,", StochVar/TotalVar:%7.2f",krfit.stochratio[0]);
    fprintf(ofp,"\n");
    printvecr(*krfit.d,krfit.alpha,"alpha:");
    printvecr(*krfit.d,krfit.theta,"theta:");
}
/************************************************************************/
/* Purpose : save current corpar's into tempalpha and temptheta		*/
/*           save stochratio into temp_stochratio 			*/
/*	     save corresponding logklike into krigloglike (!!!)		*/
void  mle_save_corpar(real y0,real *temp_stochratio,real *tempalpha,
      real *temptheta)
{ 
       *krfit.loglike = -y0;
       copy_real_vector(*krfit.d,krfit.alpha,tempalpha);
       copy_real_vector(*krfit.d,krfit.theta,temptheta);
       *temp_stochratio = *krfit.stochratio; 
}
/************************************************************************/
/* Purpose : retrieve corpar's from tempalpha and temptheta		*/
/*	     retrieve stochratio from temp_stochratio			*/
void  mle_retrieve_corpar(real y0,real *temp_stochratio,real *tempalpha,
      real *temptheta)
{ 
      copy_real_vector(*krfit.d,tempalpha,krfit.alpha);
      copy_real_vector(*krfit.d,temptheta,krfit.theta);
      *krfit.stochratio = *temp_stochratio;
}
/************************************************************************/
/* Purpose : print out warning if number of iterations is 0 or too large*/
void  mle_minim_warning(int iteradd)
{
   if (iteradd==0) 
       nrwarning("One MLE try failed to start"); 
   else if (iteradd>MAX_ITER_FIT)
       nrwarning("MLE try was interrupted when max iteration was reached");
}
/************************************************************************/
/* Purpose : display warnings for info_infty and info_infty_warn        */
inline void  mle_warning_numerical_difficulty()
{
  if (info_infty>0) 
     fprintf(ofp,"INFO: Some numerical difficulties in computing likelihood. Frequency:%d.(T4)\n",info_infty);
  if (info_infty_warn>0) 
     fprintf(ofp,"WARNING: Some numerical difficulties in computing likelihood. Frequency:%d.(T5)\n",info_infty_warn);
}
/************************************************************************/
/* Purpose : print mle and mle-estimates of parameters			*/
void  mle_print_result()
{
   //cout<<"LogLikelihood:"<<*krfit.loglike<<" StochVariance:"
   //    <<krfit.stochratio[0]*krfit.var[0];
   fprintf(ofp,"LogLikelihood:%f StochVariance:%f",
        *krfit.loglike,krfit.stochratio[0]*krfit.var[0]);
   if (RandomError)
        fprintf(ofp,", ErrorVar:%7.2f",(1.0-krfit.stochratio[0])*krfit.var[0]);
   fprintf(ofp,"\n");
   printvecr(*krfit.d,krfit.alpha,"mle alpha:");
   printvecr(*krfit.d,krfit.theta,"mle theta:");
}
/************************************************************************/
/* Version : July 10, 1996                                       	*/
/* Purpose : Find the mle for the parameters alpha's and theta's 	*/  
/* on input : details = TRUE : print details about this fit 		*/
/*	              = FALSE: do the fit , but don't give much output  */
/* On output : kr.alpha and kr.theta are the mle's,              	*/ 
/*             kr.beta , kr.var, kr.F  are updated	          	*/ 
/* Note    : kr.X and kr.y must exist before				*/
/* Note    : makes the kriging model file-global                 	*/
/* Note    : uses random numbers 					*/
void krig_mle(KrigModel *kr,boolean details)
{
   void fit_global_alloc();
   void fit_global_free();
   void start_values_krfit(int k, int n_params,real **x,real *y);
   int iter, iteradd, n,d,  i,k, n_params;
   real **x=NULL ;  /* n_params+1, n_params matrix of parameters */  
   real  *y=NULL;   /* n_params+1 vector of objfunc evaluations at parameters */
   real *tempalpha=NULL;/*needed to keep best params from previous tries*/
   real *temptheta=NULL;/*needed to keep best params from previous tries*/
   real temp_stochratio;   //needed to keep best params from previous tries

   cputime(11,"Fitting");
   krfit = *kr; /* copy kr to file-global krfit, needed in krigloglike */
   fit_global_alloc();
   kr_calcF(&krfit); /* only depends on x (and number of elems in beta) */
   info_infty=info_infty_warn=0;

   n = *krfit.n; d = *krfit.d; 
   iter=0;
   *krfit.loglike= -INFINITY; /* nec. in case Fit is called more than once */

   for (i=0; i<n;i++)
     transF[0][i] = -1.0; /* probably not needed */

   n_params = calc_n_params(d);
   alloc(n_params+1,n_params,&x); alloc(n_params+1,&y);
   alloc(d,&tempalpha); alloc(d,&temptheta);

   for (k=0; k< MleTries; k++)
   {
     iter =0;  //number of iteration within a try 
     warnflag=FALSE; // ill-conditioning warnings printed already?

     start_values_krfit(k+1,n_params,x,y);
     iter += d+1;

     iteradd = Minimize(krigloglike,10^(-16),0.00001,MAX_ITER_FIT,
           n_params,x,y,optim_method);
     iter += iteradd;
     mle_minim_warning(iteradd);

     invtrafo_corpar(&krfit, x[0], *krfit.d);

     // if this trial was better than previous ones (or -INFINITY as default)
     if (-y[0] > *krfit.loglike)  
       mle_save_corpar(y[0],&temp_stochratio,tempalpha,temptheta);

     if (details) mle_details_print(k,y[0],iter);
   }
   mle_warning_numerical_difficulty();

   // if last trial was not the best one 
   if (-y[0] != *krfit.loglike) 
     mle_retrieve_corpar(y[0],&temp_stochratio,tempalpha,temptheta);

   // update R,F,U, beta, sigma because the best try might not be the last one
   update(); 

   mle_print_result();

   frei(n_params+1,&y);
   frei(d,&tempalpha); 
   frei(d,&temptheta);
   frei(n_params+1,n_params,&x); 
   fit_global_free();

   *kr = krfit;
   cputime(12,"Fitting");
}
/******************************************************************/
/* test loglike
   alloc(n_params,&vec);
   for (i=0 ; i< *kr.d; i++)
   {
     vec[i] = kr.alpha[i];
     vec[*kr.d+ i]= kr.theta[i];
   }
   s= krigloglike(vec, n_params); 
    fprintf(ofp,"loglike : %f ", s);
*/
