/***************************************************************************/
// Copyright (c) 1996-97 by Matthias Schonlau
/***************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h> /* also used  for atoi */
#include <math.h>
#include <unistd.h>

#include "general.h"  
#include "alloc.h"  
#include "matrix.h"  
#include "krig.h"

extern real StopTolAbs, StopTolRel,StopCritPerc;
extern int MinimizationMleFrequency, GlobalEiParam, GridSize,MleTries;
extern real MinimizationMlePerc;
extern int MaxObsForMle;
extern string StopCrit;
extern char FuncName[]; // note  *FuncName is different from FuncName[]
extern real MinTarget;
extern boolean init_FuncName;
extern boolean init_MinTarget;
extern boolean RandomError;
extern boolean SystemCommandFunction;
extern string CriterionOptimMethod;
extern string CriterionOptimStart;
extern int CriterionOptimTries;
extern FILE *ofp;

/************************************************************************/
/* Version : Feb 03, 1997 						*/
/* Purpose : cover function for calcR					*/
/* need from kr : X, alpha, theta, d, n, R				*/
/* Note : this is needed because in calcR can't do any init checks 	*/
void kr_calcR(KrigModel *kr)
{
   if (!initialized(kr->init,"alpha","theta",NULL,NULL,NULL))
      nrerror("calcR: alpha or theta not initialized"); 
   calcR(kr->X,kr->alpha,kr->theta,*kr->stochratio,*kr->n,*kr->d,kr->R);
   (kr->init)->R = TRUE;
}
/************************************************************************/
/* Version : July 09, 1996                                              */
/* on input : X, alpha, theta, d, n	  				*/
/* on output : R calculated 						*/
/* Note: when first called then StochasticProcessVariance ==0		*/
void calcR(real **X,real *alpha,real *theta,real stochratio,
   int n,int d,real **R)
{
  int i,j,k;
  real temp;

  if (stochratio > 1.0 || stochratio <0.0)
     nrerror("calcR: stochratio not within [0,1]",stochratio,"stochratio: ");

  for (i=0;i< n;i++)
    for (j=i+1;j<n;j++)
    { // note i!=j always 
        R[i][j]=1.0;
        for (k=0;k<d;k++)
        {
          temp = (fabs(X[i][k]-X[j][k]));
          R[i][j] *= exp(- theta[k] * pow(temp,(2.0-alpha[k])));
        }
        R[i][j] *= stochratio;
        R[j][i]= R[i][j];
    }
  for (i=0; i<n;i++)
    R[i][i] = 1.0;
}
/************************************************************************/
/* Version : July 10, 1996 - Jan 30,97                                  */
/* input: newx  new x coordinates                                 	*/
/*	X,alpha,theta,n,d 	corresponding entries from krigmodel	*/
/* output: r                                                 		*/
/* Note : if stochratio < 1.0 then r<1 , even for newx = x 		*/
void calcr(real *newx, real *r,real **X,real *alpha,real *theta, 
   real stochratio, int n,int d)
{
  int i,j ;
  real temp;

  if (stochratio > 1.0 || stochratio <0.0)
     nrerror("calcr: stochratio not within [0,1]",stochratio,"stochratio: ");

  for (i=0;i< n;i++)
  {
    r[i]=1.0;
    //if (!identical_vec(d,newx,X[i])) 
    if (stochratio == 1.0 || !identical_vec(d,newx,X[i]))
    { // not identical , if identical 1.0 was fine
      for (j=0; j< d;j++)
      {
        temp = (fabs(X[i][j]-newx[j]));
        r[i] *= exp(- theta[j] * pow(temp, (2.0-alpha[j]) ));
      }
    }
    r[i] *= stochratio; 
  }
}
/************************************************************************/
/* Version : July 24, 1996                                              */
/* Purpose : calculates R(x_1,x_2), i.e. the correlation		*/
/*           between two points						*/
/* on input: vec1, vec2 : the two points				*/
/* on return: R(x_1,x_2)						*/
/* need from kr : alpha, theta                                      	*/
/* NOTE: this is needed covariance in krigblup				*/ 
/* uses global variable RandomError					*/
real calcRij(KrigModel *kr, real *vec1, real *vec2)
{
  int d;
  real temp, R;

  if (*kr->stochratio > 1.0 || *kr->stochratio <0.0)
     nrerror("calcRij: stochratio not within [0,1]",*kr->stochratio,
     "stochratio: ");

  //if no error don't want to go in identical_vec every time 
  if (*kr->stochratio!=1.0 && identical_vec(*kr->d,vec1,vec2))
    return 1.0; // correlation is 1, 

  R=1.0;
  for (d=0; d< *kr->d;d++)
  {
        temp = (fabs(vec1[d]-vec2[d]));
        R *= exp(- kr->theta[d] * pow(temp, (2.0-kr->alpha[d]) ));
  }
  return (R*kr->stochratio[0]);
}
/************************************************************************/
/* Version : July 12, 1996 						*/
/* creates the F matrix  						*/
/* needs to be improved when regression model comes			*/
void kr_calcF(KrigModel *kr)
{
  int i,j;
 
  if (*kr->r !=1) nrerror("krCalcF : r!=1, that's  not implemented yet ");
  for (i=0;i< *kr->n;i++)
    for (j=0;j< *kr->r;j++)
      kr->F[i][j] = 1.0;
  kr->init->F = TRUE;
}
/************************************************************************/
/* Version : Feb 05, 1997                                              */
/* creates the F matrix                                                 */
/* on input : n, r  as dimensions of F					*/
/* on output : F							*/
/* needs to be improved when regression model comes                     */
void calcF(int n,int r,real **F)
{
  int i,j;

  if (r !=1) nrerror("calcF : r!=1, that's  not implemented yet ");
  for (i=0;i<n;i++)
    for (j=0;j<r;j++)
       F[i][j] = 1.0;
}
/************************************************************************/
/* Version : June 10, 1996                                              */
/* Purpose : construct augmented matrix and store its EV decomposition  */
/*           in kr.U, kr.eigval, and kr.eigvalinv                       */
/*   	     suitable for both  cross validation and regular  		*/
/* on input: cvindex : -1   no cross validation                         */
/*                      i   cross validation at position i              */
/* Note    : F and R need to be calc prior to this                      */
void kr_calc_augmR(KrigModel *kr, int cvindex)
{
  int i,j, n , wn,r ;
  int adjusti , adjustj;

  n = *kr->n; wn = n;
  if (cvindex != -1) wn--;
  r = *kr->r;

  /* construct augmented R in kr->U through copying from R, F, and adding 0's */
  for (i=0; i<r; i++)
    for (j=0; j<r; j++)
      kr->U[i][j]=0.0;
  for (i=0, adjusti=0; i<n; i++)
    for (j=0, adjustj=0; j<n; j++)
    {
      if (i == cvindex) adjusti= -1;
      if (j == cvindex) adjustj= -1;
      if (i != cvindex && j != cvindex)
        kr->U[i+r+adjusti][j+r+adjustj]= kr->R[i][j];
    }
  for (i=0,adjusti=0; i<n; i++)
    for (j=0; j<(r); j++)
    {
      if (i == cvindex)
        adjusti = -1;
      else
        kr->U[i+r+adjusti][j]= kr->F[i][j];
    }
  for (i=0; i<r; i++)
    for (j=0, adjustj=0; j<n; j++)
    {
      if (j == cvindex)
        adjustj = -1;
      else
        kr->U[i][j+r+adjustj]= kr->F[j][i];
    }

  /* if (*kr->n <8) printmatr(wn+r, wn+r,kr->U,"augmR"); */

  /* EV decompostion of augmented R, store in kr->U and kr->eigval */
  /* kr->U on entry contains augmented R and is overwritten */
  eigen(kr->U,wn+r,kr->eigval);

  /* construct kr->eigvalinv */
  for (i=0; i<(r+wn); i++)
  {
    if (kr->eigval[i] > REAL_EPS || kr->eigval[i] < - REAL_EPS  )
      kr->eigvalinv[i] = 1.0/kr->eigval[i];
    else
    {
      kr->eigvalinv[i] = 0.0;
      nrwarning("Correlation Matrix is singular. Eigenvalue set to zero. (T3)");
    }
  }

  kr->init->U = TRUE;
}
/************************************************************************/
/* mainly intended for testing purposes */
/* used in krigpredict*/
/* possibly, if we can integrated the cv part, may be used permanently*/
void kr_cover_augmR(KrigModel *kr)
{
  calc_augmR(*kr->n,*kr->r,kr->R,kr->F,kr->U,kr->eigval,kr->eigvalinv);
  kr->init->U = TRUE;
}
/************************************************************************/
/* Version : Feb 04, 1997                                               */
/* Purpose : construct augmented matrix and store its EV decomposition  */
/*           in U, ev (eigenvalues) and evinv (inverse of eigenvalues)  */
/*   	     suitable for both  cross validation and regular  		*/
/* on input: R  n,n 	correlation matrix				*/
/* 	     F  n,r 	regression matrix				*/
/* on output:U  n+r,n+r decomposition matrix				*/
/*	     ev n+r     eigenvector vector				*/
/* 	     evinv n+r  inverse of eigenvectors vector			*/
void calc_augmR(int n,int r,real **R,real **F,real **U,real *ev,real *evinv)
{
  int i,j;

  if (r>1) nrerror("calc_augmR: r>1 not implmented yet ");
  /* construct augmented R in U through copying from R, F, and adding 0's */
  for (i=0; i<r; i++)
    for (j=0; j<r; j++)
      U[i][j]=0.0; 	   // upper left corner
  for (i=0; i<n; i++)
    for (j=0; j<n; j++)
      U[i+r][j+r]= R[i][j];// lower right corner
  for (i=0; i<n; i++)
    for (j=0; j<r; j++)
      U[i+r][j]= F[i][j];  // lower left corner 
  for (i=0; i<r; i++)
    for (j=0; j<n; j++)
      U[i][j+r]= F[j][i];  // upper right corner 

  //if (n <8) printmatr(n+r, n+r,U,"augmR"); 

  /* EV decomposition of augmented R, store in U and ev */
  /* kr->U on entry contains augmented R and is overwritten */
  eigen(U,n+r,ev);

  /* construct eigvalinv */
  for (i=0; i<(r+n); i++)
  {
    if (ev[i] > REAL_EPS || ev[i] < - REAL_EPS  )
      evinv[i] = 1.0/ev[i];
    else
    {
      evinv[i] = 0.0;
      nrwarning("Correlation Matrix is singular. Proceed with fingers crossed. (T1)");
    }
  }
}
/************************************************************************/
/* Version : January 15, 1997                                           */
/* Purpose : outputs all information from database                      */
/* input   n_token 	number of tokens 				*/
void database(KrigModel *kr,KrigArray *ka)
{
   fprintf(ofp,"---------------------------------------------\n");
   fprintf(ofp,"Database :\n");
   if (kr->init->X)
   {
     fprintf(ofp,"Number of Observations :                %3d\n",*kr->n);
     fprintf(ofp,"Number of x Variables  :                %3d\n",*kr->d);
   }
   if (ka->init)
     fprintf(ofp,"Number of y Variables  :                %3d\n",ka->p);
   if (ka->init)
     fprintf(ofp,"Number of Constrained y Variables  :    %3d\n",ka->c);
   if (*kr->r > 1)
     fprintf(ofp,"Number of Regression Terms:             %3d\n",*kr->r);
   if (RandomError)
     fprintf(ofp,"RandomError:                           TRUE\n");
   else 
     fprintf(ofp,"RandomError:                          FALSE\n");
   fprintf(ofp,"Tries:                                  %3d\n",MleTries);
   fprintf(ofp,"StoppingCriterion:               %10s\n",StopCrit);
   if (MinimizationMleFrequency!=1)
     fprintf(ofp,"MleFrequency:                        %3d\n",MinimizationMleFrequency);
   if (MinimizationMlePerc!=0.0)
     fprintf(ofp,"MlePercentage:                       %3.1e\n",MinimizationMlePerc);
   if (init_MinTarget)
     fprintf(ofp,"MinimizationTarget:                    %4.1e\n",MinTarget);
   fprintf(ofp,"MleMaxObs:		 	       %4d\n",MaxObsForMle);
   if (GlobalEiParam!=1)
     fprintf(ofp,"GlobalExpectedImprovement:              %3d\n",GlobalEiParam);
   if (strcmp("tolerance",StopCrit)==0)
   {
     fprintf(ofp,"StoppingAbsoluteTolerance:          %3.1e\n",StopTolAbs);
     fprintf(ofp,"StoppingRelativeTolerance:          %3.1e\n",StopTolRel);
   }
   else if (strcmp("quantile",StopCrit)==0) 
   {
     fprintf(ofp,"GridSize:                               %3d\n",GridSize);
     fprintf(ofp,"StoppingQuantile:                   %3.1e\n",StopCritPerc);
   }
   if (strcmp("distanceneighbour",CriterionOptimStart)!=0) 
     fprintf(ofp,"CriterionOptimStart:         %15s\n",CriterionOptimStart);
   if (strcmp("simplex",CriterionOptimMethod)!=0) 
     fprintf(ofp,"CriterionOptimMethod:       %15s\n",CriterionOptimMethod);
   if (CriterionOptimTries!= -1 && CriterionOptimTries!= 2*(*kr->n))
    fprintf(ofp,"CriterionOptimTries:                    %3d\n",CriterionOptimTries);
   if (init_FuncName && !SystemCommandFunction)
     fprintf(ofp,"ObjectiveFunction:      %15s\n",FuncName);
   if (init_FuncName && SystemCommandFunction)
     fprintf(ofp,"SystemCommandObjectiveFunction:\n\t%15s\n",FuncName);
   fprintf(ofp,"---------------------------------------------\n");
}
/************************************************************************/
/* Version : January 15, 1997 						*/
/* Purpose : Print Info 						*/
void info(int  n_token)
{
  /* void printf(char *text); */

  if (n_token==1)
  {
   fprintf(ofp,"List of valid commands: \n");
   fprintf(ofp,"\nIN/OUTPUT:\n"); 
   fprintf(ofp,"(X|y|RegressionModel|StochasticProcessModel|range) (<|>) ([filename]|screen)\n");
   fprintf(ofp,"(R|F|YDescription|XDescription)                    (<|>)  ([filename]|screen)\n");
   fprintf(ofp,"\nFUNCTIONS:\n");
   fprintf(ofp,"Fit\n");
   fprintf(ofp,"CrossValidate                  {< xfilename yfilename} > ([filename]|screen)\n");
   fprintf(ofp,"Predict                         < {[filename]|screen} > {[filename]|screen}\n");
   fprintf(ofp,"Minimize\n");
   fprintf(ofp,"MinimizeStage  {number}         > {[filename]|screen}\n");
   fprintf(ofp,"MainEffects                     > {[filename]|screen}\n");
   fprintf(ofp,"JointEffects                    > {[filename]|screen}\n");
   fprintf(ofp,"\nOPTIONS:\n");
   fprintf(ofp,"GridSize 			= {number} #default is 10\n"); 
   fprintf(ofp,"GlobalExpectedImprovement	= {number} #default is 1\n");
   fprintf(ofp,"MinimumSample	 		= {number} #default is 0\n");
   fprintf(ofp,"MleFrequency		 	= {number} #default is 1\n");
   fprintf(ofp,"MleMaxObs	            	= {number} #default is 200\n");
   fprintf(ofp,"MlePercentage			= {number} #absent by default\n");
   fprintf(ofp,"MinimizationTarget             	= {number} #equals min(y_1 ... y_n) by default\n");
   fprintf(ofp,"ObjectiveFunction 		= {name of executable}\n");
   fprintf(ofp,"RandomError			= {TRUE/FALSE}\n");
   fprintf(ofp,"RandomSeed 			= {number} #default is 10\n");
   fprintf(ofp,"SystemCommandObjectiveFunction  = {system command for function}\n"); 
   fprintf(ofp,"StoppingCriterion 		= {quantile/tolerance} #default is tolerance\n"); 
   fprintf(ofp,"StoppingAbsoluteTolerance 	= {number} #default is 0.001\n"); 
   fprintf(ofp,"StoppingRelativeTolerance 	= {number} #default is 0.001\n");
   fprintf(ofp,"Tries	 			= {number} #default is 3\n");
   fprintf(ofp,"CriterionOptimStart             = {random|displacement|distanceneighbour}\n");
   fprintf(ofp,"                                 #default is distanceneighbour\n"); 
   fprintf(ofp,"CriterionOptimMethod            = {simplex|powell} #default is simplex\n");
   fprintf(ofp,"CriterionOptimTries             = {number} #default is 2*n\n");
   fprintf(ofp,"\nOTHER:\n");
   fprintf(ofp,"check_allocation (as first command only)\n"); 
   fprintf(ofp,"#{comment}\n");
   fprintf(ofp,"Info\n");
   fprintf(ofp,"Database\n");
   fprintf(ofp,"Quit\n");
  }
  else fprintf(ofp,"Unexpected input. No tokens should follow 'info' (other than comments)\n");
}
/************************************************************************/
/* Version : July 10 , 1996 - Feb25 , 97			     	*/
/* Purpose : checks whether matrix is initialized and does initialize 	*/
/*           if necessary 					      	*/
void ensure_init(KrigModel *addr_kr, char name[])
{
    if (!initialized(addr_kr->init,name,NULL,NULL,NULL,NULL,FALSE))
    {
       if (strcmp("R",name)==0)
          kr_calcR(addr_kr);
       else if (strcmp("F",name)==0)
          kr_calcF(addr_kr);
       else if (strcmp("U",name)==0)
          kr_cover_augmR(addr_kr);
       else 
          nrerror("ensure_init: This option is not recognized");
    }
}
/************************************************************************/
/* Version : Feb 10, 1997 						*/
/* Purpose : determine whether one component initialized		*/
/* output : name  name to be output					*/
/* return  : TRUE if all (initialized or NULL)				*/
boolean initialized_check(InitStruct *init, char *p,string name)
{
   strcpy(name,p);

   if (strcmp("X",p)==0)
     return (init->X);
   else if (strcmp("y",p)==0)
     return (init->y);
   else if (strcmp("R",p)==0)
   {
     strcpy(name,"Correlation Matrix");
     return (init->R);
   }
   else if (strcmp("F",p)==0)
     return (init->F);
   else if (strcmp("beta",p)==0)
   {
     strcpy(name,"Regression Matrix");
     return (init->beta);
   }
   else if (strcmp("alpha",p)==0)
     return (init->alpha);
   else if (strcmp("theta",p)==0)
     return (init->theta);
   else if (strcmp("U",p)==0)
     return (init->U);
   else if (strcmp("range",p)==0)
   {
     strcpy(name,"Range for X");
     return (init->range);
   }
   else if (strcmp("var",p)==0)
   {
     strcpy(name,"Stochastic Process Variance");
     return (init->var);
   }
   else if (strcmp("FuncName",p)==0)
   { // this one different because global var
     strcpy(name,"Objective Function");
     return (init_FuncName);
   }
   else 
   {
      fprintf(ofp,"%s\n",p);
      nrerror("initialized_check code bug. Unrecognized option.");
   }
   return FALSE ; // to avoid warning while compiling
}
/************************************************************************/
/* Version : Feb 10, 1997 						*/
/* Purpose : Prints error message if not initialized 			*/
/* return  : TRUE if all (initialized or NULL)				*/
/*	     FALSE if at least one (!NULL && !initialized)		*/
boolean initialized(InitStruct *init, char *n1, char *n2, 
  char *n3, char *n4, char *n5,boolean output)
{
  boolean initialized_check(InitStruct *init, char *p, string name);
  boolean status=TRUE;
  string name;
 
  if (n1 != NULL && initialized_check(init, n1,name)==FALSE) 
  {
    status=FALSE;
    if (output)
       fprintf(ofp,"Sorry, can't do that. %s is not initialized.\n",name);
  }
  if (n2 != NULL && initialized_check(init, n2,name)==FALSE)
  {
    status=FALSE;
    if (output)
       fprintf(ofp,"Sorry, can't do that. %s is not initialized.\n",name);
  }
  if (n3 != NULL && initialized_check(init, n3,name)==FALSE)
  {
    status=FALSE;
    if (output)
       fprintf(ofp,"Sorry, can't do that. %s is not initialized.\n",name);
  }
  if (n4 != NULL && initialized_check(init, n4,name)==FALSE)
  {
    status=FALSE;
    if (output)
       fprintf(ofp,"Sorry, can't do that. %s is not initialized.\n",name);
  }
  if (n5 != NULL && initialized_check(init, n5,name)==FALSE)
  {
    status=FALSE;
    if (output)
       fprintf(ofp,"Sorry, can't do that. %s is not initialized.\n",name);
  }

  return status;
}
/************************************************************************/
/* Version : Feb 17, 1997 						*/
/* Purpose : not used right now, checks whether ka is initialized	*/
/* return  : TRUE if ka->p>0						*/
/*	     FALSE otherwise 						*/
boolean initialized(KrigArray *ka)
{
  if (ka->p>0) return TRUE;
  fprintf(ofp,"Sorry, there are no response variables.\n"); 
  return FALSE;
}
/************************************************************************/
/* Version : November 17, 1996                                          */
/* Purpose : Output several matrices in case of fatal error 		*/ 
void output_error(KrigModel *kr, char error_text[])
{
  fprintf(stderr,"Matt's runtime errors ... \n");
  fprintf(stderr,"%s \n", error_text); 
  fprintf(stderr,"Emergency output of X,y,RegressionModel,StochasticProcessModeland yDescription\n");
  menu_print(&kr,NULL,2,"X","error_x.mat");
  menu_print(&kr,NULL,2,"y","error_y.mat");
  menu_print(&kr,NULL,2,"RegressionModel","error_beta.mat");
  menu_print(&kr,NULL,2,"StochasticProcessModel","error_corpar.mat");
  menu_print(&kr,NULL,2,"yDescription","error_ydescrip.mat");
  fprintf(stderr,"Now exiting to system ... \n ");
  exit(1);
}
/************************************************************************/
/* Version : Feb 3, 1997	                                        */
/* Purpose : Output several matrices (intermediate saving of results)   */
void output_temp(KrigModel *kr,char *top_text)
{
  nrinfo("Output of matrices in temp files: \n\tX,y,RegressionModel,StochasticProcessModel,yDescription");
  menu_print(&kr,NULL,2,"X","temp_x.mat",top_text);
  menu_print(&kr,NULL,2,"y","temp_y.mat",top_text);
  menu_print(&kr,NULL,2,"RegressionModel","temp_beta.mat",top_text);
  menu_print(&kr,NULL,2,"StochasticProcessModel","temp_corpar.mat",top_text);
  menu_print(&kr,NULL,2,"yDescription","temp_ydescrip.mat",top_text);
}
/************************************************************************/
/* Version : Jan 29, 1997 						*/
/* Purpose : split a string up into tokens 				*/
/* input  :  l   	string of unknown length			*/
/*           token      (n_token,len_token) array empty			*/
/* output :  token 	(input n_token,len_token) array for the tokens 	*/	
/*	     n_token    number of tokens found (n_token is changed) 	*/
/* Note   :  tokens are any signs other than ' ' and '\t'		*/
/*           seperator for tokens are ' ','<','=','>','\t'		*/	
/* Example:  "Predict <x.mat >    y.mat"				*/  
/*           Predict,<,x.mat,>,y.mat are tokens				*/
/* Note  :   l mustn't start with a seperator , 			*/
/*		e.g. "< x.mat Predict " is illegal			*/ 
/* Note(BUG):   if only two seperators appear in sequence , two diff. 	*/
/*	     tokens will be constructed					*/
/* return : ok ok, 1 unexecpeted input,2 token_num_error		*/
#define  token_num_error "More Tokens in line than code can handle\n"
int tokenize(char *l,int *n_token,int len_token,char **token)
{ 
  int j; //counter of elems within a token
  int i; // counter of elems of the whole line
  int token_count; //counter of tokens
  int length ; // length of line
  char ch;

  length = int(strlen(l));
  i=0; j=0; token_count=0;
  while (i<length && (l[i]==' ' || l[i]=='\t')) i++;//needed if line empty
  while(i<length && l[i]!='#' && l[i]!='\n')
  {
     if (i<length && (l[i]=='<' || l[i]=='>' || l[i]=='=')) 
     { //error 
         fprintf(ofp,"Unexpected input: \n\t(1)Input Line should not start with seperator or\n\t(2) Two seperators should not appear in sequence.\n"); 
        return 1;
     }
     while (i<length && (ch=l[i])!='<' && ch!='>' && ch!='=' && ch!=' ' 
          && ch!='\t' && ch!='#' && ch!='\n')
     { // read a word (not a seperator) 
        token[token_count][j] = ch ;
        i++; j++;
        if (j>= len_token-1) 
          nrerror("tokenize: attempt to write over max length of a token");
     }
     token[token_count][j] = '\0';
     j=0; token_count++;
     if (token_count >= *n_token) {fprintf(ofp,token_num_error);return 2;} 
     while (i<length && (l[i]==' ' || l[i]=='\t')) i++; // jump over extra space
     if ( (ch=l[i])=='<' || ch=='>' || ch=='=')
     { // Is there <,>,= following spaces ? If so enter it as next token 
       token[token_count][0]= ch;
       token[token_count][1]='\0';
       token_count++; 
       i++;
       if (token_count >= *n_token) {fprintf(ofp,token_num_error);return 2;}
       while (i<length && (l[i]==' ' || l[i]=='\t')) i++;//jump over extra space
     }
  }
  *n_token = token_count; //token_count already increased to offset for 0 start
  return OK;
}
#undef  token_num_error
/************************************************************************/
/* Version : Feb 13, 1997                                               */
/*    modified Jan 20, 2000  Brian's suggestion				*/
/* Purpose : Construct y mat from different Kriging Models 		*/
/* input:  kr, ka, ymat,ylab allocated					*/
/* output : ymat  kr->n,(ka->c+ka->p)					*/
/* return : 0 ok, 1 danger of overwriting ylab array			*/
int construct_ymat(KrigModel *kr, KrigArray *ka,real **ymat,char *ylab,
   int maxylab)
{
  int i,j; 

  if (ka==NULL)
     nrerror("Code Bug construct_ymat : ka == NULL");
  for (i=0;i<(*kr->n);i++)
    for (j=0;j<(ka->c+ka->p);j++)
      ymat[i][j]= kr[j].y[i];
  strcpy(ylab,"");
  for (i=0;i<(ka->c+ka->p);i++)
  {
    if (int(strlen(ylab))+int(strlen(kr[i].ylabel))+5 > (maxylab-1))
       return 1;
    strcat(ylab,kr[i].ylabel);
    strcat(ylab,"     ");
  }
  return 0;
}
/************************************************************************/
/* Version : Nov 15, 1997                                               */
/* Purpose : Construct corpar.mat from different Kriging Models         */
/* input:  kr, ka 						 	*/
/*         corlab   1+2*(p+c) vec of strings, allocated			*/
/*         cormat   d,1+2*(p+c) matrix, allocated 			*/
/* NOTE : '1+' is for the labels of the x 's				*/
/* output : ymat  kr->n,(ka->c+ka->p)                                   */
void construct_cormat(KrigModel *kr,KrigArray *ka,string **cormat,
string *corlab)
{
  int i,j;

  if (ka==NULL)
     nrerror("Code Bug construct_cormat : ka == NULL");
  for (i=0;i<(*kr->d);i++)
  {
    sprintf(cormat[i][0],"%s",kr->xlabel[i]);
    for (j=0;j<(ka->c+ka->p);j++) 
    { 
      sprintf(cormat[i][2*j+1],"%f",kr[j].theta[i]);
      sprintf(cormat[i][2*j+2],"%f",kr[j].alpha[i]);
    } 
  }
  strcpy(corlab[0],"Variable");
  for (i=0;i<(ka->c+ka->p);i++)
  {
    strcpy(corlab[2*i+1],"Theta.");
    strcat(corlab[2*i+1],kr[i].ylabel);
    strcpy(corlab[2*i+2],"Alpha.");
    strcat(corlab[2*i+2],kr[i].ylabel);
  }
}
/************************************************************************/
/* Version : March 10, 1997                                             */
/* Purpose : test whether all elems in X are within range		*/ 
/* input : mat  (n,d) matrix						*/
/*         range (d,2) matrix with lower and upper limits on matrix elems*/
/* return : 0 all elements of mat are within limits			*/
/*          1 at least one elem is not					*/ 
int test_X_in_range(int n,int d,real **mat,real **range)
{
  int i,j;

  for (i=0;i<n;i++)
  {
    for (j=0;j<d;j++)
       if (mat[i][j]<range[j][0] || mat[i][j]>range[j][1]) 
          return 1;
  }
  return 0;
}
/************************************************************************/
/* Version : March 27, 1997                                             */
/* Purpose : do all elements of a matrix meet the constraints?		*/
/* input : ymat  (n,pc) matrix , a row representing one set of y values   */
/*         ydescrip (pc,3) matrix with simple constraints		*/
/*	 	(i.e. lower and upper limits) 				*/
/* return : 0 all rows of mat are meet constraints 			*/
/*          1 at least one row is not	                                */
/* NOTE : NA values in ymat yield error					*/
#define NA_ERROR "ERROR : y contains NA values when checked for constraints"
int check_constraints(int n,int pc,real **ymat,real **ydescrip)
{
  int i,j;

  for (i=0;i<n;i++)
  {
    for (j=0;j<pc;j++)
    {
      if (ydescrip[j][2]==TRUE)
      { // this is a constraint, go check
       if (ymat[i][j]!= -REAL_INFINITY)
       { // not a missing value
         if (ydescrip[j][0]!= -REAL_INFINITY && ymat[i][j]<ydescrip[j][0])
           return 1; // lower limit not met
         if (ydescrip[j][1]!= -REAL_INFINITY && ymat[i][j]>ydescrip[j][1])   
           return 1; // upper limit not met
       }
       else nrerror(NA_ERROR);
      }
    }
  }
  return 0;
}
#undef NA_ERROR
/************************************************************************/
/* Version : March 27, 1997                                             */
/* Purpose : put beta's into matrix and label	  			*/
/* input   : label and M not allocated 					*/
/* output  : M 	(p+c,2) Matrix						*/ 
/*           label (2)  vector						*/
/*           pc  = p+c							*/
/*           n_label =2							*/
/* return  : error condition						*/
int construct_beta(KrigArray *ka, KrigModel *kr,string ***addrM, string **addrlabel, int *pc,int *n_label)
{
  string **M=NULL, *label=NULL;
  int i;

  if (ka!= NULL && ka->init) *pc = ka->p+ka->c;
  else          *pc=1;
  if (*pc==0 || strcmp(kr->ylabel,"EMPTY")==0) return 1;
  else if (*pc<0) nrerror("construct_ydescrip: Code bug , pc <0");

  *n_label=3;
  label= *addrlabel; M = *addrM;
  alloc(*pc,*n_label,&M);
  alloc(*n_label,&label);
   
  strcpy(label[0],"Variable");
  strcpy(label[1],"Term");
  strcpy(label[2],"Coefficients");
  for (i=0;i< *pc;i++)
  {
    sprintf(M[i][0],kr[i].ylabel);
    sprintf(M[i][1],"1"); 
    if (kr[i].init->beta)
      sprintf(M[i][2],"%f",*kr[i].beta); 
    else 
      sprintf(M[i][1],"NA");
  }
   
  *addrlabel = label; *addrM = M;
  return 0;
}
/************************************************************************/
/* Version : March, 1997                                              */
/* Purpose : put summary into matrix and label  			*/
/* input   : label and M not allocated 					*/
/* output  : label      (n_label) vector, labels of matrix		*/
/*	     M	 	(p+c,n_label) matrix 				*/ 
/*	     n_label	length of label					*/ 
/* return  : error condition */
int construct_ydescrip(KrigArray *ka,KrigModel *kr,string ***addrM, string **addrlabel, int *pc,int *n_label)
{
  int i,j,max_n_label=15;
  string **M=NULL, *label=NULL;

  if (ka!=NULL && ka->init) *pc = ka->p+ka->c; 
  else          *pc=1;
  if (*pc==0 || strcmp(kr->ylabel,"EMPTY")==0) return 1;
  else if (*pc<0) nrerror("construct_ydescrip: Code bug , pc <0");

  // establish label
  label= *addrlabel; M = *addrM;
  alloc(max_n_label,&label);
  j=0;
  strcpy(label[j],"Variable"); j++;  // can't have summary without variable name
  if (*kr[0].loglike != -REAL_INFINITY)
      {strcpy(label[j],"LogLikelihood"); j++;}
  if (kr[0].init->var)
      {strcpy(label[j],"StochasticProcessVariance"); j++;}
  if (RandomError)
      {strcpy(label[j],"ErrorVariance"); j++;}
  if (kr[0].init->y)
      {strcpy(label[j],"NumberObs"); j++;}
  if (ka!= NULL && ka->ydescrip_init)
  {
      strcpy(label[j],"Min"); j++;
      strcpy(label[j],"Max"); j++;
      strcpy(label[j],"Constraint"); j++;
  } 
  *n_label = j;
  if (*n_label==0) 
     {frei(max_n_label,&label);*addrlabel = label; *addrM = M;return 2;}
  morealloc(max_n_label,&label,*n_label);
  alloc(*pc,*n_label,&M);
    
  for(i=0;i<*pc;i++) 
    for(j=0;j< *n_label;j++) 
    { 
      if (strcmp(label[j],"Variable")==0)
         strcpy(M[i][j],kr[i].ylabel);   
      else if (strcmp(label[j],"LogLikelihood")==0)
      {
         if (*kr[i].loglike != -REAL_INFINITY)
           sprintf(M[i][j],"%f",*kr[i].loglike);
         else 
           strcpy(M[i][j],"NA");
      }
      else if (strcmp(label[j],"StochasticProcessVariance")==0)
         if (*kr[i].var != -REAL_INFINITY)
           sprintf(M[i][j],"%f",kr[i].stochratio[0]*kr[i].var[0]);
         else
           strcpy(M[i][j],"NA");
      else if (strcmp(label[j],"ErrorVariance")==0)
         if (*kr[i].var!= -REAL_INFINITY)
           sprintf(M[i][j],"%f",(1.0- kr[i].stochratio[0])*kr[i].var[0]);
         else
           strcpy(M[i][j],"NA");
      else if (strcmp(label[j],"NumberObs")==0)
         sprintf(M[i][j],"%d",*kr[i].n);
      else if (strcmp(label[j],"Min")==0)
      {
         if (ka->ydescrip[i][0]!= -REAL_INFINITY)
           sprintf(M[i][j],"%f",ka->ydescrip[i][0]);
         else 
           strcpy(M[i][j],"NA");
      }
      else if (strcmp(label[j],"Max")==0)
      {
         if (ka->ydescrip[i][1]!= +REAL_INFINITY)
           sprintf(M[i][j],"%f",ka->ydescrip[i][1]);
         else
           strcpy(M[i][j],"NA");
      }
      else if (strcmp(label[j],"Constraint")==0)
      {
         if (ka->ydescrip[i][2]==0)      strcpy(M[i][j],"FALSE");
         else if (ka->ydescrip[i][2]==1) strcpy(M[i][j],"TRUE");
         else nrerror("construct_ydescrip : Code BUG, ka->ydescrip!=0 or 1"); 
      }
    }

  *addrlabel = label; *addrM = M;
  return 0;
}
/************************************************************************/
/* Version : May 06, 1997                                               */
/* Purpose : put suffixes into label					*/
/* input   : kr    Full KrigModel					*/
/*	pc	number of responses+constraints				*/
/*	col_pred number of labels for each response/constraint		*/
/*	label	(pc*size_set) vector, 					*/
/*		first size_set spots have labels without suffices	*/
/* output  : all label spots have labels with suffices '.{varname}'    	*/
void label_suffix(KrigModel *kr,int pc,int size_set,string *label)
{
  int i;
  string temp;

  //copy labels everywhere
  for (i=size_set; i<(pc*size_set); i++)
    strcpy(label[i],label[i%size_set]);
  //create suffices everywhere
  for (i=0; i<(pc*size_set); i++)
  {
    sprintf(temp,"%s.%s",label[i],kr[i/size_set].ylabel);
    //note i/size_set gives division rounded down to integer
    strcpy(label[i],temp);
  }
}
/************************************************************************/
/* Purpose : print parameters alpha, theta, and stochratio              */
/* Version : Nov 02, 1997                                               */
void print_corpars(KrigModel *kr)
{
    if (RandomError)
      fprintf(ofp,", StochVar/TotalVar:%7.2f",kr->stochratio[0]);
    fprintf(ofp,"\n");
    printvecr(*kr->d,kr->alpha,"alpha:");
    printvecr(*kr->d,kr->theta,"theta:");
}
/************************************************************************/
/* Purpose : Put vector of parameters into alpha, theta, and stochratio */
/* Version : Nov 02, 1997                                               */
/* on input : parvec: vector of size n_params that contains params:     */
/*                    alpha, theta, stochratio (if RandomError)         */
void copy_vec_corpar(KrigModel *kr, real *parvec, int  n_params)
{
  int d,i;

  d = *kr->d;
  for (i=0; i<n_params; i++)
  {
    if (i<d)
       (*kr).alpha[i] = parvec[i];
    else if (i < 2*d)
       (*kr).theta[i-d] = parvec[i];
    else if (RandomError && i == 2*d)
       *kr->stochratio = parvec[i];
    else nrerror("CodeBug in set_start_corpar in kriglib.cc");
  }
}
/************************************************************************/
/* Purpose : get all ylabels in one vector 				*/
/* Version : Nov 02, 1997                                               */
void assemble_ylab(KrigArray *ka,KrigModel *kr,string *ylab)
{
   int i;
  
   for (i=0;i< (ka->p + ka->c); i++)
   {
     strcpy(ylab[i],kr[i].ylabel);
   }
}
/************************************************************************/

/*
int main()
{
  void tokenize(char *l,int *n_token,int len_token,char **token);
  char *p="Minimize # for better ";
  char **token=NULL;
  int i,*n_token=NULL, max_n_token,len_token; 
  
  max_n_token =10; 
  len_token=30;  
  alloc(1,&n_token);
  *n_token = max_n_token;
  alloc(max_n_token,len_token,&token); 
  tokenize(p,n_token,len_token,token); 
  cout << "n_toekn " << *n_token << endl;
  for (i=0;i< *n_token;i++)
     fprintf(ofp,"*%s*\n",token[i]);
  frei(max_n_token,len_token,&token); 
  frei(1,&n_token);
  return 0;
}
*/

