// A note on randomness : Used by program at 3 places
// (1) starting values for local searches of E(I) in krigmin.cc
// (2) starting values for local searches of MLE in krigfit.cc
// (3) generate random normal numbers in gasdev() (feeding monte carlo method) 

#include <stdio.h>
#include <string.h>
#include <stdlib.h> /* also used  for atoi */
#include <math.h>
#include <unistd.h>
#include <signal.h>

#include "general.h"  
#include "alloc.h"  
#include "matrix.h"  
#include "krig.h"


extern int SPEED; /* global variable defined in alloc.cc*/

#define DEBUG_TOKEN  0

real VERSION = 1.304;
FILE *ofp=NULL;         /* global file pointer for space.out */
FILE *ifp=NULL; 	 // input file pointer
boolean interactive;  // interactive mode or input through file ? 
static boolean _FIRSTCOMMAND_=TRUE; //is this the first command (for menu) ? 

// note: OptTolAbs < StopTolAbs needed always (otherwise won't converge)
// note: it is CRUCIAL to have OptTolAbs=0.0, i.e. no absolute criterion
// note: instead of OptTolRel, we use right now StopTolRel/100.0
real OptTolAbs=0.0, OptTolRel=.001; //for optimization criterion
real StopTolAbs=.001, StopTolRel=.001;//for tolerance convergence criterion

real StopCritPerc=0.01;//Critical level for cdf of the minimum order statistic
real MinTarget;  // target for minimization
int GlobalEiParam=1;  // global parameter 
int MinimizationMleFrequency = 1; // frequency of mle update
real MinimizationMlePerc = 0.0; // mle updated once percentage data added
int GridSize=10; // scalar, number of levels for the orderalg grid 
int MinSample=0; //scalar, for Minimize want at least a sample of this size
string StopCrit; // for minimization : order or tolerance (initialized in main) 
int MleTries = 3; // for Fitting
int MaxObsForMle= 200; // Minimization: mle not done with more than MaxObsForMle
boolean RandomError=FALSE; // indicate whether random error present 
boolean PtsOutOfRange=FALSE; // some X pts out of range given by XDescription ? 
boolean DesMinOption=FALSE;//if true,change calc of z; init. in case `minimize`
boolean init_MinTarget=FALSE; //target set for minimization?  
boolean init_FuncName=FALSE;  //is FuncName initialized? 
char FuncName[MAXLINE]; //name of function to be minimized 
boolean SystemCommandFunction;// False if FuncName is an executable, 
		   	// TRUE otherwise (e.g. Splus BATCH command) 
string CriterionOptimStart= "distanceneighbour";//method finding starting values
			//(random|displacement|distanceneighbour)
string CriterionOptimMethod = "simplex";//local optimization method 
                        //(simplex|powell)
int CriterionOptimTries= -1;//number of local tries to find MaxCriterion
			// default -1 indicates to be set to 2*n later 
// on the choice of CriterionOptim parameters : 
// if (RandomError) distanceneighbour may lead to distances zero (identical pts)
// hence if (RandomError && distanceneighbour) 
//    simplex doesn't start because if distances==0 ill conditioning 
// powell : "distanceneighbour" disturbs the first point in all directions 
//	    hence we are not starting searches from x's (but between x's)

/************************************************************************/
/* Version : May 05, 1997                                               */
/* Purpose : determine System Command Objective Function                */
#define USAGE_SYSCOM "Usage: SystemCommandObjectiveFunction = system command (several words allowed) \n\t that interacts with space.temp\n"
#define ERROR_LEN "ERROR: SystemCommandObjectiveFunction is too long.\n"
void init_SystemCommandFunction(KrigModel *kr,char **token,int n_token)
{
   if (n_token>2 && strcmp(token[0],"SystemCommandObjectiveFunction")==0
      && strcmp(token[1],"=")==0)
   {
      if (concat_string(n_token-2,&(token[2]),MAXLINE,FuncName)==0)
      {
         init_FuncName = TRUE;
         SystemCommandFunction = TRUE;
         fprintf(ofp,"SystemCommandObjectiveFunction = %s\n",FuncName);
      }
      else
        fprintf(ofp,ERROR_LEN);
   }
   else fprintf(ofp,USAGE_SYSCOM);

   // strcpy(FuncName,"Splus BATCH /u1/mschonlau/acedir/executable/pistonfunc.s temp.out");
}
#undef USAGE_SYSCOM
#undef ERROR_LEN
/************************************************************************/
/* Version : July 18, 1996 						*/
/* Purpose : get data component (p), action (<,>,>screen) and 		*/
/*	file name  							*/
/* input   : n_token  number of tokens					*/
/*	     token    array of tokens					*/	
void inout(KrigModel **ptrkr,KrigArray *ka,char **token,int n_token)
{
  KrigModel *kr=NULL;
  char *filename=NULL, *field=NULL;
  extern FILE *ofp;
  int action;
  
  kr = *ptrkr;
  if (n_token==3 && (strcmp(token[1],"<")==0 || strcmp(token[1],">")==0)) 
  {
   if (strcmp(token[1],"<")==0)
      action = 3;
   else if (strcmp(token[1],">")==0)
      action = 2;
   else nrerror("inout : Code Bug (1)"); 
   if ((filename=token[2]) ==NULL )
      nrerror("inout : Code Bug (2)");
   if ((field=token[0]) ==NULL )
      nrerror("inout : Code Bug (3)");
   if (strcmp(filename,"screen")==0 || strcmp(filename,"s")==0)
   { 
     if  (action==2) 
       action=1;
     else 
     {
       fprintf(ofp,"Sorry, can't read a file onto the screen. Line ignored.\n");
       return;
     }
   }
   menu_print(ptrkr,ka,action,field,filename);
   kr = *ptrkr; // because ptrkr might have changed
  }
  else
  { // input errors
    if (strcmp(token[1],">")!=0 || strcmp(token[1],"<")!=0)
     fprintf(ofp,"Unexpected input. A '<' or a '>' sign was expected. Input line ignored.\n");
    if ( n_token!=3)
     fprintf(ofp,"Unexpected input. Exactly 3 tokens are expected. (excluding comments '#')\n\tInput line ignored.\n");
  }
}
/************************************************************************/
/* Version : July 09, 1996 - Nov 15, 97					*/
/* Print (items) in menu 						*/
/* on input : action : 1 Print , 2 Output , 3 Read 			*/
/*            filename: name of a file					*/
/* 	      p	     : data component (i.e. X or y ..)			*/
/* WARNING:  ptrkr is not nec. the beginning of the krigarray when called */
/*  from output_error for example					*/
void menu_print(KrigModel **ptrkr,KrigArray *ka,int action,char *p, char *filename,char *top_text)
{
   KrigModel *kr=NULL;
   int n ,r ,nr,d ,w, n_label,pc;
   real **ymat=NULL;  
   string **SM=NULL, *slabel=NULL; // needed for output YDescription 
   char ylab[MAXLINE];

   kr = *ptrkr;
   n = *kr->n; r= *kr->r; nr = n+r, d = *kr->d; 
   if (ka==NULL && action==3) nrerror("menu_print: Code Bug , ka==NULL");

   if (strcmp("X",p)==0) 
   {
     if (action==3)  {w=read_x(kr,filename);}
     else if (initialized(kr->init,"X",NULL,NULL,NULL,NULL))
     {
       if (action==1) 
       {
         printvecs(d,kr->xlabel,"Labels:");
         printmatr(n,d,kr->X,"X:");
       }
       else if (action==2) 
       {
         outputmatr(n,d,kr->X,filename,kr->xlabel,NULL,"X Matrix");
       }
     }
   }
   else if (strcmp("y",p)==0 || strcmp("Y",p)==0)
   { 
     if (action==3 && initialized(kr->init,"X",NULL,NULL,NULL,NULL)) 
     {  
        w=read_y(ptrkr,ka,filename);
        kr = *ptrkr; //because *ptrkr might have changed
     }
     else if (action<3 && initialized(kr->init,"y","X",NULL,NULL,NULL))
     {
      if (ka!=NULL)
      { // normal case , output all y's
       alloc(n,ka->p+ka->c,&ymat);
       if (construct_ymat(kr,ka,ymat,ylab,MAXLINE)!=0)
          nrerror("menu_print: y labels are too long .");
       if (action==1) 
       {
          fprintf(ofp,"y Label: %s\n",ylab);
          printmatr(n,ka->p+ka->c,ymat,"y:");
       }
       else if (action==2)
       {
         outputmatr(n,ka->p+ka->c,ymat,filename,NULL,ylab,"y Matrix");
       }
       frei(n,ka->p+ka->c,&ymat);
      }
      else 
      { // ka == NULL (e.g. output tempmat), just output the vector y  
       if (action==1)
       {
         fprintf(ofp,"y Label: %s\n",kr->ylabel);
         printvecr(n,kr->y,"y:");
       }
       else if (action==2)
         outputvecr(n,kr->y,filename,kr->ylabel,"y Matrix");
      }
     }
   }
   else if (strcmp("R",p)==0)
   {
     if (action == 3) fprintf(ofp,"Sorry, can't read that .. \n");
     else if (initialized(kr->init,"R",NULL,NULL,NULL,NULL))
     {
       if (action==1) printmatr(n,n,kr->R,"R:");
       else if (action==2)
          outputmatr(n,n,kr->R,filename,kr->xlabel,NULL,"Correlation Matrix",
          TRUE);
     }
   }
   else if (strcmp("F",p)==0)
   {
    if (action == 3) fprintf(ofp,"Sorry, can't read that .. \n");
    else if (initialized(kr->init,"F",NULL,NULL,NULL,NULL))
    {
     if (action==1) printmatr(n,r,kr->F,"F:");
     else if (action==2) outputmatr(n,r,kr->F,filename,NULL,"F","F Matrix");
    }
   }
   else if (strcmp("beta",p)==0 || strcmp("RegressionModel",p)==0)
   { 
     if (action == 3) { w=read_beta(ptrkr,ka,filename); }
     else 
     {
        //if ( initialized(kr->init,"beta",NULL,NULL,NULL,NULL))
        if (construct_beta(ka,kr,&SM,&slabel,&pc,&n_label)==0)
        {
         if (action ==2)
           outputmats(pc,n_label,SM,filename,slabel,NULL,"RegressionModel");
         else
         {
           printvecs(n_label,slabel,"");
           printmats(pc,n_label,SM,"");
         }
         frei(pc,n_label,&SM);
         frei(n_label,&slabel);
        }
        else fprintf(ofp,"Sorry, there are no RegressionCoefficients.\n");
     }
   }
   else if (strcmp("range",p)==0 || strcmp("XDescription",p)==0
     || strcmp("xdescription",p)==0)
   {
     if (action == 3) 
     { // need to know d , i.e. X has to have been read  
       if (initialized(kr->init,"X",NULL,NULL,NULL,NULL))  
          w=read_xdescrip(kr,ka,filename);  
     }
     else if ( initialized(kr->init,"range",NULL,NULL,NULL,NULL))
     {
       if (action==1) printmatr(d,2,kr->range,"range:"); 
       else if (action==2) 
          output_range(kr,filename,"XDescription");
     }
   }
   else if (strcmp("corpar",p)==0 || strcmp("StochasticProcessModel",p)==0)
   {
     if (action == 3) 
     {
       if (initialized(kr->init,"X","y",NULL,NULL,NULL)) 
          { w=read_corpar(kr,ka,filename); }
     }
     else if ( initialized(kr->init,"theta","alpha",NULL,NULL,NULL))
     {
       if (ka==NULL || !(ka->init))
         nrerror("init corpar : ka is not initialized or NULL ");
       
       string **cormat=NULL;
       string *corlab=NULL;
       int n_col;

       n_col = 1+2*(ka->p+ka->c);
       alloc(d,n_col,&cormat);
       alloc(n_col,&corlab);
       construct_cormat(kr, ka,cormat,corlab); 
       outputmats_cover(d,n_col,cormat,filename,corlab,"CorrelationParameters");
       frei(d,n_col,&cormat);
       frei(n_col,&corlab);
     }
   }
   else if (strcmp("summary",p)==0 || strcmp("Summary",p)==0
     || strcmp("YDescription",p)==0 || strcmp("ydescrip",p)==0
     || strcmp("yDescription",p)==0)
   {  
     if (action == 3)
     { 
        w=read_ydescrip(ptrkr,ka,filename); 
        kr = *ptrkr; //because ptrkr might have changed
     }
     else if (action ==2 || action==1)
     { 
        if (construct_ydescrip(ka,kr,&SM,&slabel,&pc,&n_label)==0)
        {
         if (action ==2)
           outputmats(pc,n_label,SM,filename,slabel,NULL,"YDescription");
         else
         {
           printvecs(n_label,slabel,"");
           printmats(pc,n_label,SM,"");
         }
         frei(pc,n_label,&SM);
         frei(n_label,&slabel);
        }
        else fprintf(ofp,"Sorry, there are no response variables.\n");
     }
   }
   else 
   {
    if (action==1) fprintf(ofp,"Print pardon what ?  %s ?\n",p);
    else if (action==2) fprintf(ofp,"Output pardon what ?  %s ?\n",p);
    else  fprintf(ofp,"Read pardon what ?  %s ?\n",p);
   }
}
/************************************************************************/
/* Version : July 13, 1996 - Nov 23, 1997				*/
/* Purpose : read commands seperated by and = sign 			*/
/*           (1)  Tries = {integer} 					*/
/*           (2)  ObjectiveFunction = {string}			 	*/
/*	     (3)  RandomSeed = {integer} 				*/
/*	     (4)  MinimumSample = {integer}				*/
/*	     (5)  GridSize = {integer}					*/
/*	     (6)  StoppingCriterion = {string}				*/
/*	     (7)  StoppingAbsoluteTolerance = {real}			*/
/*	     (8)  StoppingRelativeTolerance = {real}			*/
/*	     (9)  MleFrequency = {integer}				*/
/*	          MlePerc = {real}					*/
/*	    (10)  GlobalExpextedImprovement = {integer}			*/ 
/*	    (11)  StoppingQuantile = {real}				*/
/*	    (12)  MleMaxObs						*/
void equal_init(KrigModel *kr,char **token,int n_token)
{
  int k, k0;
  float k1; /* real doesn't work for some reason with sscanf */ 
  char *s=NULL;

  if (strcmp(token[1],"=")==0 && n_token==3)
  {
     s = token[2]; // point at the same location with s, don't allocate 
     if (strcmp("Tries",token[0])==0) 
     {
       if ((k=atoi(s))>0) 
       {
         MleTries = atoi(s);
         fprintf(ofp,"Tries = %d\n",MleTries);
       }
       else fprintf(ofp,"Sorry, command ignored. Tries>0, integer.\n");
     }   
     else if (strcmp("ObjectiveFunction",token[0])==0)
     {
       strcpy(FuncName,s);
       init_FuncName = TRUE;
       SystemCommandFunction = FALSE;
       fprintf(ofp,"ObjectiveFunction = %s\n",FuncName);
     }
     else if (strcmp("RandomSeed",token[0])==0)
     { 
       if ((k=atoi(s))>0)
       {
         k = atoi(s);
         //RandInit(10, 20, 1000);
         RandInit(k, k*2, k*100);
         fprintf(ofp,"RandomSeed = %d\n",k); 
       }
       else 
        fprintf(ofp,"Sorry, command ignored. RandomSeed>0, integer value .\n");
     }
     else if (strcmp("MinimumSample",token[0])==0)
     {
       if ((k=atoi(s))>0)
       {
          k=atoi(s);
          MinSample= k;
       }
       else fprintf(ofp,"Sorry, command ignored. MinimumSample>0, integer .\n");
     }
     else if (strcmp("GridSize",token[0])==0)
     {
       if ((k=atoi(s))>0)
       {
          k=atoi(s);
          GridSize= k;
       }
       else fprintf(ofp,"Sorry, command ignored. GridSize>0 , integer.\n");
     }
     else if (strcmp("MleFrequency",token[0])==0)
     {
       if ((k=atoi(s))>0)
       {
          k=atoi(s);
          MinimizationMleFrequency = k;
       }
       else 
       fprintf(ofp,"Sorry, command ignored. MinimizationMleFrequency>0,integer.\n");
     }
     else if (strcmp("MleMaxObs",token[0])==0)
     {
       if ((k=atoi(s))>0)
       {
          k=atoi(s);
          MaxObsForMle = k;
       }
       else 
        fprintf(ofp,"Sorry, command ignored. MaxObsForMle>0,integer .\n");
     }
     else if (strcmp("MlePercentage",token[0])==0)
     {
       if ((k1=atof(s))>0 && k1<=1.00)
          MinimizationMlePerc = k1; 
       else 
        fprintf(ofp,"Sorry, command ignored.  0.0< MinimizationMlePercentage <=1.0\n");
     }
     else if (strcmp("MinimizationTarget",token[0])==0)
     {
       if ((k=sscanf(s,"%f",&k1))!=0)
       { 
          MinTarget = k1;
          init_MinTarget = TRUE;
       }
       else
        fprintf(ofp,"Sorry, command ignored.  MinTarget must be a real number.\n");
     }
     else if (strcmp("GlobalExpectedImprovement",token[0])==0)
     { // NOTE ==0 is also error condition, this was bad        
       //if ((k=atoi(s))>=0)
       if ((k0=sscanf(s,"%d",&k))!=0 && k>=0)
          GlobalEiParam = k; 
       else 
         fprintf(ofp,"Sorry, command ignored. GlobalExpextedImprovement>=0, integer.\n");
     }
     else if (strcmp("CriterionOptimTries",token[0])==0)
     {
       if ((k0=sscanf(s,"%d",&k))!=0 && k>=0)
          CriterionOptimTries = k;
       else
         fprintf(ofp,"Sorry, command ignored. CriterionOptimTries>=0, integer.\n");
     }
     else if (strcmp("StoppingAbsoluteTolerance",token[0])==0)
     {
       if ((k1=atof(s))!=0) 
       { 
         if (k1 > OptTolAbs) 
            StopTolAbs = k1;
         else fprintf(ofp,"Sorry, can't do so. You are attempting to set OptTolAbs >= StopTolAbs\n");
       }
       else fprintf(ofp,"Sorry, command ignored. Not a valid real number.\n");
     }
     else if (strcmp("StoppingRelativeTolerance",token[0])==0)
     {
       if ((k1=atof(s))>0 && k1<1.0)
          StopTolRel = k1;
       else 
         fprintf(ofp,"Sorry, command ignored.  0.0< StoppingRelativeTolerance <1.0\n");
     }
     else if (strcmp("StoppingQuantile",token[0])==0)
     {
       fprintf(ofp,"Option StoppingQuantile is depreciated. Use \n"); 
       fprintf(ofp," default option (Tolerance) .\n");
       /*
       if ((k1=atof(s))>0 && k1<=1.00)
          StopCritPerc = k1; 
       else fprintf(ofp,"Sorry, command ignored.  0.0< StoppingQuantile <=1.0\n");
       */
     }
     else if (strcmp("StoppingCriterion",token[0])==0)
     {
       if (strcmp("quantile",s)==0 || strcmp("tolerance",s)==0)
          strcpy(StopCrit,s);
       else if (strcmp("order",s)==0)
          strcpy(StopCrit,"quantile");//maintained for upwards compatibility
       else 
          fprintf(ofp,"Sorry, only 'quantile' and 'tolerance' are valid options. Default remains.\n"); 
     }
     else if (strcmp("CriterionOptimMethod",token[0])==0)
     {
       if (strcmp("simplex",s)==0 || strcmp("Simplex",s)==0)
          strcpy(CriterionOptimMethod,"simplex");
       else if (strcmp("powell",s)==0 || strcmp("Powell",s)==0)
          strcpy(CriterionOptimMethod,"powell");
       else
          fprintf(ofp,"Sorry, only 'simplex' and 'powell' are valid options. Default remains.\n");
     }
     else if (strcmp("CriterionOptimStart",token[0])==0)
     { 
       if (strcmp("random",s)==0 || strcmp("Random",s)==0)
          strcpy(CriterionOptimStart,"random");
       else if (strcmp("displacement",s)==0 || strcmp("Displacement",s)==0)
          strcpy(CriterionOptimStart,"displacement");
       else if (strcmp("DistanceNeighbour",s)==0 
          || strcmp("distanceneighbour",s)==0)
          strcpy(CriterionOptimStart,"distanceneighbour");
       else
          fprintf(ofp,"Sorry, only 'random', 'displacement' and 'distanceneighbour'  are valid options. Default remains.\n");
     }
     else if (strcmp("RandomError",token[0])==0)
     {
       if (strcmp("TRUE",s)==0 || strcmp("true",s)==0 || strcmp("True",s)==0)
       {
          RandomError=TRUE;
       }
       else if (strcmp("false",s)==0 || strcmp("FALSE",s)==0 || 
          strcmp("False",s)==0)
       {
          if (RandomError == TRUE)
          {
           fprintf(ofp,"Error: Database indicates that RandomError == TRUE. Read ");
           fprintf(ofp,"Ydescription without \n\t RandomError if no RandomError is desired.\n");
          } 
          else RandomError=FALSE;
       }
       else
          fprintf(ofp,"Sorry, only 'FALSE' and 'TRUE' are valid options. Default remains.\n");
     }
     else 
     {
       fprintf(stderr,"%s\n",token[0]);
       nrerror("equal_init: code bug,  missing option ");
     }
  }
  else 
  { 
    if (strcmp(token[1],"=")!=0)
     fprintf(ofp,"Unexpected input. A '=' sign was expected. Input line ignored.\n");
    if ( n_token!=3)
     fprintf(ofp,"Unexpected input. Only 3 tokens were expected. (excluding comments '#')\n\tInput line ignored.\n"); 
  }
}
/************************************************************************/
/* Version : July 09, 1996 						*/
/* Purpose : read the input from stdin or input 			*/
void menu(KrigModel **ptrkr, KrigArray *ka)
{
   KrigModel *kr=NULL; 
   int error;
   extern FILE *ifp; // input file pointer
   extern boolean interactive;
   char p[130];  // a full line (when changed also change for fgets  below  
   char *s=NULL; // put token[0] in s
   char **token=NULL;
   int i,*n_token=NULL, max_n_token=10,len_token=MAXLINE;
   
   kr = *ptrkr; 
   alloc(1,&n_token);
   alloc(max_n_token,len_token,&token);
   //alloc(len_token,&s);
   fprintf(ofp,"space> ");
   /* note: gets reads extra zero's */
         //((!interactive && fgets(p,130,ifp))!=NULL)) 
   while ((interactive && gets(p)!=NULL) || 
         (!interactive && fgets(p,130,ifp))) 
   {
    *n_token = max_n_token;
    if (tokenize(p,n_token,len_token,token)!=OK) 
       *n_token=0; // indicator to read next line
    if (DEBUG_TOKEN) 
    {
      fprintf(ofp,"Number of tokens %d\n",*n_token);
      for (i=0;i< *n_token;i++)
         fprintf(ofp,"*%s*\n",token[i]);
    }
    if (strcmp(p,"         ")>0 && !interactive)
       fprintf(ofp,"%s",p);
    //strcpy(s,token[0]);
    s = token[0]; // don't allocate , only point to the same thing 
    if (*n_token>0 && s)
    {
     if (strcmp("Fit",s)==0 || strcmp("f",s)==0) 
        {init_fit(kr,ka,token,*n_token);}
     else if ((strcmp("Minimize",s)==0) || (strcmp("m",s)==0))
        { init_Minimize(kr,ka,token,*n_token,RandomError); }
     else if (strcmp("Predict",s)==0 || strcmp("p",s)==0) 
        { init_predict(kr,ka,token,*n_token);}
     else if (strcmp("MinimizeStage",s)==0) 
        { init_MinStage(kr,ka,token,*n_token);}
     else if (strcmp("MainEffects",s)==0)
       { init_MainEffects(kr,ka,token,*n_token); }
     else if (strcmp("JointEffects",s)==0)
       { init_JointEffects(kr,ka,token,*n_token); }
     else if (strcmp("CrossValidate",s)==0 || strcmp("cv",s)==0) 
       { init_krigcv(kr,ka,token,*n_token); }
     else if (strcmp("X",s)==0 
        || strcmp("y",s)==0 || strcmp("Y",s)==0 
        || strcmp("beta",s)==0 || strcmp("RegressionModel",s)==0
        || strcmp("F",s)==0 || strcmp("R",s)==0  || strcmp("range",s)==0
        || strcmp("XDescription",s)==0 || strcmp("xdescription",s)==0
        || strcmp("corpar",s)==0 || strcmp("StochasticProcessModel",s)==0 
        || strcmp("summary",s)==0 || strcmp("Summary",s)==0
        || strcmp("ydescrip",s)==0 || strcmp("YDescription",s)==0
        ) 
        { inout(ptrkr,ka,token,*n_token); kr= *ptrkr;} 
     else if (strcmp("RandomError",s)==0)
     { equal_init(kr,token,*n_token); }
     else if (strcmp("SystemCommandObjectiveFunction",s)==0)
     { init_SystemCommandFunction(kr,token,*n_token);}
     else if (strcmp("Tries",s)==0 || strcmp("ObjectiveFunction",s)==0 ||
        strcmp("RandomSeed",s)==0 || strcmp("MinimumSample",s)==0 ||
        strcmp("GridSize",s)==0 || strcmp("StoppingCriterion",s)==0 ||
        strcmp("MleFrequency",s)==0 || 
        strcmp("MlePercentage",s)==0 || 
        strcmp("MleMaxObs",s)==0 || 
        strcmp("GlobalExpectedImprovement",s)==0 || 
        strcmp("StoppingAbsoluteTolerance",s)==0 || 
        strcmp("StoppingQuantile",s)==0 || 
        strcmp("MinimizationTarget",s)==0 || 
        strcmp("CriterionOptimStart",s)==0 ||
        strcmp("CriterionOptimTries",s)==0 ||
        strcmp("CriterionOptimMethod",s)==0 ||
        strcmp("StoppingRelativeTolerance",s)==0 ) 
       { equal_init(kr,token,*n_token);} 
     else if (strcmp("database",s)==0 || strcmp("Database",s)==0)
     {
       if (*n_token==1)
         database(kr,ka);
       else 
         fprintf(ofp,"Unexpected input. No tokens should follow 'database' (other than comments)\n"); 
     }
     else if (strcmp("Info",s)==0 || strcmp("i",s)==0 || strcmp("info",s)==0
        || strcmp("help",s)==0 || strcmp("Help",s)==0 || strcmp("h",s)==0) 
     {  info(*n_token); }
     else if (strcmp("check_allocation",s)==0)
     { 
       if (!_FIRSTCOMMAND_)
         fprintf(ofp,"Sorry, check_allocation must be the first input command \n");
     } 
     else if (strcmp("Quit",s)==0 || strcmp("End",s)==0 || strcmp("quit",s)==0
       || strcmp("end",s)==0 || strcmp("q",s)==0)
       { break;}
     else
     {
        fprintf(ofp,"\nCannot parse this line. \n");
     }

     if (_FIRSTCOMMAND_) 
     { // note : if the first line is a comment line , we won't enter here
        if (strcmp("check_allocation",s)!=0) 
          SPEED =1;
        _FIRSTCOMMAND_ = FALSE; // not the first command anymore
     }
    }
    fprintf(ofp,"space> ");
   }
   frei(max_n_token,len_token,&token);
   //frei(len_token,&s);
   frei(1,&n_token);
}
/************************************************************************/
/* Version : July 26 , 1996						*/
/* Purpose : deal with argc and argv (params to main) and  		*/
void main_init(int argc, char *argv[])
{
  extern FILE *ifp; // input file pointer
  extern FILE *ofp; // input file pointer
  extern boolean interactive;

  // deal with arguments to main
  if (argc==1)
  {
    interactive = TRUE;
  }
  else if (argc==2 || argc==3)
  { // deal with inputfile only, set outputfile to stdout
    interactive = FALSE ;
    if ( (ifp =fopen(argv[1],"r")) == NULL)
    {
      fprintf(stderr,"Sorry, cannot open %s\n",argv[1]);
      exit(1);
    }
  }
  else
  {
    fprintf(ofp,"Usage:  space <inputfile  outputfile> \n");
    exit(1);
  }

  if (argc==3)
  { // deal with  outputfile only
    interactive = FALSE ;
    if ( (ofp =fopen(argv[2],"w")) == NULL)
      nrerror("Sorry, can't write to outputfile.");
  }

  if (argc==2 || argc==3) fprintf(ofp,"Reading commands from '%s'.\n",argv[1]); 
  if (argc==3) fprintf(ofp,"Writing commands to '%s'.\n",argv[2]);

  //fprintf(ofp,"Please do not input words of more than 30 characters. \n");
}
/**********************************************************************/
/* Version : Nov 11 , 97 					      */
/* Purpose avoid core dumps  which may be too big 		      */
void avoid_coredump()
{
  signal(SIGSEGV,exit);
  signal(SIGBUS,exit);
  signal(SIGFPE,exit);
  signal(SIGTERM,exit);
}
/**********************************************************************/
/*Version : Nov 6, 98 */
void output_header()
{
  extern FILE *ofp;

  fprintf(ofp,"      SPACE - Stochastic Process Analysis of Computer Experiments \n");
  fprintf(ofp,"                       (Version %5.3f)\n\n",VERSION);
}
/**********************************************************************/
int main(int argc, char *argv[])
{
 void main_init(int argc, char *argv[]);
 KrigModel *mainkr=NULL;
 KrigArray *ka=NULL; 
 extern real VERSION;
 int i;


  avoid_coredump();
  SPEED=0;
  cputime(1);
  RandInit(10, 20, 1000);
  srand48( getpid() );
  ofp = stdout;

  if ((mainkr =  (KrigModel *) malloc(1*sizeof(KrigModel)))==NULL)
       nrerror("main realloc KrigModel Failure");
  else NULL_KrigModel(&mainkr[0]);
  if ((ka =  (KrigArray *) malloc(1*sizeof(KrigArray))) ==NULL) 
       nrerror("main realloc KrigArray Failure");
  else NULL_KrigArray(&ka[0]);
  main_init(argc,argv);
  output_header();
  init_KrigModel(mainkr,21,2,1);
  init_KrigArray(ka);
  strcpy(StopCrit,"tolerance");
  menu(&mainkr,ka);

  fclose(ofp);
  if (!interactive) fclose(ifp); 
  free_Krig_cover(ka,mainkr);
  cputime(2,"Program");
  return 0;
}
