#include <stdio.h> 
#include <math.h> 
#include <stdlib.h> //needed for atof
#include <string.h>

#include "general.h"
#include "matrix.h"
#include "alloc.h"
#include "krig.h"

//#define RandomGenerator drand48()
#define RandomGenerator RandUnif()
#define DEBUG 0
#define DEBUG_EI 0

static KrigModel *krmin=NULL;/* needed to communicate with ExpImp,Constprob */
static KrigArray *kam=NULL;/* needed to communicate with ExpImp */
static real  ymin;         /* needed to communicate with ExpImp */
static boolean des_to_min; /*needed to communicate with ExpImp */
static boolean constraint; /*needed to communicate with ExpImp */
static real **anyX=NULL;   /*needed to communicate with ExpImp */
static int anyX_rows;      /*needed to communicate with ExpImp */
static int anyX_cols;      /*needed to communicate with ExpImp */
static initial_des_n;	  //communicate with MaxExpImp, needed for displacement

// tranformation method (sin|exp) 
static string trafo_method = "sin";          

static int MAX_ITER = 2000;  // maximal iterations for global ei search

// for StopCrit = "quantile"
static real *h=NULL, *yvec=NULL, **ymat=NULL, **xpred=NULL; 
static int npred;       // number of points of full grid 

extern FILE *ofp;
extern real StopTolAbs,StopTolRel;
extern real  OptTolAbs ,OptTolRel;
extern int MinimizationMleFrequency, GlobalEiParam; 
extern real  MinimizationMlePerc;
extern int GridSize;
extern string StopCrit;
extern int MinSample;
extern int MaxObsForMle;
extern char FuncName[]; // note  *FuncName is different from FuncName[]
extern boolean init_FuncName;
extern boolean SystemCommandFunction;
extern real MinTarget;
extern boolean init_MinTarget;
extern int CriterionOptimTries;
extern string CriterionOptimStart;
extern string CriterionOptimMethod;
extern boolean RandomError;
extern boolean PtsOutOfRange; // set in readmat.cc

/************************************************************************/
/* Version : May 05, 1997 						*/
/* Purpose : wait until space.temp contains enough values		*/
/*  (presumably batch command is running) 				*/
/* NOTE : sleep 10 command only works in unix !				*/
void wait_spacetemp(int n_entries)
{
  FILE *waitfp=NULL;
  int i,j;
  int MAX_TRIES=1000; // max number of 10-sec interval tries to read space.temp
  boolean ready;
  char word[MAXLINE];

  j=0;
  do
  { 
    ready = TRUE; 

    if ((waitfp =fopen("space.temp","r")) == NULL)
      nrerror("wait_spacetemp: Can't open  space.temp for reading. ");
    for (i=0; i<(n_entries); i++)
    {
      if (fscanf(waitfp, "%s",word)!=1)
         ready= FALSE; // space.temp does not yet contain all entries
      //else cout << j << " " << word << endl;  
    }
    fclose(waitfp);

    if (j>=MAX_TRIES) 
      nrerror("wait_spacetemp: space.temp does not have enough values.\n May be the system command takes longer than anticipated. ",real(10*MAX_TRIES)," (real time secs)"); 
    if (!ready) 
    {
      exec_system("sleep 10"); // this command only works in unix!
      j++;
    }
  } while (!ready);
  fprintf(ofp,"Execution of System Command : %d cycles of 10 real time seconds\n",j);
  return;
}
/************************************************************************/
/* Version : July 13, 1996 - May 05 , 97				*/
/* Purpose : Sample a new Function value at xbest			*/
/* On input : FuncName  : name of executable which can calc. the function*/
/*            xbest     : x coordinates for point to be evaluated       */ 
/*            d         : dimension of xbest 				*/
/* on return : y value							*/
/* Note :  (1) open space.temp file					*/
/*         (2) write d and xbest into space.temp and close file		*/
/*         (3) pass  command to operating system that executes FuncName */
/*         (4) read  d,xbest , and the y-value from space.temp		*/
/*         (5) remove space.file (not done yet)				*/
#define SAMPLE_ERROR "sample_new_y: Possible reasons:\n\t(1) Minimization Function File does not exist in directory supplied. \n\t(2) Not enough values in  space.temp for reading.\n\t(3) Number of x-vars in Minimization Function and Program Database \n\t  don't match."
#define NON_NUMERIC "Sample New Points: Not all new points can be interpreted as a number"
void sample_new_y(char FuncName[MAXLINE],int d,real *xbest, int ny,real *newy) 
{
  FILE *fp;
  int i; 
  char line[MAXLINE];
   
  if ( (fp =fopen("space.temp","w")) == NULL)
     nrerror("sample_new_y: Can't open/create space.temp for writing. ");
  
  fprintf(fp,"%d ",d);
  for (i=0; i<d; i++)
    fprintf(fp," %f ",xbest[i]);
  fclose(fp);

  if (!SystemCommandFunction)
  { // FuncName is an ordinary executable, append space.temp
    strcpy(line,FuncName);
    strcat(line," space.temp");
    exec_system(line);
  }
  else 
  {
    exec_system(FuncName); /* ask operating system to execute FuncName  */
    wait_spacetemp(1+d+ny);
  }

  if ( (fp =fopen("space.temp","r")) == NULL)
     nrerror("sample_new_y: Can't open  space.temp for reading. ");

  for (i=0; i<(1+d+ny) ; i++)
  {
    if (fscanf(fp, "%s",line)!=1)
       nrerror(SAMPLE_ERROR);
    else  
      {  /* for some reason I can't read it directly into y */
         if (i>d)
         { // the responses/ constraints
           if (na_str_to_real(line,&(newy[i-d-1]))!=0)
             { fclose(fp); nrerror(NON_NUMERIC);} 
         }
      }
  }
  fclose(fp);
}
#undef SAMPLE_ERROR
#undef NON_NUMERIC
/************************************************************************/
/* Version : October 01, 1996                                           */
/* Purpose : Remove elements with 0 variance from ymat and xpred 	*/
/*           by exchanging them to the end of list 			*/
/* Effect:   ymat and xpred are reordered				*/
/*	     elements with non-zero cov. are between 0..(adj_npred-1),	*/
/*	     elements with zero cov are between adj_npred..(npred-1)	*/   
/* also  :   puts first col of ymat into yvec, creates h		*/
/* on return: number of total grid elements with non-zero elements	*/
/* Note :   for convenience, mean,h are not reallocated			*/
/* 	    this uses ymin , never MinTarget  				*/
/* uses npred,ymat,yvec,ymin,h,d,xpred */ 
int remove_zero_cov()
{
   int i, adj_npred;

   adj_npred = npred; 
   i=0; 
   while (i<adj_npred)              
   {
     if (ymat[i][1]>0.0)
     { // variance >0
       yvec[i]= ymat[i][0];              //copy pred value 
       h[i] = ymin;  		 	 //compute h, possibly subtract delta
       i++;
     } 
     else
     { // variance=0, discard this element: 
       // exchange useless rows to end of list : xpred, ymat 
       exchange(krmin[kam->pvec[0]].d[0],xpred,i,adj_npred-1);
       exchange(2,ymat,i,adj_npred-1);
       adj_npred--;
     }
   }
   if (npred-adj_npred!=0)
    fprintf(ofp,"%d variables in entire grid have 0 standard error.\n)",
       npred-adj_npred);
   return adj_npred;
}
/************************************************************************/
/* Version : July 30, 1996 						*/
/* Purpose : Assess whether Algorithm has converged 			*/
/* Output  : TRUE if converged , FALSE if not converged 		*/
/* Criteria: (1) deltay < StopTolAbs     OR				*/
/*           (2) deltay/y < StopTolRel					*/
boolean ConvergeCrit1(real y,real deltay,real TolAbs,real TolRel)
{
   if (deltay < TolAbs) return TRUE;
   else if (y!=0.0 && (fabs(deltay/y)) < TolRel) return TRUE;
   else return  FALSE;
}
/************************************************************************/
/* Version : July 30, 1996                                              */
/* Purpose : Assess whether Algorithm has converged			*/
/*	     Criteria : (1st order statistic) or (tolerance) 		*/
/* Output  : TRUE if converged , FALSE if not converged                 */
/* on input : npred : number of predicted values 			*/
/*	     xpred : x location for  predicted values 			*/
/* Note    : cov : n,n covariance matrix over a xpred locations 	*/ 
/*   	     h : n vector with all entries = ymin 			*/ 
/*	     yvec : predicted values over xpred 			*/
/*           ymat : matrix of  yvec and standard errors			*/ 
/* NOTE: if order of variables changes this will result into a unnoticed*/
/* serious error. Orderalg will assume probabilities for the wrong 	*/
/* variables.  The adj_npred procedure is only changing the error for	*/
/* new variables, which is ok. 						*/
boolean ConvergeCrit(real y,real deltay)
{
 int remove_zero_cov();
 int adj_npred;  // npred - (# of elems with zero  variance)
 boolean status=FALSE;
 real work_deltay;
    
 if (strcmp(StopCrit,"tolerance")==0)
 {
   if (GlobalEiParam==1 || GlobalEiParam==0) 
   {
     if (deltay < StopTolAbs) return TRUE;
     else if (y!=0.0 && (fabs(deltay/y)) < StopTolRel) return TRUE;
     else return  FALSE;
   }
   else 
   { // take the (GlobalEiParam)th root of deltay
     if (deltay < StopTolAbs) 
        return TRUE;
     else if (y!=0.0 && (fabs(deltay/y)) < StopTolRel) 
        return TRUE;
     else return  FALSE;
   }
 }
 else if (strcmp(StopCrit,"quantile")==0)
 {
  //note : augmR calculated by mle 2 steps prior in minimize  
  krigblup(&krmin[kam->pvec[0]],npred,xpred,ymat);      //compute ymat
  //cputime(21,"Remove zero covriance");
  adj_npred = remove_zero_cov(); //remove elements corresponding to 0 variance
  //cputime(22,"Remove zero covariance");

  nrerror("This option has been removed"); 
  /* status = orderalg(&krmin[kam->pvec[0]],adj_npred,yvec,xpred,h,npred); */
  return (!status);
 }
 else nrerror("ConvergeCrit: This method doesn't exist ");
  
 return(FALSE); // can't happen, just done so that  compiler doesn't complain
}
/************************************************************************/
/* Version : June 19, 1996                                              */
/* Purpose : Find Other d components of starting values                 */
/* input   : d   :  dimension of a point		 		*/ 
/* 	     d+1 :  number of starting values				*/ 
/* 	     range: d,2 matrix of upper and lower ranges 		*/ 
/* 	     startx: d+1,d matrix (untransformed)			*/
/* output  : startei: d+1 vector of y start values			*/ 
/* 	     startx: d+1,d matrix (transformed to +- inf) 		*/ 
/* NOTE : for MinimizeStage 'distanceneighbour' only uses the initial   */
/*        points , not the additionally sampled 			*/
void start_values_krmin(int d,int n,real **range, real **startx, real *startei,
   char *method)
{
   int j ,k;
   real alpha=0.01; /* multiplicativ displacement for starting values */   
   real *sm_dist=NULL,*sm_dist2=NULL;//additive displacement for starting values
   real **trafo_startx=NULL;
   real *start_point=NULL;
   real factor ; // decides whether +alpha or -alpha in displacement
   int *direction=NULL; //same as factor, except for all dims simultaneously
   int *dir2=NULL;

    alloc(d+1,d,&trafo_startx);
    alloc(d,&start_point);

    if (strcmp(method,"displacement")==0)
    { /* find the OTHER d elements of starting vector for simplex by 
       increasing one of d dimensions by (alpha)% in turns */
     if (PtsOutOfRange) 
        nrerror("start_values_krmin: PtsOutOfRange and displacement");
     alpha = .5 / (real(n)-1.0);
     copy_real_vector(d,startx[0],start_point);
     for (j=0; j<(d+1); j++)
     { // for all (d+1) starting points 
       for (k=0; k<d; k++)
       {  // in each of the d dimensions
         if (k == j-1 || j==0)
         {//change all dims of first startpt and in one dim each for all others
           //ensure this yields a valid starting point if region is constraint
           factor = (RandomGenerator<0.5) ? (-1.0):(1.0);
           startx[j][k]=start_point[k]+ factor*alpha*(range[k][1]-range[k][0]);
           if (startx[j][k] > range[k][1] || startx[j][k] < range[k][0])
           {
             startx[j][k]=start_point[k]-factor*alpha*(range[k][1]-range[k][0]);
             if (startx[j][k] > range[k][1]  || startx[j][k] < range[k][0])
                 nrerror("start_values: range < +-alpha%,modify start values");
                 // or just build a loop that makes alpha successively smaller
           }
         }
         else
           startx[j][k] = start_point[k];
        }
      }
    }
    else if (strcmp(method,"distanceneighbour")==0)
    { /* find the OTHER d elements of starting vector for simplex by
       increasing one of d dimensions by distance in turns */
     if (PtsOutOfRange) 
        nrerror("start_values_krmin: PtsOutOfRange and distanceneighbour");
     copy_real_vector(d,startx[0],start_point);
     alloc(d,&direction); alloc(d,&sm_dist);
     alloc(d,&dir2); alloc(d,&sm_dist2);
     for (k=0; k<d; k++)
     {
         direction[k] = (RandomGenerator<0.5) ? (-1):(1);
         dir2[k] = -direction[k];
     }
     min_distance_each_dim(*krmin->n,d,start_point,krmin->X,range,
         direction,sm_dist);
     min_distance_each_dim(*krmin->n,d,start_point,krmin->X,range,
         dir2,sm_dist2); //recalculate because of end-of-range-effects 
     /*
     printvecr(d,start_point,"start_point"); 
     printveci(d,direction,"direction"); 
     printvecr(d,sm_dist,"min distance"); 
     printveci(d,dir2,"direction2"); 
     printvecr(d,sm_dist2,"min distance2"); 
     */
     for (j=0; j<(d+1); j++)
     { // for all (d+1) starting points
       for (k=0; k<d; k++)
       {  // in each of the d dimensions
         if (k == j-1 && j!=0)
         {//change all dims of first startpt and in one dim each for all others
           //pending on factor go in + or - direction 
           startx[j][k]=start_point[k]+ real(direction[k])*.5*sm_dist[k];
           //ensure this yields a valid starting point if region is constraint
           if (startx[j][k] > range[k][1] || startx[j][k] < range[k][0])
             nrerror("start_values: halfneighbour , code bug");
         }
         else if (j==0)
         { // first startpt use all - directions
           startx[j][k]=start_point[k]+ real(dir2[k])*.5*sm_dist2[k]; 
           if (startx[j][k] > range[k][1] || startx[j][k] < range[k][0])
             nrerror("start_values: halfneighbour , code bug 2");
         }
         else
           startx[j][k] = start_point[k];
       }
     }
     frei(d,&direction); frei(d,&sm_dist);
     frei(d,&dir2); frei(d,&sm_dist2);
    }
    else if (strcmp(method,"random")==0)
    {
       // transform into full range
       for (j=0; j< (d+1); j++)
         for (k=0; k< d; k++)
            startx[j][k] =(range[k][1]-range[k][0])*RandomGenerator+range[k][0];
    }
    else nrerror("start_values: Unrecognized Method"); 

    /* calc the ExpectedImprovement of startx, including for the first row */
    /* note : need to transform startx to +- inf */ 
    trafo_ab_inf(d+1,d,startx,range,trafo_startx,trafo_method);
    copy_real_matrix(d+1,d,trafo_startx,startx);
    for (j=0; j< (d+1); j++)
      startei[j] = ExpectedImprovement(startx[j], d);

    frei(d+1,d,&trafo_startx);
    frei(d,&start_point);
}
/************************************************************************/
/* Version : November 15, 1996                                          */
/* Purpose : calculate the recursive integral that belongs to expected  */
/*           improvement^p						*/
real recurse_integral(int p,real pdfprob, real cdfprob, real zmin)
{
   if (p<0)
      nrerror("recurse_integral: p<0");
   else if (p==0)
      return cdfprob;
   else if (p==1)
      return (-pdfprob);
   return ( (real(p)-1.0)*recurse_integral(p-2,pdfprob,cdfprob,zmin)
                - pow(zmin,real(p-1)) * pdfprob );
}
/************************************************************************/
/* Version : Feb 14,1997			 			*/
/* Purpose : calc. prod=P(Y_c1<b_c1)*P(Y_c2<b_c2)*...			*/ 
/* input :     x    d,1 vector untransformed !				*/	/* 	       d    dimension of x					*/	
/* on return : prod							*/
real ConstraintProb(real **trafo_x, int d)
{
  int i;
  real prob = 1.0, mu, sigma, b1,b0,cdfprob1,cdfprob0;
  real *ypred=NULL;

  alloc(2,&ypred);
  for (i=0;i<(kam->c);i++)
  {
    krigblup(&krmin[kam->cvec[i]],1,trafo_x,&ypred,NoCV);
    mu = ypred[0];
    sigma = ypred[1];
    if (sigma==0) sigma = REAL_EPS; // otherwise division over 0
    b0= kam->ydescrip[kam->cvec[i]][0];
    b1= kam->ydescrip[kam->cvec[i]][1];
    // if des_to_min do we want to change the sigma ? 
    if (b0> -REAL_INFINITY)
      cdfprob0 = cdfgauss((b0-mu)/sigma,TRUE);
    else cdfprob0 = 0.0;
    if (b1< REAL_INFINITY)
      cdfprob1 = cdfgauss((b1-mu)/sigma,TRUE);
    else cdfprob1=1.0;
    prob *= (cdfprob1 - cdfprob0); 
  }
  if (DEBUG_EI)
     fprintf(ofp,"b0%f b1%f mu%f sigma%f finalprob%f \n",b0,b1,mu,sigma,prob);
  frei(2,&ypred);
  return prob;
}
/************************************************************************/
/* Version : June 15, 1996; Nov15,1996; Oct13,1997 			*/
/* Purpose : calculate the negative expected improvement -E(I) at ONE point*/ 
/* input   : x :   d,1 vector , transformed (!), ranging (-inf,inf)     */ 
/*           dopt: d   scalar , int 					*/ 
/* output  : returns -E(I) at *x					*/
/* Note    : dopt is not explicitly needed, but required for the optimization*/
/* Note    : U, eigvalinv ,eigval need to be calc prior to this 	*/
real ExpectedImprovement(real *x, int dopt)
{
  int d,i;
  real ei, mu, sigma, sigma_anyX, z, pdfprob, cdfprob; 
  real *ypred=NULL, *trafo_x=NULL;

  d = krmin[kam->pvec[0]].d[0]; /* can't use 'dopt' because of size_t */
  alloc(d,&trafo_x); 
  alloc(2,&ypred);  /* for mu and sigma */

  trafo_inf_ab(1,d,&x,krmin[kam->pvec[0]].range,&trafo_x,trafo_method);// care 
  krigblup(&krmin[kam->pvec[0]],1,&trafo_x,&ypred,NoCV);

  mu = ypred[0];
  sigma = ypred[1]; // ypred is later overwritten
  if (des_to_min) {
    krigblup_anyXY(&krmin[kam->pvec[0]],anyX,anyX[0], anyX_rows,anyX_cols,1, 
      &trafo_x, &ypred);  // ypred= anyX[0] is garbage, blup not needed though
    sigma_anyX = ypred[1]; // mse does not depend on ypred (which is garbage)  
  }
  else 
    sigma_anyX=sigma;

  if (DEBUG_EI)
  {
    printvecr(d,trafo_x,"--x :");
    fprintf(ofp,"ymin %f mu:%f sigma:%f new sigma:%f",ymin,mu,sigma,sigma_anyX);
  }
  if (sigma!=0.0)
  {
     z  = (MinTarget-mu) / sigma ;
     if (GlobalEiParam== 1)
       ei = sigma_anyX*( z* cdfgauss(z,TRUE) + pdfgauss(z));
     else if (GlobalEiParam==0)
     { 
         ei = cdfgauss(z,TRUE); 
         if (ei==0) ei = -log(-z); // numerical trick 
     }
     else if (GlobalEiParam>1)
     {
       pdfprob = pdfgauss(z);
       cdfprob = cdfgauss(z,TRUE);
       for (i=0, ei=0.0;i<=GlobalEiParam;i++)
       {
         ei += choose(GlobalEiParam,i) * pow(-1.0,real(i%2)) * 
               pow(z,GlobalEiParam-i)*
               recurse_integral(i,pdfprob,cdfprob,z);
       }
       // ei *= pow(sigma_anyX,real(GlobalEiParam));
       // if sigma_anyX is very small pow may fail
       ei = pow(ei,1.0/real(GlobalEiParam)); // criterion is p^th root 
       ei *= sigma_anyX; //equivalent to  ei*=((sigma_anyX)^p)^(1/p)
     }
     else
       nrerror("ExpectedImprovement: GlobalEiParam<0 ");
     if (constraint)
        ei *= ConstraintProb(&trafo_x,dopt);
  }
  else ei = 0.0 ;
  if (DEBUG_EI) fprintf(ofp,"ei^p*constraints : %f \n",ei); 

  frei(2,&ypred);
  frei(d,&trafo_x);
  return -ei;
}
/************************************************************************/
/* Version : June 26, 1996 - Nov 04,97					*/
/* Purpose : Find the Maximum Expected Improvement over the range of X  */
/*           given the previous  data	 		 		*/    
/* output  : xbest :   d vector of x's with maximal expected improvement*/
/*	     MaxExpImp: vector (only one elem.), contains  EI(xbest) 	*/     
/* on return : 0 if ok 							*/
/*             1 if no EI>0.0 could be found (numerically =0.0)	 	*/
/*             2 if the simplex algorithm never started a search	*/
/*	       3 if pts out of range and not random local searches 	*/
/*WARNING: uses only first kr model all the way through !!!!		*/
/* NOTE   : U,eigenvalue, eigenvalinv, R must be calc before this 	*/
#define MIN(a,b) ((b) < (a) ? (b) : (a))      /* min(a,b)*/
#define RANDOM_SEARCH_OUT_OF_RANGE "ERROR: When one or more X points are outside of the optimization range,\n    CriterionOptimStart=random needed.\n" 
int MaximalExpectedImprovement(KrigModel *kr,real *MaxExpImp,real *xbest)
{
   void start_values_krmin(int d,int n,real **range,real **startx,real *ExpImp,       char *method);
   int i,d, iter=0, iteradd; 
   int status=1; // lcoal optimum not found so far 
   real *startExpImp=NULL, **startx=NULL, *trafo_xbest=NULL, *tempx =NULL;
   int search_local_opt=0; // number of succesfull searches for local optimum 
   int count_maxiter_reached=0; //number of times tries were interrupted
   int search_not_started=0; //number of times local ei search didn't start 
   string MaxMethod; // for output which method found MaxExpImp


   d = *kr->d;
   alloc(d+1, d,&startx); alloc(d+1,&startExpImp);
   alloc(d,&trafo_xbest); alloc(d,&tempx);

   //kr_cover_augmR(kr); //kr_calc_augmR(kr);  needed for krigblup 
   init_real_vec(-1.0,d,trafo_xbest);
   *MaxExpImp=0.0;
   if (CriterionOptimTries== -1) CriterionOptimTries=2*(*kr->n);//MIN(*kr->n,50)
   if (PtsOutOfRange && strcmp(CriterionOptimStart,"random")!=0)
     { fprintf(ofp,RANDOM_SEARCH_OUT_OF_RANGE); return 3; }
   i=0;
   cputime(11,"Global Optimization EI"); //  for global ei search
   do { // local optimization searches     
     /* note startx is transformed  to (-inf,+inf) afterwards */
     /* i%(*kr->n) guarantees that the expression is valid for X */
     copy_real_vector(d, kr->X[i%(*kr->n)],startx[0]);
     if (i< *kr->n)
      start_values_krmin(d,initial_des_n,kr->range,startx,startExpImp,
        CriterionOptimStart);
     else  
      start_values_krmin(d,initial_des_n,kr->range,startx,startExpImp,"random");
     //note: for abs and rel criteria read note in kriginit.cc, don't change!

     iteradd=Minimize(ExpectedImprovement,OptTolAbs,StopTolRel/100.0,MAX_ITER,
       d, startx,startExpImp,CriterionOptimMethod);
     iter += iteradd;
     if (iteradd>=MAX_ITER)   count_maxiter_reached++;
     if (iteradd>0) search_local_opt++;
     if (iteradd==0) search_not_started++;
   
     if (-startExpImp[0] > *MaxExpImp)
     {
       status = 0; /*ok */
       *MaxExpImp = -startExpImp[0]; /* this is a scalar */
       copy_real_vector(d, startx[0], trafo_xbest);
       if (i< *kr->n) strcpy(MaxMethod,"original");
       else           strcpy(MaxMethod,"random");
     }
     if (iteradd!=0 && DEBUG)
     {
      fprintf(ofp,"MaxEI-Try %3d: ",i+1);
      fprintf(ofp,"local MaxExpImp: %.2e ExpImp-evals: %3d \n",-startExpImp[0],iteradd);
      trafo_inf_ab(1,d,&startx[0],kr->range,&tempx,trafo_method);
      printvecr(d,tempx,"local Max :");
     }
     //if (i%100 ==0  && i> *kr->n)  change as needed !!!
     if (i%100 ==0  && i> 2* *kr->n) 
     {
       nrwarning("Optimization has difficulties finding local optima.");
       if (OptTolAbs > REAL_EPS)
       {OptTolAbs /=10.0;  OptTolRel /=10.0;}
     }
     i++;
   } while (( (i<CriterionOptimTries) || search_local_opt<3) && i<MAX_ITER );
   if (strcmp(MaxMethod,"random")==0)
      fprintf(ofp,"INFO: Method 'random' found MaxExpImp\n");
   if (count_maxiter_reached>0 || search_not_started>0)
   {
     fprintf(ofp,"INFO: %d Optimization Tries.\n",CriterionOptimTries); 
     if (count_maxiter_reached>0)
      fprintf(ofp,"\t %d tries interrupted. (MAXITER=%d)\n",
        count_maxiter_reached, MAX_ITER);  
     if (search_not_started>0)
      fprintf(ofp,"\t %d tries failed to start.\n",search_not_started);
   } 
   if (search_local_opt<3) 
   {
     nrwarning("Optimization explored less than 3 tries.");
     status = 2; // Optimization didn't explore in any of the attempts
   }
   /* detransform xbest */
   trafo_inf_ab(1,d,&trafo_xbest,kr->range,&xbest,trafo_method);
   frei(d+1,&startExpImp); frei(d,&trafo_xbest); 
   frei(d+1,d,&startx); 
   cputime(12,"Global Optimization EI"); 
   return (status);
}
#undef MIN
#undef RANDOM_SEARCH_OUT_OF_RANGE
/************************************************************************/
/* Version : Jan 30, 1997 						*/
/* Purpose : Decide whether to do the mle 				*/	
/* on input :  i  number of iteration starting with 0			*/
/* on return : 1  yes, do mle  0, don't do mle				*/
/* Note : if MinimizationMlePerc is active decide based on that		*/
/*	  otherwise decide based on MinimizationMleFrequency		*/
boolean do_mle(int i, int n , int last_mle_n)
{
    if (last_mle_n==0) nrerror("do_mle: Code Bug , last_mle_n==0");
    if (n==100 && MinimizationMleFrequency<10) 
       MinimizationMleFrequency=10;

    if (i==0) return TRUE; //do the mle initially
    if (n>=MaxObsForMle) return FALSE; // no mle beyond MaxObsForMle
    if (MinimizationMlePerc!=0.0 &&
        (real(n)/real(last_mle_n) -1.0) >= MinimizationMlePerc)
       return TRUE; 
    else if (MinimizationMlePerc==0.0 && (i%MinimizationMleFrequency)==0) 
       return TRUE; // only update mle every (MleFrequency)st time
    return FALSE;
}
/************************************************************************/
/* Version : March 24, 1997 						*/
/* Purpose : output for each new sampled point in minimize 		*/
void  minimize_output(int n,int d,real *xbest, int ny, real *newy, 
   real *MaxExpImp, int pos_pvec, boolean constraint)
{
  int j;

     printvecr(d,xbest,"xbest:");
     if (constraint)
     { // print prob and all y's
        fprintf(ofp,"Responses at xbest: "); 
        for (j=0;j<ny;j++)  
        {
          if (newy[j]!= -REAL_INFINITY)
            fprintf(ofp," %7.3e ",newy[j]);
          else fprintf(ofp," NA ");
        }
        fprintf(ofp,"\n");
        fprintf(ofp,"Criterion:%6.4e ",*MaxExpImp);
        fprintf(ofp,"RelCrit:%6.4e ",*MaxExpImp/fabs(MinTarget));
        fprintf(ofp,"P(constr. met)=%5.3f ",ConstraintProb(&xbest,d));
        fprintf(ofp,"ymin:%6.4e",ymin); // note MinTarget may not be ymin
     }
     else 
     {
        fprintf(ofp,"y(xbest): %7.5e\n",newy[pos_pvec]);
        fprintf(ofp,"Criterion: %7.5e ",*MaxExpImp);
        fprintf(ofp,"RelCrit: %7.5e ",*MaxExpImp/fabs(ymin));
        fprintf(ofp,"ymin:%7.5e ",ymin); 
     }
     fprintf(ofp,"n=%3d\n",n+1);
     if (ymin<MinTarget) nrwarning("MinTarget< Minimal Sampled y Value");
}
/************************************************************************/
/* Version : March 28, 1997 - Nov 27 , 97                               */
/* Purpose : compute ymin at start of Minimization/Design to Minimize   */
/* uses static global vars : constraint,ymin , RandomError		*/
/* NOTE: don't really need ok_case if !RandomError, but allocated anyways*/
void calc_ymin(KrigArray *ka,KrigModel *kr)
{
   int i,counter,n,r,q, d,index;
   real *ok_y=NULL,  **ymat=NULL;
   int *ok_case=NULL;
   char ylab[MAXLINE];
   real *ypred=NULL;

   n = *kr[ka->pvec[0]].n;
   d = *kr[ka->pvec[0]].d;
   r = *kr[ka->pvec[0]].r;
   if (constraint)
   {
     alloc(n,&ok_y);
     alloc(n,&ok_case);
     alloc(n,ka->p+ka->c,&ymat);
     if (construct_ymat(kr,ka,ymat,ylab,MAXLINE)!=0)
        nrerror("calc_ymin: y labels are too long .");
     for (i=0,counter=0; i<n; i++) 
     {
       if (check_constraints(1,ka->p+ka->c,&(ymat[i]),ka->ydescrip)==0)
       {  
          ok_y[counter]= kr[ka->pvec[0]].y[i]; 
          ok_case[counter]=i+1; // keep track of which cases meet constraint
          counter++;
       }
     }
     ymin = RealMinimumIndex(counter,ok_y,&index);
     index = ok_case[index-1]; // get corresponding index of X matrix  
     frei(n,&ok_y); frei(n,&ok_case);
     frei(n,ka->p+ka->c,&ymat);
   }
   else
     ymin = RealMinimumIndex(n,kr[ka->pvec[0]].y,&index); // return index,ymin

   if (RandomError)
   { // don't use ymin but E(ymin) for the x corresponding to the lowest y  
     // in future may want to compute all E(y|design x) and take the smallest  
     //    one 
     // rationale for using E(y| design x) rather than just any x: 
     //    design points have  stochstic process variance ==0 , hence 
     //    know design points with more certainty 
     alloc(2,&ypred);
     q = ka->pvec[0];
     init_krigblup_anyXY(&kr[q],kr[q].X,n,d,r);
     krigblup_anyXY(&kr[q],kr[q].X,kr[q].y,n,d,1,&kr[q].X[index-1],&ypred);
     free_krigblup_anyXY(n,r);
     ymin = ypred[0];
     frei(2,&ypred);
   }

   return;
} 
/************************************************************************/
/* Version : March 28, 1997 - Nov 12, 97                                */
/* Purpose : update ymin when one new point is sampled 			*/
/* uses static global vars : constraint,ymin, RandomError		*/
/* new y is a matrix because check_constraints needs it 		*/
void update_ymin(KrigArray *ka,real **newy)
{
  real thisy;

  thisy= newy[0][ka->pvec[0]];
  if (constraint && thisy<ymin  
     && check_constraints(1,ka->p+ka->c,newy,ka->ydescrip)==0) 
         ymin = thisy;
  else if (!constraint && thisy<ymin) 
         ymin = thisy; 
  // else leave ymin untouched 
  return;
}
/************************************************************************/
/* Version : June 26, 1996 						*/
/* Purpose : minimize a user-supplied function with the Expected 	*/
/*           Improvement algorithm					*/ 
/* return  : 0 ok   							*/
/*	     1 failed to find max. Improvement numerically > 0.0  	*/
/*	     2 failed to find 3 successful optimization searches        */ 
/* Note    : F, R need to be calc. prior to this			*/
/* NOTE    : all y's are sampled in the order the appear in y.mat 	*/
int minimize(KrigModel *kr,KrigArray *ka) 
{
 void  minimize_output(int n,int d,real *xbest, int ny, real *newy, 
     real *MaxExpImp, int pos_pvec, boolean constraint);
 boolean constraint_init(KrigModel *kr, KrigArray *ka, boolean fit_params);
 boolean do_mle(int i, int n , int last_mle_n);
 void update_ymin(KrigArray *ka,real **newy);
 int i,j,n ,d, nr, error = 0; 
 int last_mle_n=1; // 0 would result into error/ devision by 0
 real *MaxExpImp=NULL, *xbest=NULL, *newy;
 int converged; // number of consecutive converged trials
 char top_line[MAXLINE];

 //don't need to check for alpha,theta,beta because krig_mle will create them
 if (initialized(ka) && (ka->c==0 || constraint_init(kr,ka,FALSE)) && 
   initialized(kr[ka->pvec[0]].init,"X","y","range","FuncName",NULL))
 {
  krmin = kr; kam=ka; 
  des_to_min=FALSE;
  if (ka->c>0)  constraint=TRUE;
  else          constraint=FALSE;
  if (ka->p>1)  nrinfo("Minimize for first non - constraint only.");
  database(krmin,ka); // care this is the first one , not first non-response
  n = krmin[ka->pvec[0]].n[0]; d= krmin[ka->pvec[0]].d[0];
  nr = n + krmin[ka->pvec[0]].r[0];
  initial_des_n =n;
  calc_ymin(kam,krmin); 
  if (!init_MinTarget) MinTarget=ymin;

  alloc(1,&MaxExpImp);
  alloc(d,&xbest);
  alloc(ka->p+ka->c,&newy);
  if (strcmp(StopCrit,"quantile")==0)
  { 
     // create xpred 
     grid(GridSize,d,krmin[ka->pvec[0]].range,&xpred,&npred);
     //printmatr(npred,d,xpred,"xpred:");
     alloc(npred,&h);
     alloc(npred,&yvec);
     alloc(npred,2,&ymat);
  }
 
  i=0; // only for initial mle
  converged = 0;
  do 
  { // sample while not converged 
    if  (do_mle(i,n,last_mle_n))
    { // do the mle , done initially, and also at the end 
       fprintf(ofp,"Fitting Response %s...\n",krmin[ka->pvec[0]].ylabel);
       krig_mle(&krmin[ka->pvec[0]],FALSE); 
       for (i=0;i< ka->c;i++)
       {
         fprintf(ofp,"Fitting Constraint %s...\n",krmin[ka->cvec[i]].ylabel);
         krig_mle(&krmin[ka->cvec[i]],FALSE);
       }
       last_mle_n = n;
       if (n>=100) 
       {
         sprintf(top_line,"Temporary storage of response with %d observations.",           n);
         output_temp(&krmin[ka->pvec[0]],top_line); // save temporary matrices 
       }
    }
    else 
    {//update R, F and U because columns corresponding to new point not calc.  
       kr_calcR(&krmin[ka->pvec[0]]);
       kr_calcF(&krmin[ka->pvec[0]]);
       kr_cover_augmR(&krmin[ka->pvec[0]]); //needed for krigblup
       for (i=0;i< ka->c;i++)
       {
         kr_calcR(&krmin[ka->cvec[i]]);
         kr_calcF(&krmin[ka->cvec[i]]);
         kr_cover_augmR(&krmin[ka->cvec[i]]);
       }
    }
    i++; // only needed for do_mle()
    error = MaximalExpectedImprovement(&krmin[ka->pvec[0]],MaxExpImp,xbest);
    if (MinSample==0 || MinSample <= krmin[ka->pvec[0]].n[0]) 
    {
      if (strcmp(StopCrit,"quantile")==0) 
         cputime(11,"1st Order Statistic");
      if (ConvergeCrit(MinTarget, *MaxExpImp))  // care, MinTarget used !!! 
         converged++;
      else 
      { 
         if (converged>0) 
            nrinfo("3 Consecutive Convergences Interrupted");
         converged =0; // number of consecutive convergences is zero
      }
      if (strcmp(StopCrit,"quantile")==0) 
         cputime(12,"1st Order Statistic");
    }
    else 
        converged = 0;
    if (!error && converged<3)
    { 
     /* sample at xbest and get newy */
     sample_new_y(FuncName,d,xbest,ka->p+ka->c,newy); 
     //newy[0]=log(GPFunc(xbest,d)); nrwarning("GP function integrated");
     if (RandomError) calc_ymin(kam,krmin); // somewhat inefficient
     else        update_ymin(ka,&newy); 
     if (!init_MinTarget) MinTarget = ymin;
     minimize_output(n,d,xbest,ka->p+ka->c,newy,MaxExpImp,ka->pvec[0],
        constraint);

     /* reallocate and adjust krmin.X, krmin->y, n */
     for (j=0;j<(ka->p+ka->c);j++)
     {
      morealloc(n,d,&krmin[j].X,n+1,d);
      copy_real_vector(d,xbest,krmin[j].X[n]);//the (n+1)th elem of X
      morealloc(n,&krmin[j].y,n+1);
      krmin[j].y[n] = newy[j]; // all y's are sampled in order of appearance 
      //eigval, eigvalinv, U, update done in krigblup, but need right dimension
      morealloc(nr,nr,&krmin[j].U,nr+1,nr+1);
      morealloc(nr,&krmin[j].eigval,nr+1);
      morealloc(nr,&krmin[j].eigvalinv,nr+1);
      krmin[j].init->U = FALSE;
      /*mle updates R and F automatically as long as dimensions ok */
      morealloc(n,n,&krmin[j].R,n+1,n+1);
      morealloc(n,krmin[j].r[0],&krmin[j].F,n+1,krmin[j].r[0]);
      krmin[j].init->R = FALSE;
      krmin[j].init->F = FALSE;
      (krmin[j].n[0])++; 
     }
     n++; 
     nr= n+ krmin[0].r[0];// must be last, assuming all r are equal
     fprintf(ofp,"--------------------------------------------------------\n"); 
    }
    else
    { //error or converged 
       fprintf(ofp,"Final MaxCriterion : %f\n",*MaxExpImp);  
       fprintf(ofp,"Best Value found: ymin=%6.4e\n",ymin);//also give x values?
    }
   } while (converged<3 && !error);

   frei(1,&MaxExpImp);
   frei(d,&xbest);
   frei(ka->p+ka->c,&newy);
   if (strcmp(StopCrit,"quantile")==0)
   { 
      frei(npred,&yvec); frei(npred,&h);
      frei(npred,krmin[ka->pvec[0]].d[0],&xpred); frei(npred,2,&ymat);
   }
 }
 exec_system("rm space.temp");
 return error;
}
/************************************************************************/
/* Version : Feb 14, 1997                                               */
/* Purpose : are all constraints initialized for X,y, and range         */
/* input  : fit_params   TRUE if to be checked for fitting parameters   */
/*                       FALSE otherwise 				*/ 
/* return : 0 ok, 1 at least one constraint not initialized             */
/*  NOTE : what is to be checked depends on fit_params			*/
boolean constraint_init(KrigModel *kr, KrigArray *ka, boolean fit_params)
{
   boolean init;
   int i;

   init = TRUE;
   for (i=0;i<ka->c;i++)
   { 
     if (!initialized(kr[ka->cvec[i]].init,"X","y","range",NULL,NULL))  
        init=FALSE;
     if (fit_params && !initialized(kr[ka->cvec[i]].init,"alpha","beta",
     "theta","var",NULL))
        init=FALSE;
   }
   if (init==FALSE) 
       fprintf(ofp,"The previous message(s) refer to the constraints\n");
   return init;
}
/************************************************************************/
/* Version : June 26, 1996                                              */
/* Purpose : design n points such that they are likely to find the minimum*/
/* input   : number   number of points to be designed 			*/
/*	     outputfile name of the file for outputting design 		*/
/* on return : 0  ok							*/
/*	       1, 2 problems						*/   	
int MinimizeStage(KrigModel *kr, KrigArray *ka,int number, char *outfile)
{
   boolean constraint_init(KrigModel *kr, KrigArray *ka,boolean fit_params);
   int i, error=0, n,d,r, anyX_max_rows;
   real *xbest=NULL, *MaxExpImp=NULL;
   real root_crit;

 //don't need to check for alpha,theta,beta because krig_mle will create them
 if (initialized(ka) && (ka->c==0 || constraint_init(kr,ka,TRUE))  && 
  initialized(kr[ka->pvec[0]].init,"X","y","range",NULL,NULL) &&
  initialized(kr[ka->pvec[0]].init,"alpha","beta","theta","var",NULL))
 {
   krmin = kr; kam =ka;
   des_to_min=TRUE; 
   if (ka->c>0)  constraint=TRUE;
   else 	 constraint=FALSE;
   if (ka->p>1)  nrinfo("MinimizeStage for first non - constraint only.");
   n = krmin[ka->pvec[0]].n[0]; d= krmin[ka->pvec[0]].d[0]; 
   r = krmin[ka->pvec[0]].r[0];
   anyX_rows=n; anyX_cols=d; anyX_max_rows = n+number;
   initial_des_n= n;
   calc_ymin(kam,krmin); 
   if (!init_MinTarget) MinTarget=ymin;
   else if (ymin<MinTarget) nrwarning("MinTarget< Minimal Sampled y Value");

   database(krmin,ka); // care , this is the first , not the first response!
   fprintf(ofp,"minimal y: %f",ymin);  
   if (MinTarget!=ymin) fprintf(ofp," MinTarget: %f",MinTarget);
   fprintf(ofp,"\n");

   alloc(1,&MaxExpImp); alloc(d,&xbest);
   alloc(anyX_max_rows,anyX_cols,&anyX);
   copy_real_matrix(n,d,krmin[ka->pvec[0]].X,anyX);//note anyX is larger than X
   for (i=0; i< ka->p+ka->c; i++) 
   {  // needed if no fitting done
     ensure_init(&krmin[i],"R");
     ensure_init(&krmin[i],"F");
     ensure_init(&krmin[i],"U");
   }

   for (i=n;i<(n+number); i++)
   { // find 'number' design points
      init_krigblup_anyXY(&krmin[ka->pvec[0]],anyX,anyX_rows,anyX_cols,r);
      error = MaximalExpectedImprovement(&krmin[ka->pvec[0]],MaxExpImp,xbest);
      free_krigblup_anyXY(anyX_rows,r);
      if (!error) 
      {
         copy_real_vector(d,xbest,anyX[i]); //add xbest to X 
         anyX_rows++;
         printvecr(d,xbest,"xbest :");
         root_crit = pow(*MaxExpImp,1.0/real(GlobalEiParam));
         fprintf(ofp,"Criterion: %10g ",root_crit);
         fprintf(ofp,"RelCrit:%6.4e ",root_crit/fabs(MinTarget));
         if (constraint)
           fprintf(ofp,"P(constraints met): %5.3f ",ConstraintProb(&xbest,d));
         fprintf(ofp,"n= %3d\n",anyX_rows);
         fprintf(ofp,"--------------------------------------------------------\n"); 
      }
      else break;
   }
   if (!error)
   { //output anyX
     if (strcmp(outfile,"screen")==0 || strcmp(outfile,"s")==0) 
     {
       printvecs(d,krmin[ka->pvec[0]].xlabel,"Labels:");
       printmatr(anyX_rows,anyX_cols,anyX,"anyX");
     }
     else
     {
       fprintf(ofp,"Writing design points to '%s' file...\n",outfile);
       outputmatr(anyX_rows,anyX_cols,anyX,outfile,krmin[ka->pvec[0]].xlabel,
        NULL,"MinimizeStage Matrix");
     }
   }
   frei(1,&MaxExpImp); frei(d,&xbest);
   frei(anyX_max_rows,anyX_cols,&anyX);
 }
 return error;
}
/************************************************************************/
// test ei at one point
/*
real temp, *temp_trafo_x=NULL, *mypoint=NULL;

alloc(d,&temp_trafo_x);
alloc(d,&mypoint);
mypoint[0]=3.142; mypoint[1]= 2.275;
trafo_ab_inf(1,2,&mypoint,krmin->range.mat,&temp_trafo_x,trafo_method);
temp =  ExpectedImprovement(temp_trafo_x,2);
cout << "###EI at center global minimum : " << -temp << endl;
*/
