#include <stdio.h>
#include <string.h>
#include <stdlib.h> /* also used  for atoi */
#include <math.h>

#include "general.h"  
#include "alloc.h"  
#include "matrix.h"  
#include "krig.h"

extern FILE *ofp;

/************************************************************************/
/* Version : May 27, 1996 - Feb 11, 97					*/
/* Purpose : for 1 KrigingModel, allocate space for its structure elems */
/* and initialize  some of them a KrigingModel				*/
void init_KrigModel(KrigModel *kr,int n,int d,int r)
{
  int i,j;

  /* all this needs to be reallocated if the n,d,r,p change */
  alloc(d,&kr->theta);
  alloc(d,&kr->alpha);
  alloc(r,&kr->beta); 
  alloc(n,&kr->y);
  alloc(n+r,&kr->eigval); 
  alloc(n+r,&kr->eigvalinv); 
  alloc(n,d,&kr->X);
  alloc(n,n,&kr->R);
  alloc(n+r,n+r,&kr->U); 
  alloc(n,r,&kr->F); 
  alloc(1,&kr->d);
  alloc(1,&kr->n);
  alloc(1,&kr->r);
  alloc(1,&kr->var);
  alloc(1,&kr->stochratio);
  alloc(1,&kr->loglike);
  alloc(d,2,&kr->range);
  alloc(d,&kr->xlabel);

  kr->init->X = FALSE;
  kr->init->y = FALSE;
  kr->init->R = FALSE;
  kr->init->F = FALSE;
  kr->init->beta = FALSE;
  kr->init->alpha = FALSE;
  kr->init->theta = FALSE;
  kr->init->U = FALSE;
  kr->init->var = FALSE;
  kr->init->range = FALSE;

  kr->init->d = TRUE;
  kr->init->n = TRUE;
  *kr->d = d;  /* final d is set in kr.X */
  *kr->n = n;  /* final n is set in kr.X */
  *kr->r = r;  /* final r is set in kr.beta */
  *kr->stochratio = 1.0; /* change that if error allowed */ 
  *kr->loglike= -REAL_INFINITY;
  strcpy(kr->ylabel,"EMPTY");
  
  init_real_vec(0.0,d,kr->alpha);
  init_real_vec(0.0,d,kr->theta);
  init_real_vec(0.0,r,kr->beta);
  /* might need to modify this if n changes */
  for (i=0; i<n; i++)
     for (j=0; j<r; j++)
     {
        if (i==0) kr->F[i][j] = 1.0;
        else      kr->F[i][j] = 0.0;
     }
  // note U, R, X, y , eigval, eigvalinv are not assigned a value 
}
/************************************************************************/
/* Version : Feb 13, 97			                                */
/* Purpose : for 1 KrigingModel, set all vectors to NULL		*/
/* Structure does not allow initialize to zero.				*/
/* this is needed only because of auxiliary allocation routines 	*/
/* if omitted may prompt error, pointer already allocated		*/ 
void NULL_KrigModel(KrigModel *kr)
{
  kr->theta=NULL;
  kr->alpha=NULL;
  kr->beta=NULL;
  kr->y=NULL;
  kr->eigval=NULL;
  kr->eigvalinv=NULL;
  kr->X=NULL;
  kr->R=NULL;
  kr->U=NULL;
  kr->F=NULL;
  kr->d=NULL;
  kr->n=NULL;
  kr->r=NULL;
  kr->var=NULL;
  kr->stochratio=NULL;
  kr->loglike=NULL;
  kr->range=NULL;
  kr->xlabel=NULL;
}
/************************************************************************/
/* Version : Feb 13, 97			                                */
/* Purpose : for 1 KrigArray , set all vectors to NULL			*/
/* Structure does not allow initialize to zero.				*/
/* this is needed only because of auxiliary allocation routines 	*/
void NULL_KrigArray(KrigArray *ka)
{
   ka->pvec=NULL;
   ka->cvec=NULL;
   ka->ydescrip=NULL;
}
/************************************************************************/
/* Version : Feb 12, 1997 						*/
/* Purpose : supply Kriging Array with default values 			*/
void init_KrigArray(KrigArray *ka) 
{
   ka->init = FALSE; 
   ka->ydescrip_init=FALSE;
   ka->p=0;
   ka->c=0;
   ka->pvec = NULL;
   ka->cvec = NULL;
   ka->ydescrip = NULL;
}
/************************************************************************/
/* Version : Feb 12, 1997 						*/
/* Purpose : supply Kriging Array with default values 			*/
/* input : n   rows of ydescrip						*/
/*         oldn (if previously allocated) old number of n, else 0	*/
/* NOTE : should only be called once because otherwise old ydescrip   */
/* may accidently be overwritten , also pvec and cvec			*/ 
void default_KrigArray(int oldn,KrigArray *ka,int n)
{
  int j;

  ka->init = TRUE;
  if (n<=0) nrerror("default_KrigArray: Code Bug n<=0"); 
  if (oldn==0) 
  {
     alloc(n,3,&ka->ydescrip);
     alloc(n,&(ka->pvec));
     alloc(n,&(ka->cvec));
  } 
  else if (oldn>0) 
  {
     morealloc(oldn,3,&ka->ydescrip,n,3);
     morealloc(oldn,&(ka->pvec),n);
     morealloc(oldn,&(ka->cvec),n);
  }
  else nrerror("default_KrigArray: Code Bug , oldn<0");

  for (j=0;j<n; j++)
  { // initialize matrix to defaults
    ka->ydescrip[j][0]= -REAL_INFINITY;
    ka->ydescrip[j][1]= +REAL_INFINITY;
    ka->ydescrip[j][2]= FALSE; // no constraints
  }

  ka->p =n;
  ka->c =0;
  for (j=0;j<n;j++)
     ka->pvec[j] = j;
}
/************************************************************************/
/* Version : Feb 17, 1997                                               */
/* Purpose : undo changes to KrigArray in default 			*/
void undo_default_KrigArray(KrigArray *ka,int n)
{
  ka->init = FALSE;
  frei(n,3,&ka->ydescrip);
  frei(n,&(ka->pvec));
  frei(n,&(ka->cvec));
  ka->p =0; 
  ka->c =0;
}
/************************************************************************/
/* Version : Feb 12, 1997 						*/
/* Purpose : adjust p, c, pvec and cvec according to ydescrip		*/
/* Example : 3 var , 1st response , 2nd constr. , 3rd constraint 	*/ 
/* then p=1, c=2, pvec=(0,.,.), cvec=(1,2,.)				*/ 
/* NOTE that  pvec and cvec start with 0!				*/
void adjust_KrigArray(KrigArray *ka) 
{
  int i,n;
 
  if (!ka->init) nrerror("Code Bug :adjust_KrigArray, ka not initialized"); 
  n = ka->p + ka->c;
  ka->p = 0;
  ka->c = 0;
  for (i=0;i<n; i++)
  {
    if (ka->ydescrip[i][2] == TRUE) 
    {
      ka->cvec[ka->c]=i;
      (ka->c)++; 
    }
    else 			 
    {
      ka->pvec[ka->p]=i;
      (ka->p)++;
    }
  }
  if (ka->p + ka->c !=n) nrerror("Code Bug adjust_KrigArray");
  if (ka->p==0) 
   nrwarning("There are no reponses among the y-variables (only constraints).");
}
/************************************************************************/
/* Version : Feb 11, 1997                                               */
/* Purpose : copy  X,n,d,p,r,range into the seconds KrigModel           */
void copy_KrigModel(KrigModel *kr1,KrigModel *kr2)
{
   int n,d ,r;

   n =  *kr1->n;
   d =  *kr1->d;
   r =  *kr1->r;
   if (kr1->init->X)
   {
     copy_real_matrix(n,d,kr1->X,kr2->X);
     copy_string(d,kr1->xlabel,kr2->xlabel);
     kr2->init->X = TRUE;
   }
   else
     nrerror("copy_KrigModel: X in o-KrigModel not initialized , why ? ");
   if (kr1->init->range)
   {
     copy_real_matrix(d,2,kr1->range,kr2->range);
     kr2->init->range = TRUE;
   }
}
/************************************************************************/
/* Version : March 11, 1997                                             */
/* Purpose : to be called whenever (p+c) changes in KrigModel           */
/* input   : newm   number of KrigModels needed                         */
/*           kr->p+kr->c old number of KriModels (>0)                   */
/* output  : reallocated kr                                             */
/*           ka initialized / reinitialized and adjusted                */
/* return  : 0 ok                                                       */
int change_krig_p(KrigArray *ka,KrigModel **ptrkr,int newm)
{
    int j, oldm;
    KrigModel *kr=NULL;

    kr = *ptrkr;

    if (!ka->init)
      oldm=0; // in KrigArray, ydescrip pvec and cvec are not yet allocated
    else
      oldm= ka->p + ka->c;
    default_KrigArray(oldm,ka,newm);
    if (oldm==0) oldm=1; //1 KrigModel was allocated already
    if (oldm!=newm)
    { // reallocate KrigModel
      if ((kr = (KrigModel *) realloc(kr,(newm)*sizeof(KrigModel)))==NULL)
        nrerror("change_krig_p: Reallocation of KrigModel failure");
      else
      { // note only do this with KrigModels not yet allocated
        for (j=oldm;j<newm;j++)
        { // only for KrigModels
          NULL_KrigModel(&kr[j]);
        }
      }
    }

    *ptrkr = kr;
    return 0;
}
/************************************************************************/
/* Purpose : print the contenst of s kriging model for debugging        */
/* Version : Nov 15, 1997                                               */
void print_krigmodel(KrigModel *kr)
{
   InitStruct is;

   fprintf(ofp," n=%d, d=%d r=%d\n",*kr->n,*kr->d,*kr->r);
   fprintf(ofp," var=%f, stochration=%f \n",*kr->var ,*kr->stochratio);
   fprintf(ofp," beta=%f \n",*kr->beta);
   print_corpars(kr);
   printvecr(*kr->n,kr->y,"y");
   printmatr(*kr->n,*kr->d,kr->X,"X");
   printmatr(*kr->n,*kr->r,kr->F,"F");
   is = kr->init[0];
   fprintf(ofp," d= %d n= %d \n",is.d,is.n);
   fprintf(ofp," X= %d y= %d \n",is.X,is.y);
   fprintf(ofp," R= %d F= %d \n",is.R,is.F);
   fprintf(ofp," beta= %d var= %d \n",is.beta,is.var);
   fprintf(ofp," alpha= %d theta= %d \n",is.alpha,is.theta);
   fprintf(ofp," U= %d range= %d \n",is.U,is.range);
}
/************************************************************************/
/* Version : Nov 30, 1997                                               */
/* Purpose : Free contents of KrigModel 				*/
void free_KrigModel(KrigModel *kr)
{
  int n,d,r;

  n = *kr->n;
  d = *kr->d;
  r = *kr->r;

  frei(d,&kr->theta);
  frei(d,&kr->alpha);
  frei(r,&kr->beta);
  frei(n,&kr->y);
  frei(n+r,&kr->eigval);
  frei(n+r,&kr->eigvalinv);
  frei(n,d,&kr->X);
  frei(n,n,&kr->R);
  frei(n+r,n+r,&kr->U);
  frei(n,r,&kr->F);
  frei(1,&kr->d);
  frei(1,&kr->n);
  frei(1,&kr->r);
  frei(1,&kr->var);
  frei(1,&kr->stochratio);
  frei(1,&kr->loglike);
  frei(d,2,&kr->range);
  frei(d,&kr->xlabel);
}
/************************************************************************/
/* Version : Nov 30, 1997                                               */
/* Purpose : Free contents of KrigArray and KrigArray itself            */
void free_KrigArray(KrigArray *ka, int pc)
{
    if (ka->init)
    {
      frei(pc,3,&ka->ydescrip);
      frei(pc,&(ka->pvec));
      frei(pc,&(ka->cvec));
    }
    free(ka);
}
/************************************************************************/
/* Version : Nov 30, 1997                                               */
/* Purpose : cover function for freeing Krigarray and KrigModel s       */
void free_Krig_cover(KrigArray *ka,KrigModel *kr)
{
  int pc,i,n;
  
  pc = ka->p+ ka->c;
  for (i=0;i< pc; i++)
  {
    free_KrigModel(&(kr[i]));
  }
  free(kr);
  free_KrigArray(ka,pc);

}
 
