/***************************************************************************/
// Copyright (c) 1996 by Matthias Schonlau
/***************************************************************************/
/* general utility procedures */

#include <stdio.h>
#include <malloc.h>
#include <string.h>  // for minimize
#include <stdlib.h>  // for str_to_real


#include "general.h"
#include "alloc.h"
#include "matrix.h"

extern FILE *ofp ;


/**********************************************************************/
/* Version : March 14, 1996 */
void nrerror(char error_text[])
{
  fprintf(stderr,"SPACE ERROR... \n");
  fprintf(stderr,"%s \n",error_text);
  fprintf(stderr,"Now exiting to system ... \n ");
  exit(1);
}
/**********************************************************************/
/* Version : January 05, 1997 						*/
/* Purpose : error message if value< critical value			*/
void nrerror(char *error_text,real value,char *value_text)
{
    fprintf(stderr,"SPACE error ... \n");
    fprintf(stderr,"%s %f \n", value_text, value);
    fprintf(stderr,"%s \n", error_text);
    fprintf(stderr,"Now exiting to system ... \n ");
    exit(2);
}
/**********************************************************************/
/* Version : January 24, 1997 */
void nrinfo(char *s)
{
  fprintf(ofp,"Info : %s\n",s);
}
/**********************************************************************/
/* Version : January 24, 1997 */
void nrwarning(char *s)
{
  fprintf(ofp,"WARNING: %s\n",s);
}
/**********************************************************************/
/* Version : March 14, 1996 */
void printveci(int n ,int *vector, char text[])
{
  int i ;

  fprintf(ofp,"%s ",text);
  for (i=0; i<n; i++)
     fprintf(ofp," %i ",vector[i]);
  fprintf(ofp," \n");
}
/**********************************************************************/
/* Version : Feb 13, 1997 */
void printvecs(int n ,string *vector,char text[])
{
  int i ;
  int items_per_line = 6;

  fprintf(ofp,"%s ",text);
  for (i=0; i<n; i++)
  {
     fprintf(ofp," %10s ",vector[i]);
     if (i%items_per_line ==5) fprintf(ofp,"\n");
  }
  fprintf(ofp,"\n");
}
/**********************************************************************/
/* Version : March 14, 1996 */
void printvecr(int n ,real *vector,char text[])
{
  int i ;
  int items_per_line = 6;

  fprintf(ofp,"%s ",text);
  if (n>items_per_line) 
    fprintf(ofp,"\n");
  for (i=0; i<n; i++)
  {
     fprintf(ofp," %10.5f ",vector[i]);
     if (i%items_per_line ==5) fprintf(ofp,"\n");
  }
  fprintf(ofp,"\n");
}
/**********************************************************************/
/* Version : April 06, 1996 */
void printmatr(int n, int m, real **matrix,char text[],boolean exact)
{
  int i ,j ;
  int items_per_line = 6;

  fprintf(ofp,"%s \n",text);
  for (i=0; i<n; i++)
  {
    for (j=0; j<m; j++)
    {
       if (!exact)
         fprintf(ofp," %10.5f ",matrix[i][j]);
       else
         fprintf(ofp," %10.15f ",matrix[i][j]);
       if (j%items_per_line ==5) fprintf(ofp,"\n");
    }
    if (i!= n-1) fprintf(ofp,", \n");
  }
  fprintf(ofp,"\n");
}
/**********************************************************************/
/* Version : July 12, 1996 */
void printmats(int n, int m, string matrix[][MAXTERMS],char text[])
{
  int i ,j ;

  fprintf(ofp,"%s \n",text);
  for (i=0; i<n; i++)
  {
    for (j=0; j<m; j++)
      fprintf(ofp,"%s ",matrix[i][j]); 
    if (i!= n-1) fprintf(ofp,", \n");
  }
  fprintf(ofp,"\n");
      //  fprintf(ofp," %s ",matrix[i][j]);
}
/**********************************************************************/
/* Version : March 04, 1997 						*/
void printmats(int n, int m, string **matrix,char text[])
{
  int i ,j ;

  fprintf(ofp,"%s \n",text);
  for (i=0; i<n; i++)
  {
    for (j=0; j<m; j++)
      fprintf(ofp,"%s ",matrix[i][j]);
    if (i!= n-1) fprintf(ofp,", \n");
  }
  fprintf(ofp,"\n");
      //  fprintf(ofp," %s ",matrix[i][j]);
}
/**********************************************************************/
/* Version : May 06, 1996 */
void printmati(int n, int m, int **matrix,char text[])
{
  int i ,j ;

  fprintf(ofp,"%s \n",text);
  for (i=0; i<n; i++)
  {
    for (j=0; j<m; j++)
       fprintf(ofp," %d ",matrix[i][j]);
    if (i!= n-1) fprintf(ofp,", \n");
  }
  fprintf(ofp,"\n");
}
/**********************************************************************/
/* Version : June 26, 1996 */
/* Purpose : initialize all elements of vec to x */
void init_real_vec(real x, int n, real *vec)
{
  int i;
  
  if (n==0) nrerror("init_real_vec: Vector has n==0");
  for (i=0; i<n; i++)
  {
    vec[i] = x;
  }
}
/************************************************************************/
/* Version : May 14, 1997 						*/
/* Purpose : initialize all elements of int vec to x 			*/
void init_int_vec(int x, int n, int *vec)
{
  int i;

  if (n==0) nrerror("init_int_vec: Vector has n==0");
  for (i=0; i<n; i++)
  {
    vec[i] = x;
  }
}
/**********************************************************************/
/* Version : June 27, 1996 */
/* Purpose : initialize all elements of m to x */
void init_real_mat(real x, int n , int m , real **mat)
{
  int i,j;

  if (n==0 || m==0) nrerror("init_real_mat: Matrix has n==0 or m==0");
  #ifdef DEBUG_ALLOC
  check_alloc(mat,0,0,"util:init_real_mat");
  check_alloc(mat,n-1,m-1,"util:init_real_mat");
  #endif
  for (i=0; i<n; i++)
    for (j=0; j<m; j++)
    {
       mat[i][j] = x;
    }
}
/**********************************************************************/
/* Version : Jan 28, 1997						*/
/* Purpose : initialize all off-diagonal elements of m to 0, 		*/
/* 	and diagonal elements to x					*/ 
void init_diag_mat(real x, int n, real **mat)
{
   int i,j;

   if (n==0) nrerror("init_diag_mat: Matrix has n==0");   
   #ifdef DEBUG_ALLOC
   check_alloc(mat,0,0,"util:init_diag_mat");
   check_alloc(mat,n-1,n-1,"util:init_diag_mat");
   #endif 
   for (i=0; i<n; i++)
    for (j=0; j<n; j++)
    {
       if (i==j) mat[i][j] = x;
       else      mat[i][j] = 0.0;
    }
}
/************************************************************************/
/* Version : March 29, 1997 						*/
/* Purpose:  check whether two real vectors are identical		*/ 
/* on return : TRUE  if yes, FALSE otherwise				*/
boolean identical_vec(int n,real *vec1,real *vec2)
{
  int i;
  for (i=0;i<n;i++)
     if (vec1[i]!=vec2[i]) return FALSE;
  return TRUE;
}
/**********************************************************************/
/* Version : April 08, 1996 */
 /* check whether square real matrix is symmetric:
    on exit: 1 is not symmetric;  0 is symmetric */
int symmetry_realmat_check(int n, real **mat)
{
   int i,j;
   int temp;

  temp=0;
  for (i=0; i<n; i++)
    for (j=i+1; j<n; j++)
       if (mat[i][j]!=mat[j][i]) temp=1;
  return(temp);
}
/**********************************************************************/
/* Version : June 19, 1996 */
/* copies the real matrix "mat" into matrix "mat_copy" */
void copy_real_matrix(int n, int m, real **mat,real **mat_copy)
{
  int i,j;

  #ifdef DEBUG_ALLOC
  check_alloc(mat,n-1,m-1,"util:copy_real_matrix");
  check_alloc(mat_copy,n-1,m-1,"util:copy_real_matrix 2");
  #endif
  for (i=0; i<n; i++)
    for (j=0; j<m; j++)
    {
      mat_copy[i][j] = mat[i][j];
    }
}
/************************************************************************/
/* Version : Feb 25, 1997 						*/
/* Purpose : copies the real matrix "mat" into matrix "mat_copy" 	*/
/*	     starting from the (offset)th column of mat			*/
/*	     into the (cp_offset)th column of mat_copy			*/ 
/* Note    : e.g. when col 2 required, offset is 1			*/
void copy_matrix(int n, int m, real **mat,int offset,real **mat_copy, int cp_offset)
{
  int i,j;

  if (offset<0 || cp_offset<0) 
    nrerror("copy_matrix with offsets: one of the  offset <0 ");  
  #ifdef DEBUG_ALLOC
  check_alloc(mat,n-1,m+offset-1,"util:copy_real_matrix");
  check_alloc(mat_copy,n-1,m+cp_offset-1,"util:copy_real_matrix 2");
  #endif
  for (i=0; i<n; i++)
    for (j=0; j<m; j++)
    {
      mat_copy[i][j+cp_offset] = mat[i][j+offset];
    }
}
/************************************************************************/
/* Version : Feb 26, 1997                                               */
/* Purpose : copies the real matrix "mat" into STRING matrix "mat_copy" */
/*           starting from the (offset)th column of mat                 */
/*           into the (cp_offset)th column of mat_copy                  */
void copy_matrix(int n, int m, real **mat,int offset,string **mat_copy, 
  int cp_offset)
{
  int i,j;
  string string_elem="";

  if (offset<0 || cp_offset<0)
    nrerror("copy_matrix real to string with offsets: one of the  offset <0 ");
  #ifdef DEBUG_ALLOC
  check_alloc(mat,n-1,m+offset-1,"util:copy_matrix");
  #endif DEBUG_ALLOC
  for (i=0; i<n; i++)
    for (j=0; j<m; j++)
    {
      sprintf(string_elem,"%f",mat[i][j+offset]);
      strcpy(mat_copy[i][j+cp_offset],string_elem);
    }
}
/************************************************************************/
/* Version : May 27, 1996 						*/
/* Purpose : copies real vector vec1 into vector vec2  			*/
/* NOTE    : if p ==0 , vec1 and vec2 stay untouched , no error 	*/
void copy_real_vector(int p, real *vec1,real *vec2)
{
  int i;

  #ifdef DEBUG_ALLOC
  check_alloc(vec1,0,"util:copy_real_vector");
  check_alloc(vec2,0,"util:copy_real_vector");
  check_alloc(vec1,p-1,"util:copy_real_vector");
  check_alloc(vec2,p-1,"util:copy_real_vector");
  #endif DEBUG_ALLOC
  for (i=0; i<p; i++)
    vec2[i] = vec1[i];
}
/************************************************************************/
/* Version : January 03, 1997                                          */
/* Purpose : copies int vector vec1 into vector vec2                   */
/* NOTE    : if p ==0 , vec1 and vec2 stay untouched, no error         */
void copy_int_vector(int p, int *vec1,int *vec2)
{
  int i;

  #ifdef DEBUG_ALLOC
  check_alloc(vec1,0,"util:copy_int_vector");
  check_alloc(vec2,0,"util:copy_int_vector");
  check_alloc(vec1,p-1,"util:copy_int_vector");
  check_alloc(vec2,p-1,"util:copy_int_vector");
  #endif DEBUG_ALLOC
  for (i=0; i<p; i++)
    vec2[i] = vec1[i];
}
/************************************************************************/
/* Version : January 03, 1997 - Nov 05 , 97                             */
/* Purpose : copies int vector orig into vector copy			*/
/*	except the i th element (0,...,p-1)				*/
/* NOTE    : if p ==0 , omit==0 no error, everything untouched         	*/
/* NOTE    :  0<= omit < p must be true otherwise error		*/
void copy_int_vector_minus_one(int p,int *orig,int omit,int *copy)
{
  int i;
   
   if (omit>=p || omit<0) 
      nrerror("copy_int_vector_minus_one:(omit index>vector length)or <0");
  for (i=0; i<p; i++)
  {
     if (i<omit)
     {
       copy[i] = orig[i]; 
       #ifdef DEBUG_ALLOC
       check_alloc(orig,i,"util:copy_int_vector_minus_one");
       check_alloc(copy,i,"util:copy_int_vector_minus_one");
       #endif DEBUG_ALLOC
     }
     else if (i>omit)
     {
       copy[i-1] = orig[i];
       #ifdef DEBUG_ALLOC
       check_alloc(orig,i,"util:copy_int_vector_minus_one");
       check_alloc(copy,i-1,"util:copy_int_vector_minus_one");
       #endif DEBUG_ALLOC
     }
     // if i==omit do nothing
  }
}
/************************************************************************/
/* Version : Nov 05 , 97                             			*/
/* Purpose : copies real vector orig into vector copy                   */
/*      except the i th element (0,...,p-1)                             */
/* NOTE    : if p ==0 , omit==0 no error, everything untouched          */
/* NOTE    :  0<= omit < p must be true otherwise error                 */
void copy_real_vector_minus_one(int p,real *orig,int omit,real *copy)
{
  int i;
  
  if (omit>=p || omit<0)
    nrerror("copy_real_vector_minus_one:(omit index>vector length)or <0");
  for (i=0; i<p; i++)
  {
     if (i<omit)
       copy[i] = orig[i];
     else if (i>omit)
       copy[i-1] = orig[i];
     // if i==omit do nothing
  }
}
/************************************************************************/
/* Version : May 22, 98							*/
/* Purpose : copies real matrix "orig" into real matrix "copy"          */
/*      except the i th row     (0,...,n-1)                             */
/*    orig n,p matrix 							*/
/*    copy n,p matrix 							*/
/* Note    : cannot check that indeed both matrices have the same dims  */
/* Note    : if n ==0 , omit==0 no error, everything untouched          */
/* Note    :  0<= omit <= n must be true otherwise error                */
/* NOTE    : the first row is indexed 0 not 1 				*/
void copy_real_matrix_minus_1row(int n,int p,real **orig,int omit,real **copy)
{
  int i,j;
  
  if (omit>=n || omit<0)
      nrerror("copy_real_matrix_minus_1row:(omit index>vector length)or <0");
  for (i=0; i<n; i++)
  {
    if (i<omit)
       for (j=0; j<p; j++)
       { /* copy i^th row */ 
         copy[i][j] = orig[i][j];
       }
    else if (i>omit)
       for (j=0; j<p; j++)
       { /* copy i^th row into (i-1)th row  */
          copy[i-1][j] = orig[i][j];
       }
     // if i==omit do nothing
  }
}
/************************************************************************/
/* Version : January 03, 1997                                           */
/* Purpose : Copy n characters from orig nach copy. orig and copy may   */
/*	have length greater than n					*/
void copy_string(int n,char *orig,char *copy)
{
    int i;
   
    for (i=0;i<n;i++)
       copy[i] = orig[i]; 
}
/************************************************************************/
/* Version : Feb 10, 1997                                           	*/
/* Purpose : Copy n strings from orig nach copy. orig and copy may   	*/
/*      have length greater than n                                      */
void copy_string(int n,string *orig,string *copy)
{
    int i;

    for (i=0;i<n;i++)
       strcpy(copy[i],orig[i]);
}
/************************************************************************/
/* Version : Feb 13, 1997                                               */
/*	modified Jan 20, 2000, Brian's suggestion 			*/
/* Purpose : Copy n strings from orig to copy. orig and copy may        */
/*      have length greater than n                                      */
/* return : 0 ok , 1 error line  isn't long enough to hold the n 	*/
/*	concatenated strings 						*/
/* NOTE : change 3 , if more or less space given			*/ 
//int concat_string(int n,string *vec,int maxline,char *line)
int concat_string(int n,char **vec,int maxline,char *line)
{
  int i;
    strcpy(line,"");
    for (i=0;i<n;i++)
    {
       if (int(strlen(line))+int(strlen(vec[i]))+5 > (maxline-1)) 
         return 1;
       strcat(line,vec[i]);
       strcat(line,"   ");
    }
    //fprintf(ofp,"len %d \n",int(strlen(line)));
  return 0;
}
/*************************************************************************/
/* Version : April 25, 1996 */
/* Procedure calculates   veca = s + veca */ 
/* result is returned in a  */
void VecAddScalarReal(real s, int n , real *a)
{
  int i ;

  #ifdef DEBUG_ALLOC
  check_alloc(a,0,"util:VecAddScalarReal");
  check_alloc(a,n-1,"util:VecAddScalarReal");
  #endif DEBUG_ALLOC
  for (i=0; i<n ; i++)
     a[i] += s;
}
/******************************************************************/
/* Version : May 15, 1996 */
/* Procedure calculates   vecc = veca + s* vecb */ 
/* result is returned in the function    */
void VecAddVecNewReal(real s, real *a, int n, real *b, real *c)
{
  int i;
   
  #ifdef DEBUG_ALLOC
  check_alloc(a,0,"util:VecAddVecNewReal");
  check_alloc(a,n-1,"util:VecAddVecNewReal");
  #endif DEBUG_ALLOC
  for (i=0; i<n ; i++)
     c[i] = a[i] + s * b[i];
}
/******************************************************************/
/* Version : April 25, 1996 */
/* Procedure calculates   veca = veca *s  */
/* result is returned in a  */
void VecMultScalarReal(real s, int n , real *a)
{
   int i ;
   #ifdef DEBUG_ALLOC
   check_alloc(a,0,"util:VecMultScalarReal");
   check_alloc(a,n-1,"util:VecMultScalarReal");
   #endif DEBUG_ALLOC
   for (i=0; i<n ; i++)
     a[i] *= s;
}
/******************************************************************/
/* Version : April 15, 1996 */
/* Procedure calculates   veca = veca *s + vecb */
/* result is returned in b  */
void VecAddVecReal(real s, real *a, int n, real *b)
{
  int i ;

  #ifdef DEBUG_ALLOC
  check_alloc(a,0,"util:VecAddVecReal");
  check_alloc(b,0,"util:VecAddVecReal");
  check_alloc(a,n-1,"util:VecAddVecReal");
  check_alloc(b,n-1,"util:VecAddVecReal");
  #endif DEBUG_ALLOC
  for (i=0; i<n ; i++)
     b[i] += s * a[i]  ;
}
/******************************************************************/
/* Version : May 03, 1996 */
void VecEwMultVecReal(real *a, int n, real *b)
/* Procedure calculates   veca = veca * vecb  (elementwise multiplication*/
/* result is returned in b  */
{
   int i ;

   #ifdef DEBUG_ALLOC
   check_alloc(a,0,"util:VecEwMultVecReal");
   check_alloc(b,0,"util:VecEwMultVecReal");
   check_alloc(a,n-1,"util:VecEwMultVecReal");
   check_alloc(b,n-1,"util:VecEwMultVecReal");
   #endif DEBUG_ALLOC
   for (i=0; i<n ; i++)
     b[i] *= a[i];
}
/******************************************************************/
/* Version : May 15, 1996 */
void VecEwMultVecNewReal(real *a, int n, real *b, real *c)
/* Procedure calculates   vecc = veca * vecb  (elementwise multiplication*/
/* result is returned in c  */
{
  int i ;

  #ifdef DEBUG_ALLOC
  check_alloc(a,0,"util:VecEwMultVecNewReal");
  check_alloc(b,0,"util:VecEwMultVecNewReal");
  check_alloc(c,0,"util:VecEwMultVecNewReal");
  check_alloc(a,n-1,"util:VecEwMultVecNewReal");
  check_alloc(b,n-1,"util:VecEwMultVecNewReal");
  check_alloc(c,n-1,"util:VecEwMultVecNewReal");
  #endif DEBUG_ALLOC
  for (i=0; i<n ; i++)
     c[i] = a[i] * b[i];
}
/******************************************************************/
/* Version : May 16, 1996 */
/* Purpose : compute s = vec[0]* ... *vec[n-1]
   on return : svec[0]* ... *vec[n-1]  
*/
real ProdVecReal(int n, real *vec)
{
  real s;
  int i ;
  
  s=1.0;
  #ifdef DEBUG_ALLOC
  check_alloc(vec,0,"util:ProdVecReal");
  check_alloc(vec,n-1,"util:ProdVecReal");
  #endif DEBUG_ALLOC
  for (i=0; i<n ; i++)
     s *= vec[i]; 
  return s;
}
/******************************************************************/
/* Version : May 01, 1996 */
/* on input : mat :  n,m matrix   */
/* on output: tmat :  m,n matrix, transpose of mat   */
void TransRealMat(int n, int m, real **mat, real **tmat)
{
   int i,j;

  #ifdef DEBUG_ALLOC
  check_alloc(mat,0,0,"util:TransRealMat");
  check_alloc(tmat,0,0,"util:TransRealMat");
  check_alloc(mat,n-1,m-1,"util:TransRealMat");
  check_alloc(tmat,m-1,n-1,"util:TransRealMat");
  #endif DEBUG_ALLOC
  for (i=0; i<n ; i++)
     for (j=0; j<m ; j++)
     {
       tmat[j][i] = mat[i][j];
     } 
}
/******************************************************************/
/* Version : May 01, 1996 */
/* on input : M   (n,m) matrix 
   on output: M = s*M    
*/ 
void MatMultScalarReal(real s,int n,int m,real **M)
{
   int i,j ;

  #ifdef DEBUG_ALLOC
  check_alloc(M,0,0,"util:MatMultScalarReal");
  check_alloc(M,n-1,m-1,"util:MatMultScalarReal");
  #endif DEBUG_ALLOC
  for (i=0; i<n; i++)
     for (j=0; j<m; j++)
       M[i][j] *= s;
}
/******************************************************************/
/* Version : May 01, 1996 */
/* proc calculates M1 + M2 and returns it into M2 */
/* both M1 and M2 are n,m matrices */
void MatAddMatReal(int n, int m ,real **M1, real **M2)
{
   int i,j;

  #ifdef DEBUG_ALLOC
  check_alloc(M1,0,0,"util:MatAddMatReal");
  check_alloc(M2,0,0,"util:MatAddMatReal");
  check_alloc(M1,n-1,m-1,"util:MatAddMatReal");
  check_alloc(M2,n-1,m-1,"util:MatAddMatReal");
  #endif DEBUG_ALLOC
  for (i=0; i<n ; i++)
     for (j=0; j<m ; j++)
        M2[i][j] += M1[i][j];
}
/******************************************************************/
/* Version : May 01, 1996 */
/* M1 : q,r matrix , M2 : r,s matrix */
/* on input : M1, M2  */
/* on output: M = M1 * M2   */
void MatMultMatReal(int q,real **M1, int r,real **M2, int s, real **M)
{
   int i,j,k;

  for (i=0;i<q;i++)
     for (j=0;j<s;j++) {
       M[i][j] = 0.0;
       for (k=0;k<r;k++)
       {
         #ifdef DEBUG_ALLOC
         check_alloc(M,i,j,"util:MatMultMatReal");
         check_alloc(M1,i,k,"util:MatMultMatReal");
         check_alloc(M2,k,j,"util:MatMultMatReal");
         #endif DEBUG_ALLOC
         M[i][j] += M1[i][k] * M2[k][j];
       }
     }
}
/******************************************************************/
/* Version : May 03, 1996 */
/* vec 1,n  vector, mat : n,m matrix */
/* vec2 :  1,m vector */
/* on input : vec, mat  */
/* on output: vec2 = vec * mat  */
void VecMultMatReal(real *vec, int n, real **mat, int m, real *vec2)
{
   int i,j;

   for (i=0;i<m;i++)
   { 
     vec2[i] = 0.0;
     for (j=0;j<n;j++) 
     {
        #ifdef DEBUG_ALLOC
        check_alloc(mat,j,i,"util:VecMultMatReal");
        check_alloc(vec2,i,"util:VecMultMatReal"); 
        check_alloc(vec,j,"util:VecMultMatReal"); 
        #endif DEBUG_ALLOC
        vec2[i] += vec[j] * mat[j][i];      

     }
   }
}
/******************************************************************/
/* Version : May 15, 1996 */
/* mat : m,n matrix , vec n,1  vector  */
/* vec2 :  m,1 vector */
/* on input : mat, vec  */
/* on output: vec2 =  mat * vec  */
void MatMultVecReal(int m, real **mat, int n, real *vec, real *vec2)
{
   int i,j;

   for (i=0;i<m;i++)
   {
     vec2[i] = 0.0;
     for (j=0;j<n;j++)
     {
        #ifdef DEBUG_ALLOC
        check_alloc(mat,i,j,"util:MatMultVecReal");
        check_alloc(vec2,i,"util:MatMultVecReal");
        check_alloc(vec,j,"util:MatMultVecReal");
        #endif DEBUG_ALLOC
        vec2[i] += vec[j] * mat[i][j];
     }
   }
}
/******************************************************************/
/* Version : May 15, 1996 */
/* mat : n,m matrix , vec m  vector  */
/* mat2 : n,m matrix */
/* on input : mat, vec  */
/* on output: mat2 = mat * D  
              where D is a diagonal matrix whose elements are given in vec */
void MatMultDiagMatReal(int n,real **mat,int m, real *vec, real **mat2)
{
  int i,j; 

  #ifdef DEBUG_ALLOC
  check_alloc(mat,0,0,"util:MatMultDiagMatReal");
  check_alloc(mat2,0,0,"util:MatMultDiagMatReal");
  check_alloc(vec,0,"util:MatMultDiagMatReal");
  check_alloc(mat,n-1,m-1,"util:MatMultDiagMatReal");
  check_alloc(mat2,n-1,m-1,"util:MatMultDiagMatReal");
  check_alloc(vec,m-1,"util:MatMultDiagMatReal");
  #endif DEBUG_ALLOC
  for (i=0; i<n; i++)
    for (j=0; j<m; j++)
      mat2[i][j] = mat[i][j] * vec[j]; 

}
/************************************************************************/
/* Version : Nov 18, 1997 						*/
/* Purpose : find a string in a vector of strings			*/
/* output : index (1..len) when found, -1 otherwise			*/
/* on return :       1    if not found					*/  
int matchvecs(string word,int length,string *vec,int *index)
{
  int i;
   
  for (i=0;i<length;i++)
  {
    if (strcmp(word,vec[i])==0) 
       {*index= i+1; return 0;}
  }
  *index = -1;
  return 1;
}
/******************************************************************/
/* Version : May 03, 1996 */
/* vec1 : n vector , vec2 : n vector*/
/* on input : vec1, vec2  */
/* on output: return s  = vec1 * vec2   */
real VecMultVecReal(real *vec1,int n,real *vec2)
{
   int i;
   real s;

  #ifdef DEBUG_ALLOC
  check_alloc(vec1,0,"util:VecMultVecReal");
  check_alloc(vec1,n-1,"util:VecMultVecReal");
  check_alloc(vec2,0,"util:VecMultVecReal");
  check_alloc(vec2,n-1,"util:VecMultVecReal");
  #endif DEBUG_ALLOC
  s = 0.0;
  for (i=0;i<n;i++)
       s += vec1[i] * vec2[i];
  return s;
}
/******************************************************************/
/* Version : July 03, 1996 */
/* M1 : q,r matrix , M2 : r,s matrix, M3: s,t matrix */
/* on input : M1, M2 , M3 */
/* on output: M = M1 * M2  * M3 */

void Mat3MultReal(int q,real **M1, int r,real **M2, int s,real **M3, int t,
  real **M)
{
   real **tempM=NULL;

   alloc(q,s,&tempM);

   /* multiply tempM = M1 * M2 */
   MatMultMatReal(q,M1,r,M2,s,tempM);
  
   /* multiply M = tempM * M3 */
   MatMultMatReal(q,tempM,s,M3,t,M);

   frei(q,s,&tempM);
}
/******************************************************************/
/* Version : July 03, 1996 */
/* multiplies   vec1 * mat * vec2
   vec1 : 1,p1 vector
   mat  : p1,p2 matrix 
   vec2 : p2,1 vector
*/
real quadraticform(real *vec1, int p1,  real **mat, int p2, real *vec2)
{
   real *tempvec=NULL;
   int i,j;
   real result;

   alloc(p2,&tempvec); 
   for (i=0;i<p2; i++)
     for (j=0, tempvec[i]=0.0; j<p1; j++)
     {
       #ifdef DEBUG_ALLOC
       check_alloc(tempvec,i,"util:quadraticform");
       check_alloc(vec1,j,"util:quadraticform");
       check_alloc(mat,j,i,"util:quadraticform");
       #endif DEBUG_ALLOC
       tempvec[i] += vec1[j] * mat[j][i];
     }
   for (i=0, result=0;i<p2; i++)
   {
     result += tempvec[i] * vec2[i];
   }
   frei(p2,&tempvec);
   return result;
}
/************************************************************************/
/* Version : June 18, 1996 						*/
/* Purpose : returns the minimum value of a real vector 		*/
/* Note    : if n too large this could return 0, vec filled up with 0's */ 
real Real_Minimum(int n,real *vec)
{
  int i ;
  real tempmin;

  if (n==0) nrerror("Real_Minimum: Minimum of empty set of numbers");
  tempmin = vec[0];
  /* we start nonetheless  the loop at i=0, so that hte program doesn't 
     crash in case n =1 */
  #ifdef DEBUG_ALLOC
  check_alloc(vec,0,"util:Real_Minimum");
  check_alloc(vec,n-1,"util:Real_Minimum");
  #endif DEBUG_ALLOC
  for (i=0; i<n; i++)
    if (vec[i] < tempmin ) 
       tempmin = vec[i];
  return tempmin;
}
/************************************************************************/
/* Version : June 18, 1996 - Nov 11,97                                  */
/* Purpose : returns the minimum value of a real vector                 */
/* on output : ymin (returned)						*/
/*	       *index  contains position of ymin (1..n)			*/
/* Note    : if n too large this could return 0, vec filled up with 0's */
real RealMinimumIndex(int n,real *vec, int *index)
{
  int i ;
  real tempmin;

  if (n==0) nrerror("Real_Minimum: Minimum of empty set of numbers");
  tempmin = vec[0];
  *index=1;
  /* we start nonetheless  the loop at i=0, so that the program doesn't
     crash in case n =1 */
  #ifdef DEBUG_ALLOC
  check_alloc(vec,0,"util:Real_Minimum");
  check_alloc(vec,n-1,"util:Real_Minimum");
  #endif DEBUG_ALLOC
  for (i=0; i<n; i++) 
  {
    if (vec[i] < tempmin) 
    {
       tempmin = vec[i];
       index[0] = i+1;
    }
  }
  return tempmin;
}
/************************************************************************/
/* Version : June 18, 1996                                              */
/* Purpose : Generates the next permutation in lexicographical order of */
/* 	     n nonnegative integers. 					*/
/* input   : int  n     size of whole permutation 			*/
/*  	     int *perm  previous permutation string	 		*/
/* output  : perm	next permutation string 			*/
/* return  : 0 if no more permutations remain and 1 otherwise		*/
int PermLex(int  n, int *perm)
{
  int i,j, temp;

  if (n<=1) return (All_Done); /*All_Done */

  i = n-1;
  while (i>0 && perm[i-1] >= perm[i] )
        i--;
  if (i==0)
        return (All_Done); /*All_Done */

  i--;

  j= n-1;
  while (perm[j] <= perm[i])
    j--;

  /* interchange perm[i] and perm[j] */
  temp = perm[i];
  perm[i] = perm[j] ;
  perm[j] = temp;

  if (i < n-2)
     for (i++, j=n-1; i<j; i++, j--)
     {
        temp= perm[i];
        perm[i] = perm[j];
        perm[j] = temp;
     }
  return OK; /* OK */
}
/************************************************************************/
/* Version : June 24, 1996 - Feb 05 , 97				*/
/* Purpose : allocate and create a grid of n_levels points in d dimensions*/ 
/* input :  n_levels  grid_size						*/
/*          d	      d, number of dimensions 				*/
/*  	    m	      range.mat						*/
/* Note : allocates xpred but doesn't free it # frei(npred,d,&xpred);   */
/* output:  xpred     *npred,d matrix , contains grid			*/
/*          *npred    pointer to scalar npred, needed to free xpred later*/
/* Note : xpred on entry should be NULL , may be wanna check that       */
void grid(int n_levels, int d,real **m,real ***xpred, int *npred)
{
  int i,j,  *levels, *gridelem;
  real **my_xpred=NULL;

  alloc(d,&levels); alloc(d,&gridelem);
  my_xpred= *xpred;

  for (j=0; j<d; j++)
  { // needed to start PermGrid
    gridelem[j]=0;
    levels[j]= n_levels;
  }
  if (n_levels<=1) nrerror("grid(): Code Bug : <=1 on grid");
  for (i=0,*npred=1; i<d; i++)
     *npred *= n_levels;  // *npred=n_levels^d

  alloc(*npred,d,&my_xpred);
  i=0;
  do 
  { // loop runs *npred times
    //printveci(d,gridelem,"cont ");
    for (j=0;j<d;j++)
      my_xpred[i][j] = m[j][0] + (real(gridelem[j])/(real(levels[j]-1))) * 
                    (m[j][1]-m[j][0]);
    i++; //next grid point
  } while (PermGrid(d,levels,gridelem));

  *xpred = my_xpred;
  frei(d,&levels); frei(d,&gridelem);
}
/************************************************************************/
/* Purpose : Generate a sequence of vectors that fill up a grid 	*/
/* on input :  d : number of dimensions					*/
/* 	       *gridelem: d vector: last point on the grid generated	*/ 
/*             *levels  : d vector: number of levels in each dimensions */
/* on output:  gridelem : next point on the grid generated		*/
/* on return:  Returns 0 if no more points remain and 1 otherwise  	*/
/* NOTE: i levels means , gridelems go from 0..(i-1)			*/
/* NOTE: the very first elem is  (0,0,...,0)				*/
/* NOTE: no checking of levels or gridelem is done 			*/ 
int PermGrid(int d,int *levels, int *gridelem)
{
 int i;
 boolean next_digit=TRUE;

 if (d<=0) nrerror("PermGrid: dimension is zero or negative.");
 i=d-1;
 while (next_digit && i>=0)
 {
    if (gridelem[i]==(levels[i]-1))
    {
       gridelem[i]=0;
       next_digit=TRUE;
    }
    else
    {
       gridelem[i]+=1;
       next_digit=FALSE;
    }
    i--;
 }
 if (next_digit) return 0; // all_done
 else return 1;   // more points to come
}
/************************************************************************/
/* Version : August 10, 1996                                            */
/* Purpose : exchange two rows of a matrix                              */
/* on input : n:     length of row                                      */
/*            matrix: real matrix                                       */
/*            row1,row2 : rows to be exchanged                          */
/* on output : matrix has exchanged rows                                */
/* NOTE :  row1, row2 range from 0..N-1 , and NOT from 1..N		*/
void exchange(int n,real **matrix,int row1,int row2)
{
  int i;
  real temp;

  for (i=0;i<n;i++)
  {
    #ifdef DEBUG_ALLOC
    check_alloc(matrix,row2,i,"util:exchange");
    check_alloc(matrix,row1,i,"util:exchange");
    #endif DEBUG_ALLOC
    temp = matrix[row2][i];
    matrix[row2][i] = matrix[row1][i];
    matrix[row1][i] = temp;
  }
}
/************************************************************************/
/* Version : November 15, 1996                                          */
/* Purpose : calc. mathematical operation n!				*/
/* on return : n! 							*/
int factorial(int n)
{
   int result,i;

   if (n<0) nrerror("factorial: (n<0) or (n is not an integer)");
   else if (n>30)
     nrerror("factorial: n! where n>30, number too large");
   for  (i=2,result=1;i<=n;i++)
   {
     result *= i;
   }
   return result;
}
/************************************************************************/
/* Version : November 15, 1996                                          */
/* Purpose : calc. mathematical operation (n choose i)			*/
/* on return : (n choose i)						*/
int choose(int n,int i) 
{
  if (i>n) nrerror("choose: i>n , mathematical nonsense\n");
  if (n<=10)
    return (factorial(n) / (factorial(i)*factorial(n-i)));  
  else 
    nrerror("choose: rewrite procedure, can't handle n>10");
  return -1 ; // can't happen since always exit earlier 
}
/************************************************************************/
/* Version : November 24, 1996                                          */
/* Purpose : return the greater of two  numbers				*/
real maximum_two(real a, real b)
{
	if (a>b) return a;
	else return b;
}
/************************************************************************/
/* Version : March 07, 1997                                             */
/* Purpose : calc. the min. distance in direction of sm_dist(+or-)      */
/*  	to any point in all_pts for each dimension 			*/
/*      lower and upper ranges act as lower and upper limits		*/
/* NOTE : pt contained in all_pts more than once , then distance ==0    */ 
/* input : 								*/
/*  	pt	 	(d) vec with one point				*/
/*	all_pts		(n,d) matrix of all points (incl. point [once])	*/ 
/*	range 		(d,2) matrix of lower and upper) 		*/ 
/*	direction	(2) vec containing -1,+1 's			*/ 
/*      n, d								*/
/* output :								*/
/* 	sm_dist		(d) vec with smallest distances in each dim	*/
void min_distance_each_dim(int n,int d,real *pt,real **all_pts,real **range,int *direction,real *sm_dist)
{
  int i,j; 
  real temp;
  int pt_index= -1;  // elem number of pt within all_pts (0..n-1) 
  boolean btemp;

  j=0;
  do
  { // find index of pt in all_pts
    btemp=TRUE;
    for (i=0;i<d;i++)
      if (pt[i]!=all_pts[j][i])
         btemp=FALSE;
    j++;
  } while (j<n && btemp==FALSE); 
  if (btemp==TRUE) pt_index= j-1;
  else nrinfo("min_distance_each_dim: No point contained");

  for (i=0;i<d;i++) 
  { // for each dimension
    if (direction[i]==1) temp = range[i][1]-pt[i]; 
    else if (direction[i]== -1) temp = pt[i]-range[i][0]; 
    else nrerror("min_distance_each_dim: Code Bug , direction!= -1,1");
    if (temp<0.0) nrerror("min_distance_each_dim: Code Bug, temp<0");
    for (j=0;j<n;j++)
    { // for each point
       if (direction[i]==1 && all_pts[j][i]-pt[i]>=0 && pt_index!=j) 
       {
         if (all_pts[j][i]-pt[i]<temp)
           temp=all_pts[j][i]-pt[i];
       }
       else if (direction[i]==(-1) && all_pts[j][i]-pt[i]<=0 && pt_index!=j)
       {
         if (pt[i]- all_pts[j][i]<temp)
           temp=pt[i]-all_pts[j][i];
       }
    }
    sm_dist[i]=temp;
  }
}
/************************************************************************/
/* Version : March 20, 1997						*/
/* Purpose : check whether a string is a number and converti            */
/* input   : str  string    containing number                           */
/* output  : val  number (if no error)                                  */
/* return  : 0 ok, otherwise error                                      */
/* NOTE    : ' ','\t','\n' in a string will not cause an error          */
/************************************************************************/
int str_to_real( char str[] , real *val)
{
        char* end;
        int i;

        *val = strtod( str, &end );
        if ( end != str + strlen(str) )
        {
                if ( end == str )
                { // failed at beginning (empty string)
                   return 1;
                }
                else if ( *end == ' ' )
                { // There is one space after the number
                  while (*end==' ' || *end=='\n' || *end=='\t')
                  {
                    end++;
                    fprintf(ofp,"*%c*",*end);
                  }
                  if (*end!='\0')
                    return 2;
                  else
                    return 0; // there is only space after the number
                }
                else
                { // failed at end
                   return 3;
                }
        }
        return 0;
}
/************************************************************************/
/* Version : March 20, 1997                                             */
/* Purpose : check whether a string is a number and convert             */
/*           missing values indicated by NA or na are allowed and are 	*/
/* 	     assigned  na_value						*/
/* input   : str    string to be converted 				*/
/* output  : val    converted string , with na's appearing as na_value  */
/* return  : 0 if ok , !=0 otherwise					*/
int na_str_to_real( char str[], real *val,real na_value)
{
       if (str_to_real(str,val)!=0)
       {
         if (strcmp(str,"NA")==0 || strcmp(str,"na")==0)
          *val = na_value;
         else
           return 1;
       }
       return 0;
}
/************************************************************************/
/*
int main()
{
  int index ;
  string vec[3]={"one","two","three"};
  real **m, **m2;
  int i,j;

  alloc(3,4,&m);
  alloc(2,2,&m2);
  for  (i=0;i<3 ;i++)
   for  (j=0;j<2 ;j++)
   {
     m[i][j] = (real) i  + j;
     if (i<2) m2[i][j] = 0.0;
   }

   printmatr(3,4,m,"m");
   printmatr(2,2,m2,"m2");
   copy_real_matrix_minus_1row(3,4,m,1,m2);
   printmatr(3,4,m,"m");
   printmatr(2,2,m2,"m2");

  
return 0;

}
*/
