/***************************************************************************/
// Copyright (c) 1996 by Matthias Schonlau
/***************************************************************************/

#include "general.h"

#ifndef _MATRIX_
#define _MATRIX_

/* all functions from util.h */

/*****************************/
/* ... miscellaneous functions ...*/
/*****************************/
int symmetry_realmat_check(int n, real **mat);
boolean identical_vec(int n,real *vec1,real *vec2);
real Real_Minimum(int n,real *vec);
real RealMinimumIndex(int n,real *vec, int *index);
void print_rnow();
int PermLex(int n, int *perm);
int PermGrid(int d,int *levels, int *gridelem);
void grid(int n_levels, int d,real **m,real ***xpred,int *npred);
void exchange(int n,real **matrix,int row1,int row2);

/*****************************/
/* ... initialize functions ... */
/*****************************/
void init_real_vec(real x, int n, real *vec);
void init_int_vec(int x, int n, int *vec);
void init_real_mat(real x, int n , int m , real **mat);
void init_diag_mat(real x, int n , real **mat);

/*****************************/
/* ... copy + concatenate + matching functions ... */
/*****************************/
void copy_real_matrix(int n, int m, real **mat,real **mat_copy);
void copy_real_vector(int p, real *vec1,real *vec2);
void copy_int_vector(int p, int *vec1,int *vec2);
void copy_string(int n,char *orig,char *copy);
void copy_string(int n,string *orig,string *copy);
void copy_int_vector_minus_one(int p,int *orig,int omit,int *copy);
void copy_real_vector_minus_one(int p,real *orig,int omit,real *copy);
void copy_real_matrix_minus_1row(int n, int p,real **orig,int omit,real **copy);
int concat_string(int n,char **vec,int maxline,char *line);
void copy_matrix(int n, int m, real **mat,int offset,real **mat_copy, int cp_offset);
void copy_matrix(int n, int m, real **mat,int offset,string **mat_copy,
  int cp_offset);
int matchvecs(string word,int length,string *vec,int *index);

/*****************************/
/* ... print functions ... */
/*****************************/
void printveci(int n ,int *vector, char text[]="");
void printvecr(int n ,real *vector,char text[]="");
//doesn't work void printvecr(FILE *afp,int n ,real *vector,char text[]="");
void printvecs(int n ,string *vector,char text[]);
void printmatr(int n, int m, real **matrix,char text[]="",boolean exact=FALSE);
void printmati(int n, int m, int **matrix,char text[]="");
void printmats(int n, int m, string matrix[][MAXTERMS],char text[]);
void printmats(int n, int m, string **matrix,char text[]);

/*****************************/
/* ...vector algebra functions... */
/*****************************/
void VecEwMultVecReal(real *a, int n, real *b);
void VecEwMultVecNewReal(real *a, int n, real *b, real *c);
void VecAddVecReal(real s, real *a, int n, real *b);
void VecAddVecNewReal(real s,  real *a, int n, real *b , real *c);
void VecMultScalarReal(real s, int n , real *a);
void VecAddScalarReal(real s, int n , real *a);
real VecMultVecReal(real *vec1,int n,real *vec2);
real ProdVecReal(int n, real *vec);

/*****************************/
/* ...mixed matrix-vector algebra functions... */
/*****************************/
void VecMultMatReal(real *vec, int n, real **mat, int m, real *vec2);
void MatMultVecReal(int m, real **mat, int n, real *vec, real *vec2);
real quadraticform(real *vec1, int p1,  real **mat, int p2, real *vec2);

/*****************************/
/* ...matrix algebra functions... */
/*****************************/
void TransRealMat(int n, int m, real **mat, real **tmat);
void Mat3MultReal(int q,real **M1, int r,real **M2, int s,real **M3, int t,
  real **M);
void MatMultMatReal(int q,real **M1, int r,real **M2, int s, real **M);
void MatMultDiagMatReal(int n,real **mat,int m, real *vec, real **mat2);
void MatAddMatReal(int n, int m ,real **M1, real **M2);
void MatMultScalarReal(real s,int n,int m,real **M);

#endif
