#include <stdio.h>
#include <math.h>

#include "general.h"
#include "alloc.h"
#include "matrix.h" /*for test runs in main*/

#define DEBUG 0   // prints x and y values at each iteration

/************************************************************************/
/*NOTE : this may cause a floating error , for example in mnbrak when 	*/
/* the  function has a minimum at -infinity 				*/
/************************************************************************/


/************************************************************************/
/* Version : November 24, 1996                                          */
/* Purpose : Isolates the minimum to a relative tolerance of about "tol"*/
/* 	     using Brent's method					*/ 
/* input   : 								*/
/*	ax,bx,cx 	bracketing triplet with (ax < bx < cx) and	*/
/*			( f(bx) < f(ax)) and (( f(bx) < f(cx))		*/
/* output  : 								*/
/*	brent	minimal function value found				*/
/* 	xmin 	x value corresponding to brent  			*/

#include <math.h>
#define NRANSI
#define ITMAX 100
#define CGOLD 0.3819660
#define ZEPS 1.0e-10
#define SHFT(a,b,c,d) (a)=(b);(b)=(c);(c)=(d);

extern FILE *ofp;

real brent(real ax, real bx, real cx, real (*f)(real), real tol,
	real *xmin)
{
	int iter;
	real a,b,d,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm;
	real e=0.0;

	a=(ax < cx ? ax : cx);
	b=(ax > cx ? ax : cx);
	x=w=v=bx;
	fw=fv=fx=(*f)(x);
	for (iter=0;iter<ITMAX;iter++) {
		xm=0.5*(a+b);
		tol2=2.0*(tol1=tol*fabs(x)+ZEPS);
		if (fabs(x-xm) <= (tol2-0.5*(b-a))) {
			*xmin=x;
			return fx;
		}
		if (fabs(e) > tol1) {
			r=(x-w)*(fx-fv);
			q=(x-v)*(fx-fw);
			p=(x-v)*q-(x-w)*r;
			q=2.0*(q-r);
			if (q > 0.0) p = -p;
			q=fabs(q);
			etemp=e;
			e=d;
			if (fabs(p) >= fabs(0.5*q*etemp) || p <= q*(a-x) || p >= q*(b-x))
				d=CGOLD*(e=(x >= xm ? a-x : b-x));
			else {
				d=p/q;
				u=x+d;
				if (u-a < tol2 || b-u < tol2)
					d=SIGN(tol1,xm-x);
			}
		} else {
			d=CGOLD*(e=(x >= xm ? a-x : b-x));
		}
		u=(fabs(d) >= tol1 ? x+d : x+SIGN(tol1,d));
		fu=(*f)(u);
		if (fu <= fx) {
			if (u >= x) a=x; else b=x;
			SHFT(v,w,x,u)
			SHFT(fv,fw,fx,fu)
		} else {
			if (u < x) a=u; else b=u;
			if (fu <= fw || w == x) {
				v=w;
				w=u;
				fv=fw;
				fw=fu;
			} else if (fu <= fv || v == x || v == w) {
				v=u;
				fv=fu;
			}
		}
	}
	nrerror("Powell : Too many iterations in brent");
	*xmin=x;
	return fx;
}
#undef ITMAX
#undef CGOLD
#undef ZEPS
#undef SHFT
#undef NRANSI

/************************************************************************/
/* Version : November 24, 1996                                          */
/* Purpose : artificial function created by linmin to be fed to 1 dimen-*/
/*	sional minimization. Function represents a 1-dimensional line   */
/*	of the n dimensional function of linmin				*/
#define NRANSI

extern int ncom;
extern real *pcom,*xicom,(*nrfunc)(real [], int nDims);

real f1dim(real x)
{
	int j;
	real f,*xt=NULL;

	alloc(ncom,&xt);
	for (j=0;j<ncom;j++) 
            xt[j]=pcom[j]+x*xicom[j];
	f=(*nrfunc)(xt,ncom);
	frei(ncom,&xt);
	return f;
}
#undef NRANSI


/************************************************************************/
/* Version : November 24, 1996                                          */
/* Purpose : Find the minimum point on the line defined by the point p  */
/*	     and the direction xi					*/
/* input   : 	p	n dimensional point				*/
/* 		xi	n dimensional direction 			*/
/* 		*func	Minimization function				*/
/* return  :	p	reset to where func(p) takes a minimum along the*/
/*			direction xi from p				*/
/* 		fret	func(p)						*/
/* aux funcs:	brent, mnbrak , f1dim					*/


#define NRANSI
#define TOL 2.0e-4

int ncom;
real *pcom,*xicom,(*nrfunc)(real [],int nDims);

void linmin(real p[],real xi[],int n,real *fret,real (*func)(real [],int nDims))
{
	real brent(real ax, real bx, real cx,
		real (*f)(real), real tol, real *xmin);
	real f1dim(real x);
	void mnbrak(real *ax, real *bx, real *cx, real *fa, real *fb,
		real *fc, real (*func)(real));
	int j;
	real xx,xmin,fx,fb,fa,bx,ax;

	ncom=n;
	alloc(n,&pcom);
	alloc(n,&xicom);
	nrfunc=func;
	for (j=0;j<n;j++) {
		pcom[j]=p[j];
		xicom[j]=xi[j];
	}
	ax=0.0;
	xx=1.0;
	mnbrak(&ax,&xx,&bx,&fa,&fx,&fb,f1dim);
	*fret=brent(ax,xx,bx,f1dim,TOL,&xmin);
	for (j=0;j<n;j++) {
		xi[j] *= xmin;
		p[j] += xi[j];
	}
	frei(n,&xicom);
	frei(n,&pcom);
}
#undef TOL
#undef NRANSI


/************************************************************************/
/* Version : November 24, 1996                                          */
/* Purpose : Routine for initially bracketing a minimum        		*/
/* input   : 								*/
/*	ax, bx	distinct initial points 				*/
/*	func 	function to be bracketed				*/
/* output : 								*/
/*	ax,bx,cx new points that bracket a minimum of a function	*/
/* 	fa,fb,fc function values at ax, bx ,cx				*/

#include <math.h>
#define NRANSI
#define GOLD 1.618034
#define GLIMIT 100.0
#define TINY 1.0e-20
#define SHFT(a,b,c,d) (a)=(b);(b)=(c);(c)=(d);

void mnbrak(real *ax, real *bx, real *cx, real *fa, real *fb, real *fc,
	real (*func)(real))
{
	real ulim,u,r,q,fu,dum;

	*fa=(*func)(*ax);
	*fb=(*func)(*bx);
	if (*fb > *fa) {
		SHFT(dum,*ax,*bx,dum)
		SHFT(dum,*fb,*fa,dum)
	}
	*cx=(*bx)+GOLD*(*bx-*ax);
	*fc=(*func)(*cx);
	while (*fb > *fc) {
		r=(*bx-*ax)*(*fb-*fc);
		q=(*bx-*cx)*(*fb-*fa);
		u=(*bx)-((*bx-*cx)*q-(*bx-*ax)*r)/
			(2.0*SIGN(maximum_two(fabs(q-r),TINY),q-r));
		ulim=(*bx)+GLIMIT*(*cx-*bx);
		if ((*bx-u)*(u-*cx) > 0.0) {
			fu=(*func)(u);
			if (fu < *fc) {
				*ax=(*bx);
				*bx=u;
				*fa=(*fb);
				*fb=fu;
				return;
			} else if (fu > *fb) {
				*cx=u;
				*fc=fu;
				return;
			}
			u=(*cx)+GOLD*(*cx-*bx);
			fu=(*func)(u);
		} else if ((*cx-u)*(u-ulim) > 0.0) {
			fu=(*func)(u);
			if (fu < *fc) {
				SHFT(*bx,*cx,u,*cx+GOLD*(*cx-*bx))
				SHFT(*fb,*fc,fu,(*func)(u))
			}
		} else if ((u-ulim)*(ulim-*cx) >= 0.0) {
			u=ulim;
			fu=(*func)(u);
		} else {
			u=(*cx)+GOLD*(*cx-*bx);
			fu=(*func)(u);
		}
		SHFT(*ax,*bx,*cx,u)
		SHFT(*fa,*fb,*fc,fu)
	}
}
#undef GOLD
#undef GLIMIT
#undef TINY
#undef SHFT
#undef NRANSI

/************************************************************************/
/* Version : November 24, 1996                                          */
/* Purpose : Minimization of a function of n variables. 		*/
/*	     Powell's directional set method . 				*/
/* input   : 								*/
/*      p	n-vector, initial starting point			*/
/*	xi	n,n matrix, columns contain initial set of directions   */
/*			(e.g. unit vectors)				*/
/*	ftol	RELATIVE tolerance					*/
/*	*func	function to be minimized 				*/ 
/* output : 								*/
/* 	p 	best point found					*/
/*	xi 	current direction set					*/
/* 	fret	function value at p					*/
/* on return:	iter 	number of iterations taken			*/
/* aux funcs:	linmin							*/

#include <math.h>
#define NRANSI

int Powell(real (*func)(real [],int nDims), real ftol, 
       int ITMAX, int n, real p[], real *fret, real **xi) 
{
	void linmin(real p[], real xi[], int n, real *fret,
		real (*func)(real [],int nDims));
        int iter; 
	int i,ibig,j;
	real del,fp,fptt,t,*pt=NULL,*ptt=NULL,*xit=NULL;

	alloc(n,&pt);
	alloc(n,&ptt);
	alloc(n,&xit);
	*fret=(*func)(p,n);
	for (j=0;j<n;j++) 
            pt[j]=p[j];
	for (iter=0;;++(iter)) {
        if (DEBUG) {
           fprintf(ofp,"y %f",*fret);
           printvecr(n,p,"xpoint");
        }
		fp=(*fret);
		ibig=0;
		del=0.0;
		for (i=0;i<n;i++) {
			for (j=0;j<n;j++) 
                           xit[j]=xi[j][i];
			fptt=(*fret);
			linmin(p,xit,n,fret,func);
			if (fabs(fptt-(*fret)) > del) {
				del=fabs(fptt-(*fret));
				ibig=i;
			}
		}
		if (2.0*fabs(fp-(*fret)) <= ftol*(fabs(fp)+fabs(*fret))) 
                {
			frei(n,&xit);
			frei(n,&ptt);
			frei(n,&pt);
			return iter;
		}
		if (iter == (ITMAX-1)) 
                { // exceeding maximum iterations
                   frei(n,&xit);
                   frei(n,&ptt);
                   frei(n,&pt); 
                   return ITMAX;
                   //nrerror("powell exceeding maximum iterations.");
                }
		for (j=0;j<n;j++) {
			ptt[j]=2.0*p[j]-pt[j];
			xit[j]=p[j]-pt[j];
			pt[j]=p[j];
		}
		fptt=(*func)(ptt,n);
		if (fptt < fp) {
			t=2.0*(fp-2.0*(*fret)+fptt)*SQR(fp-(*fret)-del)-del*SQR(fp-fptt);
			if (t < 0.0) {
				linmin(p,xit,n,fret,func);
				for (j=0;j<n;j++) {
					xi[j][ibig]=xi[j][n];
					xi[j][n]=xit[j];
				}
			}
		}
	}
  return iter;
}
#undef ITMAX
#undef NRANSI

/*
real objfunc(real *x, int nDims)
{
  return (1+ x[0]*x[0]) + (x[1]*x[1]);
}
*/

/*
int main()
{
 int it, d;
 real **direction=NULL, *y=NULL;
 real *x=NULL;

  d = 2;
  alloc(d,&x);
  alloc(d,d,&direction);
  alloc(1,&y);
  x[0]=0.0; x[1]=-1.0;
  y[0] = BraninFunc(x,d);  

  printvecr(d,x,"xstart");
  printvecr(1,y,"ystart");
   
  init_diag_mat(1.0,d,direction);
  printmatr(d,d,direction,"direction");
  it = Powell(BraninFunc,0.000001,200,2,x,y,direction);

  printvecr(d,x,"xstart");
  printvecr(1,y,"ystart");
  fprintf(ofp,"%d it \n",it);

  return 0;
}
*/

