#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "general.h"
#include "alloc.h"
#include "matrix.h"
#include "krig.h"

#define DEBUG 0
extern FILE *ofp;
extern SPEED;
extern RandomError;
extern PtsOutOfRange;

// MAXTERMS and MAXLINES defined in general.h
# define OUT_OF_RANGE "WARNING: Not all x elements are within range limits.\n"

/************************************************************************/
/* Version : March 07, 1997 - Nov 24, 97				*/
/* Purpose : read in  StochasticProcessModel 				*/ 
/* input format:  							*/
/* Case  Term     Theta.y  Alpha.y Theta.y2 Alpha.y2			*/
#define MISMATCH_ROWS "ERROR: Matrix contains too few or too many rows.\n"
#define MISMATCH_COLS "ERROR: Matrix contains too few or too many columns.\n"
#define MISMATCH_LABEL "ERROR: Variables in Matrix do not appear in same order as in x matrix\n" 
#define MATCHVEC_THETA "ERROR: Name of a response variable (appended to Theta) unknown\n"
#define MATCHVEC_ALPHA "ERROR: Name of a response variable (appended to Alpha) unknown\n"
int read_corpar(KrigModel *kr, KrigArray *ka,char *filename)
{
  int  i,j, newn , newm, pc;
  int k, index;
  string *label=NULL, **content=NULL;
  string *ylab=NULL; //for assemble ylab 

  if (ka==NULL)
   nrerror("read_corpar: code bug : ka==NULL"); // change (newm!= condition
  pc =  ka->p + ka->c;

  kr[0].init->theta = FALSE;
  kr[0].init->alpha = FALSE;
  kr[0].init->R = FALSE; // reading corpar has impact on R and U
  kr[0].init->U = FALSE; // if more than first corpar's read , modify!

  if (read_matrix(0,0,&content,&newn,&newm,filename,&label)!=0)
  {  if (newn*newm!=0) frei(newn,newm,&content);
     if (newm!=0)      frei(newm,&label);return 4;
  }
  if (newn!= *kr->d)
    { 
       fprintf(ofp,MISMATCH_ROWS);
       frei(newn,newm,&content);frei(newm,&label);return 5;}
  // this last command implies that theta and alpha do not need reallocation 
  if (newm!= 1+2*(ka->p+ka->c))
    {fprintf(ofp,MISMATCH_COLS);frei(newn,newm,&content);frei(newm,&label);return 1;}

  alloc(pc,&ylab);
  assemble_ylab(ka,kr,ylab);

  for (j=0; j<newn; j++)
    for (i=0; i<newm;i++)
    {
      if (strcmp(label[i],"Term")==0 || strcmp(label[i],"term")==0
         || strcmp(label[i],"Variable")==0 || strcmp(label[i],"variable")==0)
      {
        if (kr->init->X && strcmp(content[j][i],kr[0].xlabel[j])!=0)
        {fprintf(ofp,MISMATCH_LABEL);frei(newn,newm,&content);frei(newm,&label);
         frei(pc,&ylab);
         return 2; }
      }
      else if (strncmp(label[i],"Theta",5)==0 || strncmp(label[i],"theta",5)==0)
      {
          if (matchvecs(&label[i][6],pc,ylab,&index)!=0) { 
            fprintf(ofp,MATCHVEC_THETA); frei(newn,newm,&content);frei(newm,&label);
            frei(pc,&ylab); return 10; 
          }
          kr[index-1].theta[j] = atof(content[j][i]);
          kr[index-1].init->theta = TRUE;
      }
      else if (strncmp(label[i],"alpha",5)==0 || strncmp(label[i],"Alpha",5)==0)
      {
          if (matchvecs(&label[i][6],pc,ylab,&index)!=0){
            fprintf(ofp,MATCHVEC_ALPHA); frei(newn,newm,&content);frei(newm,&label);
            frei(pc,&ylab); return 11;
          }
          kr[index-1].alpha[j] = atof(content[j][i]);
          kr[index-1].init->alpha = TRUE;
      }
      else 
      {
          fprintf(ofp,"ERROR: Couldn't identify label %s .\n",label[i]);
          frei(newn,newm,&content);frei(newm,&label);frei(pc,&ylab); 
          return 12;      
      }
    }

  frei(newn,newm,&content);frei(newm,&label);
  frei(pc,&ylab);
  return 0;
}
#undef MISMATCH_ROWS
#undef MISMATCH_COLS
#undef MISMATCH_LABEL
#undef MATCHVEC_THETA
#undef MATCHVEC_ALPHA
/************************************************************************/
/* Version : March 06, 1997 						*/
/* Purpose : read in x matrix 						*/
/* Note: realloc d : X, alpha, theta 					*/
/* Note: realloc n : y, R, F, U, eigval, eigvalinv			*/
/* bug:  once n has been reassigned (e.g. minimization ) this won;t work*/
/* NOTE : if X doesn't have to be read before y need to adjust for 	*/
/* 	parallel KrigingModels						*/
// if X is no longer first thing read , copy X to other KrigModels
int read_x(KrigModel *kr,char *filename)
{
  int oldnr,newnr, newm, newn;

  if (read_matrix(*kr->n,*kr->d,&(kr->X),&newn,&newm,filename,&(kr->xlabel))!=0)
  { // error, reset to original dimensions
    morealloc(newn,newm,&(kr->X),*kr->n,*kr->d); 
    morealloc(newm,&(kr->xlabel),*kr->d); 
    return 10;
  }

  if (newm != *kr->d) 
  {
    morealloc(*kr->d,2,&kr->range,newm,2);
    morealloc(*kr->d,&kr->alpha,newm); 
    morealloc(*kr->d,&kr->theta,newm);
    init_real_vec(0.0,newm,kr->alpha);
    init_real_vec(0.0,newm,kr->theta);
    *kr->d= newm;
    kr->init->d=TRUE;
  }
  if (newn != *kr->n)
  {
    oldnr= *kr->n + *kr->r;
    newnr = newn+ *kr->r;
    morealloc(*kr->n,&kr->y,newn);
    morealloc(*kr->n,*kr->n,&kr->R,newn,newn);
    morealloc(*kr->n,*kr->r,&kr->F,newn,*kr->r);
    morealloc(oldnr,oldnr,&kr->U,newnr,newnr);
    morealloc(oldnr, &kr->eigval,newnr);
    morealloc(oldnr, &kr->eigvalinv,newnr);
    *kr->n=newn; 
    kr->init->n=TRUE;
  }
  kr[0].init->X = TRUE;
  kr[0].init->R = FALSE; // reading X has impact on R and U
  kr[0].init->U = FALSE;
  if (kr->init->X && kr->init->range)
     if (test_X_in_range(*kr->n,*kr->d,kr->X,kr->range)!=0)
     {
         fprintf(ofp,OUT_OF_RANGE);
         PtsOutOfRange = TRUE;
         // kr->init->X =FALSE;
     }
     else PtsOutOfRange = FALSE;
  // if X is no longer first thing read , copy X to other KrigModels
  return 0;
}
/************************************************************************/
/* Version : March 10, 1997                                              */
/* Purpose : read in y.mat                                              */
/*           find out whether constraints present 		        */
/*           if constraints, establish all the krigmodels               */
/*           fill in the y - vectors in each of the krigmodels          */
/* Assume:there is always a "case" column                               */
/* Assume: *kr.n is given                                               */
#define ROW_MISMATCH "ERROR: Number of rows does not match with database. Matrix ignored.\n" 
#define COL_MISMATCH "ERROR: Number of columns does not match with database. Matrix ignored.\n" 
#define LABEL_MISMATCH "ERROR: The variable names do not match with earlier variable names.\n"
#define Y_WARN "You already have a y matrix. Old y-matrix will be overread.\n"
int read_y(KrigModel **ptrkr, KrigArray *ka,char *filename)
{
  int  i,j, newn, newm;
  string *label=NULL; 
  real **content=NULL; 
  KrigModel *kr=NULL;

  kr = *ptrkr; 
  if (kr->init->y) nrwarning(Y_WARN);
  if (read_matrix(0,0,&content,&newn,&newm,filename,&label)!=0)
  { if (newn*newm!=0) frei(newn,newm,&content);
    if (newm!=0)      frei(newm,&label);return 4;
  }
  // if X doesn't need to be read first this will give error with initial n=21:
  if (kr->init->n && newn!= *kr->n)
    {fprintf(ofp,ROW_MISMATCH);frei(newn,newm,&content);frei(newm,&label);return 6;}
  if (newm<=0) nrerror("read_y: Code bug newm<=0");

  if (!ka->init)
  { //YDescription not yet read 
     // establish or change parallel KrigingModels
     if (change_krig_p(ka,ptrkr,newm)!=0)
       {frei(newn,newm,&content);frei(newm,&label);return 5;}
     kr = *ptrkr;  // may have been reallocated 
     // initialize KrigingModels after 1st one
     for (i=1;i<newm;i++)
       init_KrigModel(&kr[i],*kr[0].n,*kr[0].d,*kr[0].r);
  }
  else 
  {// YDescription was already read, check labels, (#of y variables), and n
   for (i=0;i<newm;i++)
   {
    if(strcmp(kr[i].ylabel,label[i])!=0) 
     {fprintf(ofp,LABEL_MISMATCH);frei(newn,newm,&content);frei(newm,&label);return 9;}
    *kr[i].n = newn; // newn was before checked for possible mismatch  
    kr[i].init->n = TRUE;
   }
   if (ka->init && newm != (ka->p + ka->c))
     {fprintf(ofp,COL_MISMATCH);frei(newn,newm,&content);frei(newm,&label);return 7;}
  }
  for (j=1;j<newm;j++)
  { // copy over X,n,d,range into the parallel krigmodel from 1st
    // note if oldm>1,old Kriging Models beyond first are partially overwritten
    // may want to move this once y can be read before X 
    copy_KrigModel(&kr[0],&kr[j]);
  }

  // copy labels
  for (i=0;i<newm;i++)
      strcpy(kr[i].ylabel,label[i]);
  // fill in the y - vectors in each of the krigmodels 
  // (allocation of y's in copy_KrigModel done)
  for (j=0; j<newm; j++)
  {
      for (i=0; i< *kr->n; i++)
          kr[j].y[i] = content[i][j];
      kr[j].init->y= TRUE;
  }
  //printvecr(*kr->n, kr->y,"response y");
  //printvecr(*kr->n, (*kr).y,"response y again");
  //printvecr(*kr->n, (kr[1].y),"first constraint");

  frei(newn,newm,&content); frei(newm,&label); 
  return 0;
}
#undef Y_WARN 
#undef ROW_MISMATCH
#undef COL_MISMATCH
#undef LABEL_MISMATCH
/************************************************************************/
/* Version : March 06, 1997 						*/
/* Purpose : read in RegressionModel 					*/
/*    if ka not established yet , establish it . In this case read      */
/*    y-labels into ka  						*/
/* Note :kr.F , U, eigval, eigvalinv is reallocated			*/
#define MISMATCH_COLS "ERROR: Matrix contains too many columns.\n"
#define MISMATCH_ROWS "ERROR: Number of rows in matrix does not match with database.\n"
#define MISMATCH_VAR "ERROR: Unsuspected Variable names for reponse(s).\n"
#define NO_VAR "WARNING: No names for response variables given.\n" 
int read_beta(KrigModel **ptrkr, KrigArray *ka, char *filename)
{
  int  i,j,newn, newm;
  string *label, **content;
  boolean var_read, var_error;
  KrigModel *kr=NULL;

  kr = *ptrkr;
  kr[0].init->beta = FALSE;
  var_read = FALSE;
  var_error = FALSE;
  if (read_matrix(0,0,&content,&newn,&newm,filename,&label)!=0)
  {  if (newn*newm!=0) frei(newn,newm,&content);
     if (newm!=0)      frei(newm,&label);return 4;
  }
  if (newm>3)
    {fprintf(ofp,MISMATCH_COLS);frei(newn,newm,&content);frei(newm,&label);return 4;}
  if (!ka->init)
  { // establish or change parallel KrigingModels to size newn
     if (change_krig_p(ka,ptrkr,newn)!=0)
       {frei(newn,newm,&content);frei(newm,&label);return 5;}
     kr = *ptrkr;  // may have been reallocated
     for (i=0;i<newn;i++)
     {
       if (i>0)
        init_KrigModel(&kr[i],*kr[0].n,*kr[0].d,*kr[0].r);
       kr[i].init->n =FALSE; // y hasn't been read yet
       //kr[i].init->d = FALSE; x may have been  read yet, but what if not ? 
     }
  }
  else
  {
    if (newn != (ka->p + ka->c))
     {fprintf(ofp,MISMATCH_ROWS);frei(newn,newm,&content);frei(newm,&label);return 7;}
  }

  // j is row, i is column within row 
  for (j=0; j<newn; j++)
  {
   for (i=0; i<newm;i++)
   {
    if (strcmp(label[i],"Term")==0 || strcmp(label[i],"term")==0)
    {  
      if (atoi(content[j][i])!=1) 
        nrwarning("Term should be 1 for intercept, nothing else is implemented right now");
    }
    else if (strcmp(label[i],"Variable")==0 || strcmp(label[i],"variable")==0)
    {
       var_read = TRUE;
       if (kr[j].init->y && strcmp(content[j][i],kr[j].ylabel)!=0)
       {
         var_error = TRUE;
         fprintf(ofp,MISMATCH_VAR);
       }
       else strcpy(kr[j].ylabel,content[j][i]); // as soon as ka is initialized
          // the y labels need to be established
    } 
    else if (strncmp(label[i],"Coefficients",4)==0 || 
       strncmp(label[i],"coefficients",4)==0 ||
       strncmp(label[i],"Beta",4)==0)
    { 
        kr[j].beta[0] = atof(content[j][i]);
        kr[j].init->beta = TRUE;
    }
   }
  }
  frei(newn,newm,&content);frei(newm,&label);

  if (var_error) return 21;
  if (!var_read)
    { fprintf(ofp,NO_VAR); return 22;}
  if (!kr[0].init->beta)
   nrwarning("Matrix did not contain a `beta` column and could not be read.\n");

  return 0;
}
#undef MISMATCH_COLS
#undef MISMATCH_ROWS
#undef MISMATCH_VAR 
#undef NO_VAR 

/************************************************************************/
/* Version : March 10, 1997 						*/
/* Purpose : read in yDescription.mat 					*/
/* return : 0 ok, 1 LABEL_MISMATCH					*/
#define LABEL_MISMATCH "ERROR: Label for y does not match label used earlier\n"
#define ROW_MISMATCH "ERROR: Number of rows (y variables) does not match database\n" 
#define CONSTRAINT_INPUT "ERROR: Label for column 'constraint' must be 'TRUE' or 'FALSE'.\n" 
#define COL_MISMATCH "ERROR: Number of columns in matrix does not match number of x variables\n"
#define NO_VAR "ERROR: No 'Variable' column. Cannot check variable names." 
#define NON_NUMERIC "ERROR: Some of the input data is not numeric where it ought to be.\n"
int read_ydescrip(KrigModel **ptrkr, KrigArray *ka,char *filename)
{
  int  i,j,  newn,newm;
  string *label=NULL, **content=NULL;
  boolean lower, upper, constraint, variable;
  KrigModel *kr=NULL;
  real temp_errorvar;

  kr = *ptrkr;
  if (read_matrix(0,0,&content,&newn,&newm,filename,&label)!=0)
  { if (newn*newm!=0) frei(newn,newm,&content);
    if (newm!=0)      frei(newm,&label);return 4;
  }
  if (ka->init && ka->p+ka->c!=newn)
    {fprintf(ofp,ROW_MISMATCH);frei(newn,newm,&content);frei(newm,&label);return 2;}
  if (newm <=0) nrerror("read_ydescrip: Code Bug newm<=0");

  if (!ka->init)
  { // establish or change parallel KrigingModels
     if (change_krig_p(ka,ptrkr,newn)!=0)
       {frei(newn,newm,&content);frei(newm,&label);return 5;}
     kr = *ptrkr;  // may have been reallocated
     for (i=0;i<newn;i++)
     {
       if (i>0) 
        init_KrigModel(&kr[i],*kr[0].n,*kr[0].d,*kr[0].r);
       kr[i].init->n =FALSE;
     }
  }
  else if (ka->init && newn != (ka->p + ka->c)) 
   { 
     fprintf(ofp,COL_MISMATCH); 
     frei(newn,newm,&content); frei(newm,&label); 
     return 7; 
   }

  lower=FALSE; upper=FALSE; constraint=FALSE; variable=FALSE;
  for (j=0;j<newn; j++)
  { //  first to last response variable
   temp_errorvar= -REAL_INFINITY;
   for (i=0;i<newm; i++)
   { 
    if (strcmp("StochasticProcessVariance",label[i])==0) 
    { // note : if RandomError present then errorvar must be added!
       if (str_to_real(content[j][i],&(kr[j].var[0]))!=0)
       {
         if (strcmp(content[j][i],"NA")==0 || strcmp(content[j][i],"na")==0)
          *kr[j].var = -REAL_INFINITY;
         else 
          {fprintf(ofp,NON_NUMERIC);frei(newn,newm,&content);frei(newm,&label);
           return 1;}
       }
       kr[j].init->var = TRUE;
    }
    else if (strcmp("ErrorVariance",label[i])==0)
    {//can't assume that StochVar was read already, therefore put in temp_error
      if (str_to_real(content[j][i],&temp_errorvar)!=0)
      {
        if (strcmp(content[j][i],"NA")==0 || strcmp(content[j][i],"na")==0)
        {
          temp_errorvar = -REAL_INFINITY;  // in case was overwritten
          // make sure that this is not starting value for mle 
        }
        else
          {fprintf(ofp,NON_NUMERIC);frei(newn,newm,&content);frei(newm,&label);
           return 1;}
      }
      if (temp_errorvar>0.0) 
         RandomError = TRUE;
      else if (temp_errorvar <0.0 && temp_errorvar != -REAL_INFINITY)
      {
         fprintf(ofp,"ERROR: Error Variance negative. Ignored.");
      }
    }
    else if (strcmp("LogLikelihood",label[i])==0)
    {
      if (str_to_real(content[j][i],&(kr[j].loglike[0]))!=0)
      {
        if (strcmp(content[j][i],"NA")==0 || strcmp(content[j][i],"na")==0)
           *kr[j].loglike= -REAL_INFINITY;
        else 
         {fprintf(ofp,NON_NUMERIC);frei(newn,newm,&content);frei(newm,&label);
           return 1;}
      }
    }
    else if (strcmp("Variable",label[i])==0)
    {
      if (kr[j].init->y && strcmp(kr[j].ylabel,content[j][i])!=0)
        {fprintf(ofp,LABEL_MISMATCH);frei(newn,newm,&content);frei(newm,&label);
        return 1;}
      strcpy(kr[j].ylabel,content[j][i]);
      variable=TRUE;
    }
    else if (strcmp("Min",label[i])==0 || strcmp("MIN",label[i])==0
       ||  strcmp("min",label[i])==0)
    { /* assume right order of variables */
      if (str_to_real(content[j][i],&(ka->ydescrip[j][0]))!=0)
      {
        if (strcmp(content[j][i],"NA")==0 || strcmp(content[j][i],"na")==0)
         ka->ydescrip[j][0] = -REAL_INFINITY;
        else 
          {fprintf(ofp,NON_NUMERIC);frei(newn,newm,&content);frei(newm,&label);
           return 1;}
      }
      lower = TRUE;
    }
    else if (strcmp("Max",label[i])==0 || strcmp("max",label[i])==0
       || strcmp("MAX",label[i])==0)
    {  /* assume right order of variables */
      if (str_to_real(content[j][i],&(ka->ydescrip[j][1]))!=0)
      {
        if (strcmp(content[j][i],"NA")==0 || strcmp(content[j][i],"na")==0)
          ka->ydescrip[j][1] = +REAL_INFINITY;
        else  
          {fprintf(ofp,NON_NUMERIC);frei(newn,newm,&content);frei(newm,&label);
           return 1;}
      }
      upper = TRUE;
    }
    else if (strcmp("Constraint",label[i])==0 || 
       strcmp("constraint",label[i])==0)
    { /* assume right order of variables */
        if (strcmp(content[j][i],"true")==0 || strcmp(content[j][i],"TRUE")==0)
          ka->ydescrip[j][2] = TRUE;
        else if (strcmp(content[j][i],"false")==0 
          || strcmp(content[j][i],"FALSE")==0)
          ka->ydescrip[j][2] = FALSE;
        else 
         {fprintf(ofp,CONSTRAINT_INPUT);frei(newn,newm,&content);frei(newm,&label);
          return 3;} 
        constraint =TRUE;
    }
    else if (strcmp("NumberObs",label[i])==0)
    {
       if (kr[j].init->y && atof(content[j][i]) != *kr[j].n)
         nrwarning("NumberObs does not match with database.");
    }
    else if (j==0) fprintf(ofp,"WARNING: Label %s ignored.\n",label[i]);//only once
   }
   // if RandomError adjust var and stochratio 
   if (temp_errorvar>0.0 && kr[j].init->var && *kr[j].var!= -REAL_INFINITY)
   { 
      kr[j].stochratio[0] = kr[j].var[0] / (temp_errorvar+ kr[j].var[0]);
      kr[j].var[0] += temp_errorvar; 
   }
   else if (temp_errorvar>0.0) 
   {
     fprintf(ofp,"ErrorVariance ignored since no StochasticProcess Variance present\n");
   }
  }

  if (!variable) fprintf(ofp,NO_VAR);
  if (lower && upper && constraint) ka->ydescrip_init=TRUE;
  else if (!lower && !upper && !constraint) ka->ydescrip_init=FALSE;
  else 
  { // some of upper ,lower, constraint have to be filled in
   ka->ydescrip_init=TRUE;
   if (!lower)
   {
    nrinfo("Lower bounds on responses are assumed to be -infinity");
    for (j=0;j<newn; j++)
         ka->ydescrip[j][0] = -REAL_INFINITY;
   }
   if (!upper) 
   {
    nrinfo("Upper bounds on responses are assumed to be +infinity");
    for (j=0;j<newn; j++)
         ka->ydescrip[j][1] = +REAL_INFINITY;
   }
   if (!constraint)
   {
    if (newm>1)
      nrinfo("Assuming that none of the responses is a constraint.");
    for (j=0;j<newn; j++)
         ka->ydescrip[j][2] = FALSE;
   }
  }

  adjust_KrigArray(ka); // rearrange ka->p ka->c c and p accroding to ydescrip

  frei(newn,newm,&content);frei(newm,&label);
  return 0;
}
#undef LABEL_MISMATCH
#undef ROW_MISMATCH
#undef CONSTRAINT_INPUT
#undef COL_MISMATCH
#undef NO_VAR
#undef NON_NUMERIC
/************************************************************************/
/* Version : March 06, 1997                                             */
/* Purpose : read in xdescription                                       */
/* assume  : variables are in the right order to be read 		*/
/* assume  : d was read before this, i.e. d will not be changed here    */
#define COL_MISMATCH "ERROR: Number of columns in matrix does not match number of x variables\n" 
#define LABEL_ORDER "ERROR: Variables in matrix are not in same order as in x matrix\n" 
int read_xdescrip(KrigModel *kr, KrigArray *ka, char *filename)
{
  int  i,j, d, newn ,newm;
  boolean min_col, max_col,var_label;
  string *label, **content=NULL;
  int NumberLevels ; // for later use in Joint effects etc.

  d = *kr->d;
  min_col =FALSE; max_col= FALSE; var_label= FALSE;

  if (read_matrix(0,0,&content,&newn,&newm,filename,&label)!=0)
  { if (newn*newm!=0) frei(newn,newm,&content); 
    if (newm!=0)      frei(newm,&label);return 4;
  }
  // note newn is d , not newm
  if (kr->init->X && newn!= *kr->d)
    { fprintf(ofp,COL_MISMATCH);frei(newn,newm,&content);frei(newm,&label);
      return 4;}

  for (i=0;i<newm; i++)
  {
    if (strcmp("Min",label[i])==0 || strcmp("min",label[i])==0)
    {
      /* assume right order of variables */ 
      for (j=0;j<d; j++)
        kr->range[j][0] = (atof) (content[j][i]);
      min_col = TRUE;
    }
    if (strcmp("Max",label[i])==0 || strcmp("max",label[i])==0)
    {
      /* assume right order of variables */ 
      for (j=0;j<d; j++)
        kr->range[j][1] = (atof) (content[j][i]);
      max_col = TRUE;
    }
    if (strcmp("NumberLevels",label[i])==0 || 
      strcmp("numberlevels",label[i])==0)
    {
      /* assume right order of variables */
      for (j=0;j<d; j++)
        NumberLevels = (atoi) (content[j][i]);
      // ignored for now
    }
    if (strcmp("Variable",label[i])==0 || strcmp("variable",label[i])==0)
    {
      /* check right order of variables */
      for (j=0;j<d; j++)
        if (strcmp(kr->xlabel[j],content[j][i])!=0)
          {fprintf(ofp,LABEL_ORDER);return 1;} 
      var_label = TRUE;
    }
  }
  if (min_col && max_col) kr->init->range = TRUE;
  else nrwarning("'Min' or/and  'Max' column are missing. Command ignored.");
  if (!var_label) 
   nrwarning("Assuming xdescription contains variables in same order as x matrix");
  if (kr->init->X && kr->init->range) 
     if (test_X_in_range(*kr->n,*kr->d,kr->X,kr->range)!=0)
     {
         fprintf(ofp,OUT_OF_RANGE);
         PtsOutOfRange = TRUE;
         //kr->init->range =FALSE;
     }
     else PtsOutOfRange = FALSE;
  if (ka->init && kr->init->range)
  { // copy to parallel KrigModels , needed if xdescrip read after copying 
    for (i=1;i<(ka->p+ka->c); i++) 
    {
      copy_real_matrix(d,2,kr[0].range,kr[i].range);
      kr[i].init->range = TRUE;
    }
  }

  frei(newn,newm,&content);frei(newm,&label);
  return 0;
}
#undef COL_MISMATCH
#undef LABEL_ORDER
/*************************************************************************/
// file definitions
#undef OUT_OF_RANGE
/************************************************************************/
/* Version : March 04, 1997                                             */
/* Purpose : tokenizes one line, 
             checks for unwelcome seperators,
             checks for case column 
             puts tokens into vec					*/
/* on input : 								*/
/*	line 	  (MAXLINE) char vec , to be split into tokens		*/
/*	n_vec     size of vec, hence max. terms for tokens 		*/
/*	case_check TRUE, when line to be checked for case , FALSE otherw.*/ 
/*      case_col  0 or 1 pending on whether case column present or not  */ 
/* on output:								*/
/*	vec       (n_vec) real vector containing n_terms tokens		*/	
/*      n_terms	  number of tokens 					*/	
/*      case_col  0 or 1 pending on whether case column present or not  */ 
/* return 0 ok							 	*/
/*       1 too many tokens in line 					*/
/*       4 TOKEN_ERROR							*/
/*	 5 tokenize returns with error					*/
/* NOTE : maximal lenght of label is MAXTERMS				*/
/* NOTE : if first column is Case, column is omitted in vec and n_terms */
/* NOTE : Case column has to be in first column,otherwise treated normal col*/
#define TOKEN_ERROR "ERROR: The symbols <,>,= are not allowed as part of variable names\n"
int split_token(char line[],int *n_terms,string vec[],int n_vec,int *case_col,boolean case_check)
{
  int i;
  char **token=NULL;

  alloc(n_vec,STRING_LEN,&token);
  *n_terms = n_vec;
  if (tokenize(line,n_terms,STRING_LEN,token)!=0) 
    { frei(n_vec,STRING_LEN,&token); return 5;}  
  if (*n_terms>= n_vec) 
  { 
    fprintf(ofp,"Can only handle up to %d variables in one block of data.\n",
      n_vec);
    frei(n_vec,STRING_LEN,&token);
    return 1;
  }
  for (i=0; i< *n_terms;i++)
  {  
    if (strcmp(token[i],"<")==0 || strcmp(token[i],">")==0
      || strcmp(token[i],"=")==0)
       {fprintf(ofp,TOKEN_ERROR); frei(n_vec,STRING_LEN,&token); return 4;}
    if (case_check==TRUE && i==0)
    { 
      if ((strcmp(token[i],"Case")==0 
       || strcmp(token[i],"case")==0 || strcmp(token[i],"CASE")==0))
       (*case_col)=1;
      else 
       (*case_col)=0;
    }
    if (!(i==0 && *case_col==1)) 
      strcpy(vec[i- *case_col],token[i]);  
  }
  *n_terms -= *case_col; // the case column is not counted for n_terms
  frei(n_vec,STRING_LEN,&token);
  if (DEBUG) printvecs(*n_terms,vec,"vec line");

  return 0;
}
#undef TOKEN_ERROR
/************************************************************************/
/* Version : March 04, 1997                                             */
/*           Oct26: MAXLINE_ERROR added					*/
/* Purpose : read in a string (n,m) matrix of unknown dimensions        */
/* on input : 								*/
/*	mat  (oldn,oldm) real matrix					*/	
/*	oldn : number of rows                                    	*/
/*      oldm : number of cols                                    	*/
/*	fp   : file pointer						*/
/*	newn : (1) vector, empty					*/
/*	newm : (1) vector, empty					*/
/*	label: (oldm) vector, empty					*/
/* on output :								*/ 
/*	mat  (newn,newm) matrix (also when error condition occurs)	*/
/*	newn (1) vector							*/
/*	newm (1) vector							*/
/*	label (newm) string vector of column labels			*/ 
/* on return : 0 ok, 1 OPEN_ERROR, 2 init_read error 		 	*/
/*             3 MISMATCH_COLS						*/
/*	       4 NO_COLS 5 NO_ROWS					*/
/*	       6 tokenize returns with error				*/
/*	       7 TOKEN_ERROR						*/
/*	       8 MISMATCH_ROWS						*/
/*             9 NO_HEADER						*/
/*	      10 SEP_LINE						*/
/* Note: if a string with more than 30 chars is read -> segmentation err*/
/* Note: presumes for an empty matrix labelvec is empty , ok ? ? 	*/
// need to NOT allocate xlabel and X when calling this (similarly otherprocs)
#define NO_HEADER "ERROR: There is no header line with variable names.\n"
#define MAXLINE_ERROR "ERROR: Input line contains too many characters.\n"
#define SEP_LINE  "ERROR: There is no seperator line '---' after header line.\n"
#define MISMATCH_COLS "ERROR: Too few or too many entries in a row \n"
#define MISMATCH_ROWS "ERROR: Too few or too many rows in a data block\n"
#define NO_COLS "ERROR: There are no columns in header\n"
#define NO_ROWS "ERROR: There are is no data in matrix\n"
#define TOKEN_ERROR "ERROR: The symbols <,>,= are not allowed as part of variable names\n"
#define OPEN_ERROR "ERROR: Can't open this file. Does it exist?\n"
//int read_matrix(int oldn, int oldm,real ***mat1, int *newn, int *newm,
int read_matrix(int oldn, int oldm,string ***mat1, int *newn, int *newm,
   char *filename, string **labelvec1)
{
   FILE *fp=NULL;
   char line[MAXLINE]; 
   string add_label[MAXTERMS]; // label for a given block of data
   string **mat=NULL; string *labelvec=NULL;
   //real **mat=NULL; string *labelvec=NULL;
   int i;
   int currentn; // counter within a block of data 0..newn-1
   int addm;  // m for a  given line of data  
   int currentm; // m for a  given block of data
   int offset; // to adjust current row/column number for previous blocks 
   int case_col=0; //0 if not present , 1 if present , from subproc
   boolean first_block=TRUE; // first block of data 

   fprintf(ofp,"Reading from %s.\n",filename);
   mat = *mat1; labelvec= *labelvec1;
   *newn=oldn; *newm=oldm;
   if ( (fp =fopen(filename,"r")) == NULL)
   {fprintf(ofp,OPEN_ERROR);*mat1=mat;*labelvec1 = labelvec;return 1;}
   // read prelim stuff until '---'
   while(fgets(line,MAXLINE,fp)!= NULL && !(line[0]=='-'&&line[1]=='-'));
   
   if (DEBUG) fprintf(ofp,"line1 *%s*\n",line);
   while (line!= NULL && line[0]=='-' && line[1]=='-') 
   { // read next group of variables
     // read header (modify m)
     oldm= *newm; oldn= *newn; // needed for reallocation
     if ((fgets(line,MAXLINE,fp))==NULL) 
       {fprintf(ofp,NO_HEADER);fclose(fp); *mat1=mat;*labelvec1 = labelvec;return 9;}
     else if (strlen(line)>= MAXLINE-1)
       {fprintf(ofp,MAXLINE_ERROR);fclose(fp);*mat1=mat;*labelvec1=labelvec;return 43;}
     if (DEBUG) fprintf(ofp,"line2 *%s\n",line);
     if ((split_token(line,&addm,add_label,MAXTERMS,&case_col,TRUE)) != 0) 
       {fclose(fp); *mat1 =  mat;*labelvec1 = labelvec;return 2;}
     if (fgets(line,MAXLINE,fp)== NULL || !(line[0]=='-'&&line[1]=='-'))
       {fprintf(ofp,SEP_LINE); fclose(fp); *mat1=mat;*labelvec1 = labelvec;return 10; }
     if (DEBUG) fprintf(ofp,"line3 *%s\n",line);
     if (addm==0)
       {fprintf(ofp,NO_COLS); fclose(fp); *mat1= mat;*labelvec1 = labelvec;return 4;}
     // realloc labelvec and mat
     *newm=oldm; *newn=oldn;  
     if (first_block)   {offset =0; *newm =addm;}//overwrite old data
     else               {offset = oldm; *newm +=addm;}
     if (oldm>0)                  morealloc(oldm,&labelvec,*newm); 
     else if (oldm==0)            alloc(*newm,&labelvec);
     else nrerror("read_matrix: Code Bug , Unrecognized option (1)"); 
     if (oldm>0 && oldn>0)        morealloc(oldn,oldm,&mat,*newn,*newm);  
     else if (oldm==0 && oldn>0) alloc(*newn,*newm,&mat); 
     else if (oldm>=0 && oldn==0) i=i; // do nothing 
     else nrerror("read_matrix: Code Bug, Unrecognized option (2)");
     for (i=0;i< addm;i++)
       strcpy(labelvec[offset+i],add_label[i]);

     // read data (modify n)
     currentn=0; currentm=addm;
     while (fgets(line,MAXLINE,fp) != NULL 
          && !(line[0]=='-' && line[1]=='-'))
     { // there is another line
       if (strlen(line)>= MAXLINE-1)
       {fprintf(ofp,MAXLINE_ERROR);fclose(fp);*mat1=mat;*labelvec1=labelvec;return 43;}
       if (currentn>= *newn && first_block)
       {
           oldn= *newn; oldm= *newm;
           *newn+=10;
           if (oldm>0 && oldn>0)        morealloc(oldn,oldm,&mat,*newn,*newm);
           else if (oldm==0 || oldn==0) alloc(*newn,*newm,&mat);
           else nrerror("read_matrix: Code Bug , Unrecognized option (3)");
           //!first_block check is done below, MISMATCH_ROWS
       }
       // break line up into m entries, and add to mat
       if ((split_token(line,&addm,add_label,MAXTERMS,&case_col,FALSE)) != 0)
         {fclose(fp); *mat1 =  mat;*labelvec1 = labelvec;return 2;}
       if (addm>0 && currentn>= *newn)
         {fprintf(ofp,MISMATCH_ROWS);fclose(fp);*mat1=mat;*labelvec1 = labelvec;
         return 8;} // can only happen if !first_block
       if (addm>0 && addm!= currentm) 
       {fprintf(ofp,MISMATCH_COLS);fclose(fp);*mat1=mat;*labelvec1=labelvec;return 3;}
       for (i=0;i<addm;i++)
       {
         strcpy(mat[currentn][i+offset],add_label[i]);   
         //mat[currentn][i+offset]= atof(add_label[i]);   
       }
       if (addm>0) currentn++; // line not empty
     }
     if (first_block)
     { // finalize newn  
       if (currentn==0) 
         {fprintf(ofp,NO_ROWS); fclose(fp);*mat1=mat;*labelvec1 = labelvec;return 5;}
       oldn= *newn; oldm= *newm;
       *newn= currentn;
       if (oldm>0 && oldn>0)        morealloc(oldn,oldm,&mat,*newn,*newm);
       else if (oldm==0 || oldn==0) alloc(*newn,*newm,&mat);
       else nrerror("read_matrix: Code Bug , Unrecognized option (4)");
     }
     else if (!first_block && currentn!=*newn) 
       {fprintf(ofp,MISMATCH_ROWS);fclose(fp);*mat1=mat;*labelvec1=labelvec;return 8;}
     first_block = FALSE;
   }
   *mat1 =  mat; *labelvec1 = labelvec;
   fclose(fp);
   if (*newn==0) {fprintf(ofp,NO_HEADER);return 9;}
   if (DEBUG) 
   { 
     fprintf(ofp,"newm %d newn %d\n",*newm,*newn); 
     printvecs(*newm,labelvec,"label");
     printmats(*newn,*newm,mat,"matrix");
   }
   return 0;
}
#undef NO_HEADER
#undef SEP_LINE
#undef MISMATCH_COLS
#undef NO_COLS 
#undef NO_ROWS 
#undef TOKEN_ERROR
#undef MISMATCH_ROWS
#undef OPEN_ERROR
/************************************************************************/
/* Version : March 04, 1997                                             */
/* Purpose : read in a real (n,m) matrix of unknown dimensions          */
/*	     NA values in matrix are replaced with -REAL_INFINITY       */	
/* on input : 								*/
/*	mat  (oldn,oldm) real matrix					*/	
/*	oldn : number of rows                                    	*/
/*      oldm : number of cols                                    	*/
/*	fp   : file pointer						*/
/*	newn : (1) vector, empty					*/
/*	newm : (1) vector, empty					*/
/*	label: (oldm) vector, empty					*/
/* on output :								*/ 
/*	mat  (newn,newm) matrix (also when error condition occurs)	*/
/*	newn (1) vector							*/
/*	newm (1) vector							*/
/*	label (newm) string vector of column labels			*/ 
/* on return : 0 ok, 1 OPEN_ERROR, 2 init_read error 		 	*/
/*             3 MISMATCH_COLS						*/
/*	       4 NO_COLS 5 NO_ROWS					*/
/*	       6 tokenize returns with error				*/
/*	       7 TOKEN_ERROR						*/
/*	       8 MISMATCH_ROWS						*/
/*             9 NO_HEADER						*/
/*	      10 SEP_LINE						*/
/*	      11 NOT_NUMERIC						*/
/* Note: if a string with more than 30 chars is read -> segmentation err*/
/* Note: presumes for an empty matrix labelvec is empty , ok ? ? 	*/
// need to NOT allocate xlabel and X when calling this (similarly otherprocs)
#define NO_HEADER "ERROR: There is no header line with variable names.\n"
#define MAXLINE_ERROR "ERROR: Input line contains too many characters.\n"
#define SEP_LINE  "ERROR: There is no seperator line '---' after header line.\n"
#define MISMATCH_COLS "ERROR: Too few or too many entries in a row \n"
#define MISMATCH_ROWS "ERROR: Too few or too many rows in a data block\n"
#define NO_COLS "ERROR: There are no columns in header\n"
#define NO_ROWS "ERROR: There are is no data in matrix\n"
#define TOKEN_ERROR "ERROR: The symbols <,>,= are not allowed as part of variable names\n"
#define NOT_NUMERIC "ERROR: Some of the Data cannot be interpreted as a number\n"
#define OPEN_ERROR "ERROR: Can't open this file. Does it exist?\n"
//int read_matrix(int oldn, int oldm,string ***mat1, int *newn, int *newm,
int read_matrix(int oldn, int oldm,real ***mat1, int *newn, int *newm,
   char *filename, string **labelvec1)
{
   FILE *fp=NULL;
   boolean warn_NA=FALSE;
   char line[MAXLINE]; 
   string add_label[MAXLINE];  // label for a given block of data
   //string **mat=NULL; string *labelvec=NULL;
   real **mat=NULL; string *labelvec=NULL;
   int i;
   int currentn; // counter within a block of data 0..newn-1
   int addm;  // m for a  given line of data  
   int currentm; // m for a  given block of data
   int offset; // to adjust current row/column number for previous blocks 
   int case_col=0; //0 if not present , 1 if present , from subproc
   boolean first_block=TRUE; // first block of data 

   fprintf(ofp,"Reading from %s.\n",filename);
   mat = *mat1; labelvec= *labelvec1;
   *newn=oldn; *newm=oldm;
   if ( (fp =fopen(filename,"r")) == NULL)
   {fprintf(ofp,OPEN_ERROR);*mat1=mat;*labelvec1 = labelvec; return 1;}
   // read prelim stuff until '---'
   while(fgets(line,MAXLINE,fp)!= NULL && !(line[0]=='-'&&line[1]=='-'));
   
   if (DEBUG) fprintf(ofp,"line1 *%s*\n",line); 
   while (line!= NULL && line[0]=='-' && line[1]=='-') 
   { // read next group of variables
     // read header (modify m)
     oldm= *newm; oldn= *newn; // needed for reallocation
     if ((fgets(line,MAXLINE,fp))==NULL) 
       {fprintf(ofp,NO_HEADER);fclose(fp); *mat1=mat;*labelvec1 = labelvec; return 9;}
     else if (strlen(line)>= MAXLINE-1)
       {fprintf(ofp,MAXLINE_ERROR);fclose(fp);*mat1=mat;*labelvec1=labelvec;return 43;}
     if (DEBUG) fprintf(ofp,"line2 *%s*\n",line);
     if ((split_token(line,&addm,add_label,MAXTERMS,&case_col,TRUE)) != 0) 
       {fclose(fp); *mat1 =  mat;*labelvec1 = labelvec;return 2;}
     if (fgets(line,MAXLINE,fp)== NULL || !(line[0]=='-'&&line[1]=='-'))
       {fprintf(ofp,SEP_LINE); fclose(fp); *mat1=mat;*labelvec1 = labelvec; return 10;}
     if (DEBUG) fprintf(ofp,"line3 *%s*\n",line);
     if (addm==0)
       {fprintf(ofp,NO_COLS); fclose(fp); *mat1= mat;*labelvec1 = labelvec; return 4;}
     // realloc labelvec and mat
     *newm=oldm; *newn=oldn; 
     if (first_block)   {offset =0; *newm =addm;}//overwrite old data
     else 		{offset = oldm; *newm +=addm;} 
     if (oldm>0)                  morealloc(oldm,&labelvec,*newm); 
     else if (oldm==0)            alloc(*newm,&labelvec);
     else nrerror("read_matrix: Code Bug , Unrecognized option (1)"); 
     if (oldm>0 && oldn>0)        morealloc(oldn,oldm,&mat,*newn,*newm);  
     else if (oldm==0 && oldn>0) alloc(*newn,*newm,&mat); 
     else if (oldm>=0 && oldn==0) i=i; // do nothing 
     else nrerror("read_matrix: Code Bug, Unrecognized option (2)");
     for (i=0;i< addm;i++)
       strcpy(labelvec[offset+i],add_label[i]);

     // read data (modify n)
     currentn=0; currentm=addm;
     while (fgets(line,MAXLINE,fp) != NULL 
          && !(line[0]=='-' && line[1]=='-'))
     { // there is another line
       if (strlen(line)>= MAXLINE-1)
       {fprintf(ofp,MAXLINE_ERROR);fclose(fp);*mat1=mat;*labelvec1=labelvec;return 43;}
       if (currentn>= *newn && first_block)
       {
           oldn= *newn; oldm= *newm;
           *newn+=10;
           if (oldm>0 && oldn>0)        morealloc(oldn,oldm,&mat,*newn,*newm);
           else if (oldm==0 || oldn==0) alloc(*newn,*newm,&mat);
           else nrerror("read_matrix: Code Bug , Unrecognized option (3)");
           //!first_block check is done below, MISMATCH_ROWS
       }
       // break line up into m entries, and add to mat
       if ((split_token(line,&addm,add_label,MAXTERMS,&case_col,FALSE)) != 0)
         {fclose(fp); *mat1 =  mat;*labelvec1 = labelvec;return 2;}
       if (addm>0 && currentn>= *newn) 
         {fprintf(ofp,MISMATCH_ROWS);fclose(fp);*mat1=mat;*labelvec1 = labelvec;
         return 8;} // can only happen if !first_block
       if (addm>0 && addm!= currentm) 
        {fprintf(ofp,MISMATCH_COLS);fclose(fp);*mat1=mat;*labelvec1=labelvec;return 3;}
       for (i=0;i<addm;i++)
       { 
         if (str_to_real(add_label[i],&(mat[currentn][i+offset]))!=0)   
         {
           if (strcmp("NA",add_label[i])==0 || strcmp("na",add_label[i])==0)  
           {
             mat[currentn][i+offset]= -REAL_INFINITY;
             warn_NA = TRUE;
           }
           else  
             {fprintf(ofp,NOT_NUMERIC);fclose(fp);*mat1=mat;*labelvec1=labelvec; 
              return 11;}
         }
         //strcpy(mat[currentn][i+offset],add_label[i]);   
       }
       if (addm>0) currentn++; // line not empty 
     }
     if (first_block)
     { // finalize newn  
       if (currentn==0) 
         {fprintf(ofp,NO_ROWS); fclose(fp);*mat1=mat;*labelvec1 = labelvec; return 5;}
       oldn= *newn; oldm= *newm;
       *newn= currentn;
       if (oldm>0 && oldn>0)        morealloc(oldn,oldm,&mat,*newn,*newm);
       else if (oldm==0 || oldn==0) alloc(*newn,*newm,&mat);
       else nrerror("read_matrix: Code Bug , Unrecognized option (4)");
     }
     else if (!first_block && currentn!=*newn) 
       {fprintf(ofp,MISMATCH_ROWS);fclose(fp);*mat1=mat;*labelvec1=labelvec;return 8;}
     first_block = FALSE;
   }
   *mat1 =  mat; *labelvec1 = labelvec;
   fclose(fp);
   if (*newn==0) {fprintf(ofp,NO_HEADER);return 9;}
   if (DEBUG) 
   { 
     fprintf(ofp,"newm %d newn %d\n",*newm,*newn); 
     printvecs(*newm,labelvec,"label");
     printmatr(*newn,*newm,mat,"matrix");
   }
   if (warn_NA)
     nrwarning("Matrix contains NA values. Operations on NA values other than \n\tin- and output will result into errors.\n"); 
   return 0;
}
#undef NO_HEADER
#undef SEP_LINE
#undef MISMATCH_COLS
#undef NO_COLS 
#undef NO_ROWS 
#undef TOKEN_ERROR
#undef MISMATCH_ROWS
#undef OPEN_ERROR
#undef NOT_NUMERIC
/************************************************************************/
/*
int main()
{
  int read_matrix(int oldn, int oldm,string ***mat1, int *newn, int *newm,
   char *filename, string **labelvec1);
  int read_matrix(int oldn, int oldm,real ***mat1, int *newn, int *newm,
   char *filename, string **labelvec1);
  int i=0,j=0, error;
  int oldn, oldm , newm, newn; 
  //string **mat=NULL;
  real **mat=NULL;
  string *labelvec=NULL; 
  
  oldn=4; oldm=2; 
  alloc(oldn,oldm,&mat);
  SPEED=0;
  oldn=0;oldm=0;
  error =read_matrix(oldn,oldm,&mat,&newn,&newm,"x2.mat",&labelvec);
  if (error) fprintf(ofp, "ERROR in main\n");
  fprintf(ofp,"newn %d  , newm %d \n",newn,newm);
  printmatr(newn,newm,mat,"matrix");
  printvecs(newm,labelvec,"label");
  frei(newn,newm,&mat);

  return 0;
}
*/
