/* this program creates executables for sequentialdesign of gasp 
   It can be tested as following :
      executable filename
wherefile name contains #xvar, all coordinates. Example with 2 x-var:
2 3.4 1.6
*/

#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#define CANNOT_OPEN      "Cannot open %s for reading.\n"
#define CANNOT_REOPEN    "Cannot reopen %s for writing.\n"
#define CANNOT_FIRST     "Cannot read number of x's from %s.\n"
#define CANNOT_X         "Cannot read x's from %s.\n"
#define CANNOT_WRITE     "Cannot write to %s.\n"

#define MAX_X_DIM        100

double BraninFunc(double *x, int n);
double GPFunc(double *x, int n);
double Hartman6Func(double *x,int n);
double SixHumpFunc(double *x,int n);

/****************************************/
/* Branin function.                     */
/* x1 is in [-5, 10].                   */
/* x2 is in [0, 15].                    */
/* global minimum is 0.397887357729739. */
/****************************************/

double BraninFunc( double *x, int n)
{
     double    a, b, c, d, f, ff, h, p, t, x1, x2;

     x1 = x[0];
     x2 = x[1];

     p = 3.14159265358979;

     a = 1.0;
     b = 5.1 / (4.0 * p * p);
     c = 5.0 / p;
     d = 6.0;
     h = 10.0;
     ff= 1.0 / (8 * p);
     t = x2 - b * x1 * x1 + c * x1 - d;
     f = a * t * t + h * (1.0 - ff) * cos(x1) + h;

     return f;
}

/****************************************/
/* Goldstein Price function.            */
/* x1 is in [-2,2].                     */
/* x2 is in [-2,2].                     */
/* global minimum 3.0000000000          */
/****************************************/

double GPFunc(double *x,int n )
{
  double power(double z, int n); 
  double x1, x2 , sqrx1, sqrx2, f; 

  x1 = x[0];
  x2 = x[1];
  sqrx1 = x1 * x1;
  sqrx2 = x2 * x2;
  

  f = ( 1.0 + power((x1+x2+1.0),2)  *
      (19.0 -14.0*x1+ 3.0*sqrx1 -14.0*x2 +6.0*x1*x2 +3.0*sqrx2 ) ) *
      (30.0 + power((2.0*x1-3.0*x2),2)   *
      (18.0 -32.0*x1 +12.0*sqrx1 +48.0*x2 -36.0*x1*x2 +27.0*sqrx2)) ;

  return(f);
}

/****************************************/
/* Hartman 6 function.                  */
/* x1...x6 are in [0,1].                */
/* global minimum  -3.32236801141551    */
/****************************************/
/* note : matrices indices in Fortran work exactly the other way around */


double Hartman6Func(double *x,int n )
{
   double power(double z, int n);
   int i,j;
   double sumi, sumj,f;
   double cvec[4] = { 1.0,1.2,3.0,3.2};
   double a[6][4] = { 
{10.00,  0.0,   3.00, 17.00},
{ 3.00, 10.00,  3.50,  8.00},
{17.00, 17.00,  1.70,  0.05},
{ 3.50,  0.10, 10.00, 10.00},
{ 1.70,  8.00, 17.00,  0.10},
{ 8.00, 14.00,  8.00, 14.00} };
   double p[6][4] = {
 {0.1312, 0.2329, 0.2348, 0.4047}, 
 {0.1696, 0.4135, 0.1451, 0.8828},
 {0.5569, 0.8307, 0.3522, 0.8732},
 {0.0124, 0.3736, 0.2883, 0.5743},
 {0.8283, 0.1004, 0.3047, 0.1091},
 {0.5886, 0.9991, 0.6650, 0.0381}};

   sumi = 0.0;
   for (i=0; i<4; i++)
   {
     sumj = 0.0;
     for (j=0; j<6; j++)
     {
        sumj += a[j][i]* power((x[j]-p[j][i]),2);
     }
     sumi += cvec[i] * exp(-sumj);
   }
   f= -sumi;
   return(f);
}

/****************************************/
/* SixHump function.                    */
/* x1...x2 are in [-5,5].               */
/* two global minima and 4 local minima */
/****************************************/
double SixHumpFunc(double *x,int n)
{
   double x1,x2,f ;

   x1 = x[0];
   x2 = x[1];
   f  = 4.0*x1*x1 - 2.1*pow(x1,4.0) + pow(x1,6.0)/3.0 + x1*x2 - 4.0*x2*x2 
        + 4.0 * pow(x2,4.0);
   return(f);
}

double power(double z, int n)
{
  int i;
  double result;
  
  result= 1.0;  
  for (i=0;i<n;i++)
     result*=z;
  return(result);
}

/****************************************/
/*
void main()
{
 double x[100];
  double temp ;


 x[0]=.0;
 x[1]=.8;
 x[2]=.1;
 x[3]=.3;
 x[4]=.22;
 x[5]=.5;
 temp = log(-Hartman6Func(6,x));
 printf("%f ",temp);
}
*/
