#include <stdio.h>
#include <math.h>
#include <stdlib.h> /* needed for fabs*/
#include <string.h> /* needed for strcmp*/

#include "general.h"
#include "alloc.h"
#include "matrix.h"

static real alpha=0.01;     /* expansion factor for "exp" */
static real alpha_sqr=10.0; /* expansion factor for "square" */

extern FILE *ofp;

/* note : any changes should be verified with krigfit tranformations !*/
/************************************************************************/
/* Version : June 19, 1996 - Jan 27,97					*/
/* Purpose : transform n points in m dimensions from			*/ 
/*           [a_i,b_i] to [-infty,+infty] where i=1..d 			*/
/* 'exp'Transformation:  log( [(x-a)/(b-a)] / [1- (x-a)/(b-a)]) /alpha	*/   
/* 'sin'Transformation:  arcsin[2(x-a)/(b-a) -1]  			*/   
/* 'square'Transformation: sqrt(alpha_sqr*x)				*/
/* on input: x : n,d matrix (untranformed)				*/
/*           range : d,2 matrix of upper and lower limits for each dimesion*/ 
/* output  : trafox : n,d matrix, tranformed 				*/     
/* Note 'square': square only when [a_i,b_i]=[0,+infinity]:		*/
/* Note 'square': range can be NULL pointer				*/
/* Note 'sin': transforms to [-PI/2,+PI/2], because that's 1 full cyclefor sin*/
/* Note 'sin': a bit time consuming, especially for ab_inf 		*/ 
/* Note 'exp': with alpha=1, it's almost impossible to tranform beyond  */
/*	     (-15,+15) exept at the endpoints				*/
/* Note 'exp': no transformation past +-INIFINTY (10^7 or so )          */
/* //old'sin'Transformation:  [(x-a)/(b-a) -.5] * PI 			*/   
void trafo_ab_inf(int n,int d,real **x,real **range,real **trafox,string method)
{
  int i,j;
  real temp;
   
  // may want to include a check that x is indeed within range

  if (strcmp(method,"square")!=0)
  { // 'square' doesn't have a range 
    for (j=0; j<d; j++)
      if (range[j][0] >= range[j][1]) 
       nrerror("trafo_ab_inf: their are range intervals [a,b] with a>=b");
  }
   
  if (strcmp(method,"exp")==0)
  {
   for (i=0; i<n; i++)
    for (j=0; j<d; j++)
    {
      if (x[i][j]< range[j][0]  || x[i][j]> range[j][1])
      {
        fprintf(ofp,"x:  %f lower: %f upper: %f \n",x[i][j],range[j][0],range[j][1]);
        nrerror("trafo_ab_inf: Matrix has elements out of [a,b] range ");
      }
      temp = (x[i][j]-range[j][0])/(range[j][1]-range[j][0]);    
      if (temp == 1.0)
         trafox[i][j] = INFINITY;
      else if (temp == 0.0) trafox[i][j] = -INFINITY;
           else trafox[i][j] = log( temp/(1.0-temp)) / alpha;

      if (fabs(trafox[i][j]) > INFINITY) 
      {
         if (trafox[i][j] > INFINITY) trafox[i][j]= INFINITY;
         else                         trafox[i][j]= -INFINITY;
      }
    }
  }
  else if (strcmp(method,"sin")==0)
  {
   for (i=0; i<n; i++)
    for (j=0; j<d; j++)
    {
      if (x[i][j]< range[j][0]  || x[i][j]> range[j][1])
      {
        fprintf(ofp,"x:  %f lower: %f upper: %f \n",x[i][j],range[j][0],
           range[j][1]);
        nrerror("trafo_ab_inf: Matrix has elements out of [a,b] range ");
      }
      temp = 2.0 * (x[i][j]-range[j][0])/(range[j][1]-range[j][0]) -1.0;
      if (temp>= -1.0 && temp<=1.0)
        trafox[i][j] = asin(temp);
      else 
      {
        fprintf(ofp,"arcsin argument %f \n",temp);
        nrerror("krigfit, trafo_corpar: arcsin argument out of range");
      }
    }
  } 
  else if (strcmp(method,"square")==0)
  {
    for (i=0; i<n; i++)
      for (j=0; j<d; j++)
      {
        if (x[i][j]<0.0)
          nrerror("trafo_ab_inf: Attempt of (inverse) transformation 'square' with x<0"); 
        else 
           trafox[i][j] = sqrt(alpha_sqr *x[i][j]);
      }
  }
  else nrerror("trafo_ab_inf: Not a valid transformation method.");
}
/************************************************************************/
/* Version : July 16, 1996						*/
/* Purpose : transform n points in m dimensions from			*/ 
/*           [-infty,+infty] to [a_i,b_i] where i=1..d 			*/
/* 'exp'Transformation:  [ e^(alpha*y) (b-a) / (1+ e^(alpha*y) ] +a	*/   
/* /'sin'Transformation:  [(sin(y)+1)/2.0*(b-a)] + a 			*/
/* 'square'Transformation: x*x / alpha_sqr				*/
/* Note : square trafo only for [a_i,b_i] = [0,+infinity]		*/
/* on input: x : n,d matrix (untranformed)				*/
/*           range : d,2 matrix of upper and lower limits for each dimension*/ 
/* output  : trafox : n,d matrix, tranformed 				*/     
/* Note    : anything beyond about [-10,10] is transformed to the limits*/
/*           a or b       						*/ 
void trafo_inf_ab(int n,int d,real **x,real **range,real **trafox,string method)
{
 int i,j;
 real temp;

  if (strcmp(method,"square")!=0)
  {
    for (j=0; j<d; j++)
      if (range[j][0] >= range[j][1]) 
         nrerror("trafo_inf_to_ab: their are range intervals [a,b] with a>=b");
  }

  if (strcmp(method,"exp")==0)
  {
   for (i=0; i<n; i++)
    for (j=0; j<d; j++)
    {
       temp = exp(alpha*x[i][j]); 
       trafox[i][j] = (temp/(1+temp) *(range[j][1]-range[j][0])) + range[j][0]; 
       /* leave temp/(1+temp) together , rearranging may cause floating error*/
    }
  }
  else if (strcmp(method,"sin")==0)
  {
   for (i=0; i<n; i++)
    for (j=0; j<d; j++)
     trafox[i][j]=(sin(x[i][j])+1)/2.0*(range[j][1]-range[j][0]) + range[j][0];
  }
  else if (strcmp(method,"square")==0)
  {
    for (i=0; i<n; i++)
    for (j=0; j<d; j++)
     trafox[i][j]= x[i][j] * x[i][j] / alpha_sqr;
  }
  else nrerror("trafo_inf_ab: Not a valid transformation method.");
}
/**************************************************************/
/*
void main()
{
int n,d ;    
real **range, **trafox, **x;

d=2;
n=2;
alloc(n,d,&trafox);
alloc(n,d,&x);  
alloc(d,2,&range);

range[0][0] = 0.0;
range[0][1] = 2.0;
range[1][0] = 1.0;
range[1][1] = 10.0;
x[0][0]= 0.0;
x[0][1]= 5.0;
x[1][0]= 1.0;
x[1][1]= 10.0;

printmatr(n,d,x,"x");
trafo_ab_inf(n,d,x,range,trafox,"sin");
printmatr(n,d,trafox,"trafox");


}
*/
