#include <stdio.h>
#include <string.h>
#include <stdlib.h>  


#include "general.h"
#include "alloc.h"
#include "matrix.h"
#include "krig.h"

extern real VERSION;
extern FILE *ofp;

//#define DEBUG_ALLOC

/* NOTE: output option 2  currently used in : 
   "F"  , "y" ,"    Term     Value       Effect       StandardError"
   and joint effects
*/

/************************************************************************/
/* Version : Feb 25, 1997                                               */
/* Purpose : output real matrix						*/
/* exact : TRUE   15 digits, FALSE (default) 5digits                    */
/* Note: uses array of strings for column labels 			*/
/* Note: when varnames=NULL, then varnames2 is used			*/
/* NOTE : when entry == -REAL_INFINITY this means its not available (NA)*/  
#define MAX(a,b) ((b) > (a) ? (b) : (a))
void outputmatr(int n, int m, real **matrix,char matname[],
     string varnames[], char varnames2[], char comment[],
     char mode, boolean exact) 
{
  extern FILE *ofp;
  FILE *fp;
  int i,j,k, temp; 
  int len_line; //length of the header of an output block
  int start_m, end_m;//index(0..m-1) of start and end variable in a output block
  int items_per_line = 6;

  
  /* file status check*/
  if (mode=='a')
    fp =fopen(matname,"a");
  else
    fp =fopen(matname,"w");
  if (fp == NULL)
  {
     fprintf(stderr,"\n writemat.c: Can't open/create file %s. Exit(1) \n",
     matname);
     exit(1);
  }

  /* overhead output */
  fprintf(ofp,"Writing to file %s.\n",matname);
  //cout << "Writing to file " << matname << "." << endl;
  if (mode!='a')
  {
    fprintf(fp,"Version %5.3f \n",VERSION);
    fprintf(fp,"%s \n",comment);
  }

  end_m=0;
  while (end_m < m)
  { // while not all vars are printed
    start_m=end_m;

    // define end_m
    if (varnames!=NULL)
    {
      i=6; j=start_m; // take case column = 10 characters
      while (j<m && (i+= 2+ MAX(strlen(varnames[j]),10)) <75)
      j++; // 2+ is due to the extra spaces in front and after a name
      end_m = j; 
      if (j<m) len_line= i- (2+MAX(strlen(varnames[j]),10));
      else     len_line= i;
    }
    else 
    {start_m=0;end_m=m;len_line = 6+ strlen(varnames2);} // other output option
   
    fprintf(fp,"\n");
    for (i=0;i<len_line;i++) fprintf(fp,"-");
    fprintf(fp,"\n Case ");
    if (varnames!=NULL)
    {
      for (i=start_m;i<end_m;i++)
        fprintf(fp," %10s ",varnames[i]);
    }
    else if (start_m==0)
    { // more than one block, only print in first 
      fprintf(fp,"%s",varnames2);
    }
    fprintf(fp,"\n");
    for (i=0;i<len_line;i++) fprintf(fp,"-");
    fprintf(fp,"\n");

    for (i=0; i<n; i++)
    {
      fprintf(fp," %4d ",i+1);
      for (j=start_m; j<end_m; j++)
      {
        #ifdef DEBUG_ALLOC
        check_alloc(matrix,i,j,"outputmatr");
        #endif
        if (varnames!=NULL && strlen(varnames[j])>10)
        { // need extra space since name is large
          temp = strlen(varnames[j])-10;
          for (k=0;k<temp;k++)
            fprintf(fp," ");
        }
        if (matrix[i][j]!= -REAL_INFINITY)
        {
          if (!exact) fprintf(fp," %10.5f ",matrix[i][j]);
          else        fprintf(fp," %10.15f ",matrix[i][j]);
        } 
        else 
          fprintf(fp,"         NA "); //-REAL_INFINITY means NA by convention 
      }
      fprintf(fp," \n");
    }
  }
  fclose(fp);
}
#undef MAX
/************************************************************************/
/* Version : Feb 25, 1997                                               */
/* Purpose : output string matrix					*/
/* exact : TRUE   15 digits, FALSE (default) 5digits                    */
/* Note: uses array of strings for column labels                        */
/* Note: when varnames=NULL, then varnames2 is used                     */
#define MAX(a,b) ((b) > (a) ? (b) : (a))
void outputmats(int n, int m,string **matrix,char matname[],
     string varnames[], char varnames2[], char comment[],
     char mode, boolean exact) 
{
  extern FILE *ofp;
  FILE *fp;
  int i,j,k, temp, start_m, end_m, len_line ;
  int items_per_line = 6;

  /* file status check */
  if (mode =='a')
    fp =fopen(matname,"a");
  else
    fp =fopen(matname,"w");
  if (fp == NULL)
  {
     fprintf(stderr,"\n writemat.c: Can't open/create file %s. Exit(1) \n",
     matname);
     exit(1);
  }

  /* overhead output */
  fprintf(ofp,"Writing to file %s.\n",matname); 
  //cout << "Writing to file " << matname << "." << endl;
  fprintf(fp,"Version %5.3f \n",VERSION);
  fprintf(fp,"%s \n",comment);

  end_m=0;
  while (end_m < m)
  {
    start_m=end_m;

    // define end_m
    if (varnames!=NULL)
    {
    i=6; j=start_m; // take case column = 10 characters
    while (j<m && (i+= 2+ MAX(strlen(varnames[j]),10)) <75)
      j++; // 2+ is due to the extra spaces in front and after a name
    end_m = j;
    if (j<m) len_line= i- (2+MAX(strlen(varnames[j]),10));
    else     len_line= i;
    }
    else
    {start_m=0;end_m=m;len_line = 6+ strlen(varnames2);} // other output option


    for (i=0;i<len_line;i++) fprintf(fp,"-");
    fprintf(fp,"\n Case ");
    if (varnames!=NULL)
    {
      for (i=start_m;i<end_m;i++)
        fprintf(fp," %10s ",varnames[i]);
    }
    else if (start_m==0)
    { // more than one block, only print in first
      fprintf(fp,"%s",varnames2);
    }
    fprintf(fp,"\n");
    for (i=0;i<len_line;i++) fprintf(fp,"-");
    fprintf(fp,"\n");

    for (i=0; i<n; i++)
    {
      fprintf(fp," %4d ",i+1);
      for (j=start_m; j<end_m; j++)
      {
        if (varnames!=NULL && strlen(varnames[j])>10)
        { // need extra space since name is large
          temp = strlen(varnames[j])-10;
          for (k=0;k<temp;k++)
            fprintf(fp," ");
        }
        if (!exact) fprintf(fp," %10s ",matrix[i][j]);
        else        fprintf(fp," %10s ",matrix[i][j]);
      }
      fprintf(fp," \n");
    }
  }
  fclose(fp);
}
#undef MAX
/************************************************************************/
/* Version : July 10, 1996 						*/
/* outputs the vector into one row 					*/
/* input :  char mode 'a' appends, otherwise writes			*/
void outputvecr(int n, real *vector,char vecname[], 
      char varnames[], char comment[],char mode)
{
  FILE *fp;
  int i;
   
  if (mode=='a')
    fp =fopen(vecname,"a"); 
  else 
    fp =fopen(vecname,"w");
  if (fp == NULL)
  {
     fprintf(stderr,"\n writevec.c: Can't open/create file %s. Exit(1) \n",
     vecname);
     exit(1);
  }

  fprintf(fp,"Version %5.3f \n",VERSION); 
  fprintf(fp,"%s \n",comment);
  fprintf(fp,"-------");
  fprintf(fp,"\n Case %s \n",varnames);
  fprintf(fp,"-------");
  fprintf(fp,"\n");

  #ifdef DEBUG_ALLOC
  check_alloc(vector,0,"outputmatr");
  check_alloc(vector,n-1,"outputmatr");
  #endif
  for (i=0; i<n; i++)
     fprintf(fp,"  %2d   %10.5f \n",i+1,vector[i]);
  fclose(fp);
}
/************************************************************************/
/* Version : July 11, 1996                                              */
/* Purpose : output range.mat						*/
/* Note    : proc assumes range is initialized 				*/
void output_range(KrigModel *kr, char name[],char *comment)
{
  FILE *fp;
  int i;

  if ( (fp =fopen(name,"w")) == NULL)
     nrerror("output_summary: Can't open/create file corpar.mat.");

  fprintf(fp,"Version %5.3f \n",VERSION);
  fprintf(fp,"%s \n",comment);
  fprintf(fp,"----------------------------------------------------------\n");
  fprintf(fp," Case Variable Min        Max\n"); 
  fprintf(fp,"----------------------------------------------------------\n");
  for (i=0; i< *kr->d; i++)
    fprintf(fp,"  %d     x%d      %f  %f \n",i+1,i+1,kr->range[i][0],
     kr->range[i][1]);
  fclose(fp);
}
/************************************************************************/
/* Version : May 06, 1997                                               */
/* Purpose : output and print (screen) 					*/
void outputmatr_cover(int n_row,int n_col,real **matrix,char filename[] ,
   string *label, string comment)
{
  extern FILE *ofp;

     if (strcmp(filename,"screen")!=0 && strcmp(filename,"s")!=0 )
       outputmatr(n_row,n_col,matrix,filename,label,NULL,comment); 
     else if (n_row < 22)
     {
       printmatr(n_row,n_col,matrix,"sorry, no labels at this time");
     }
     else
        fprintf(ofp,"Sorry, Matrix is to big to be displayed on screen\n");
}
/************************************************************************/
/* Version : Nov 15, 1997                                               */
/* Purpose : output and print (screen)                                  */
void outputmats_cover(int n_row,int n_col,string **matrix,char filename[],
   string *label, string comment)
{
  extern FILE *ofp;

     if (strcmp(filename,"screen")!=0 && strcmp(filename,"s")!=0 )
       outputmats(n_row,n_col,matrix,filename,label,NULL,comment);
     else if (n_row < 22)
     {
       printmats(n_row,n_col,matrix,"Sorry, no labels at this time");
     }
     else
        fprintf(ofp,"Sorry, Matrix is to big to be displayed on screen\n");
}

